# Copyright(c) 2002-2004 Association of Universities for Research in Astronomy, Inc.

procedure gdisplay(image,frame)

# Display GMOS (raw) images
#
# Version Feb 28, 2002   IJ  v1.3 release
#         Apr 26, 2002   IJ  change warning comment, handle [1] images
#         Jul  7, 2002   IJ  handle dir-names+short image names correctly
#         Sept 20, 2002  IJ   v1.4 release
#         Oct 17, 2002   BM  generalized for GMOS-S, 6amps, and all readmodes
#         Oct 22, 2002   BM  fix problem with defining l_gainmode when fl_bias-
#         Jan 31, 2003   BM  Added output file
#                            
char image      {"",prompt="GMOS image to display"}                            # OLDP-1-input-primary-none
int  frame      {1,prompt="Frame to write to"}                                 # OLDP-1
char output     {"",prompt="Save pasted file to this name if not blank"}       # OLDP-3
bool fl_paste   {no,prompt="Paste images to one for imexamine"}                # OLDP-2
bool fl_bias    {no,prompt="Rough bias subtraction"}                           # OLDP-2
int  gap        {37,prompt="Size of the gap between the CCDs (in pixels)"}     # OLDP-3
real z1         {0.,prompt="Lower limit if not autoscaling"}                   # OLDP-2
real z2         {0.,prompt="Upper limit if not autoscaling"}                   # OLDP-2
char sci_ext   {"all",prompt="Name of extension(s) to display"}                # OLDP-3
char key_detsec  {"DETSEC",prompt="Header keyword for detector section"}       # OLDP-3
char key_datasec  {"DATASEC",prompt="Header keyword for data section"}         # OLDP-3
char key_ccdsum  {"CCDSUM",prompt="Header keyword for CCD binning"}            # OLDP-3
string gaindb       {"gmos$data/gmosamps.dat",prompt="Database with gain data"} # OLDP-2
bool   verbose   {no,prompt="Verbose"}                                         # OLDP-4
struct* scanfile {"",prompt="Internal use only."}                              # OLDP-4

begin

char l_image, l_sci_ext, l_key_detsec, l_key_datsec, l_key_ccdsum
char l_gaindb
bool l_verbose, l_fl_paste, l_fl_bias
int l_gap, l_frame
real l_z1, l_z2

char tmpin, l_dsec, l_datsec, l_stdimage, junk, tmpout, l_paste
char tmpim, l_imext, tmpdisp, l_output
char l_pixtype
int n_ext, Xmin, Xmax, Ymin, Ymax, x1, x2, y1, y2, n_i, n_j, n_k
int Xoff, Yoff, Xbin, Ybin
real l_xcenter, l_ycenter, l_mag
bool fl_erase, fl_zscale, fl_zrange, l_fl_mag, l_fl_chmag, l_fl_sec
struct l_struct
real l_bias[6],bias,l_gain,l_rdnoise
string l_readmode,l_gainmode,dum,s_bias
int  l_igain

l_image=image ; l_sci_ext=sci_ext ; l_key_detsec=key_detsec
l_key_datsec=key_datasec ; l_key_ccdsum=key_ccdsum ; l_z1=z1 ; l_z2=z2
l_verbose=verbose ; l_gap=gap ; l_frame=frame ; l_fl_paste=fl_paste
Xmin=1 ; Xmax=6144+2.*l_gap; Ymin=1; Ymax=4608
l_fl_bias=fl_bias
l_gaindb=gaindb
l_output=output

#tmpin="uparm$"//mktemp("tmpin")
tmpin=mktemp("tmpin")
tmpout=mktemp("tmpout")
tmpdisp=mktemp("tmpdisp")
tmpim=mktemp("tmpim1")

fl_zscale=yes ; fl_zrange=yes
if(l_z1!=0. || l_z2!=0.) {
  fl_zscale=no ; fl_zrange=no
}

cache("imgets","gimverify","fparse")
date | scan(l_struct)

gimverify(l_image)
if(gimverify.status==1) {
  print("ERROR - GDISPLAY: Cannot access image "//l_image)
  bye
}
if(gimverify.status!=0) {
  print("ERROR - GDISPLAY: Image "//l_image//" is not a MEF file")
  bye
}
fparse(gimverify.outname)
l_imext=fparse.root ; tmpout=tmpout//l_imext
fparse(l_image)
if (fparse.extension != ".fits") {
   l_image=l_image//".fits"
}
if(fl_paste) 
print("GDISPLAY: Temporary image name "//tmpout)

show stdimage | scan(l_stdimage)
printf("stdimage = %s\n",l_stdimage)
if(substr(l_stdimage,1,7)=="imtgmos" && l_stdimage!="imtgmosccd") {
#  l_imX=6400 ; l_imY=4644
  Xoff=nint(6400.-Xmax)/2.
  Yoff=nint(4644.-Ymax)/2.
  Xmax=6400 ; Ymax=4644
} else {
 print("ERROR - GDISPLAY: Please set stdimage to a full display GMOS device")
 print("                  imtgmos, imtgmos2, imtgmos4 or imtgmos8")
 bye
}
l_mag=1.
if(strlen(l_stdimage)==8) 
  l_mag=1./real(substr(l_stdimage,8,8))

if(!l_fl_paste && l_fl_bias) 
  print("WARNING - GDISPLAY: Cannot bias subtract if not pasting")

if(l_sci_ext!="all") {
  fxhead(l_image) | match("IMAGE","STDIN",stop-) | \
    match(l_sci_ext,"STDIN",stop-, > tmpin)
} else {
  n_k=0
  fxhead(l_image) | match("IMAGE","STDIN",stop-) | \
    match("SCI","STDIN",stop-) | scan(n_k)
  if(n_k>0) {
    fxhead(l_image) | match("IMAGE","STDIN",stop-) | \
    match("SCI","STDIN",stop-, > tmpin)
    print("Displaying the SCI extensions")
    l_sci_ext="SCI"
  } else {
    fxhead(l_image) | match("IMAGE","STDIN",stop-, > tmpin)
    # raw data
    l_sci_ext=""
  }
}

count(tmpin) | scan(n_ext)  # number of extensions

delete("uparm$gdisplay"//str(l_frame),verify-, >>& "dev$null")
#delete("uparm$mscdisp"//str(l_frame),verify-, >>& "dev$null")

l_fl_chmag=yes  # check magnification on first round
l_fl_mag=no     # initialize check for changed magnification

s_bias=""
n_k=0 ; l_fl_sec=yes
scanfile=tmpin
while(fscan(scanfile,n_i)!=EOF) {
 n_k=n_k+1
# Get binning
 imgets(l_image//"["//str(n_i)//"]",l_key_ccdsum)
 if(imgets.value=="0") {
   Xbin=1 ; Ybin=1
 } else {
   print(imgets.value) | \
    fields("STDIN","1,2",lines="1",quit_if_miss-,print_file-) | \
    scan(Xbin,Ybin)
 }

 imgets(l_image//"["//str(n_i)//"]",l_key_detsec, >>& "dev$null")
 l_dsec=imgets.value
 imgets(l_image//"["//str(n_i)//"]",l_key_datsec, >>& "dev$null")
 l_datsec=imgets.value
if(l_dsec=="0" || l_datsec=="0") 
 l_fl_sec=no

# Get the readout mode  (slow/fast lowgain modes only)
if (l_fl_bias) {
 if(n_k==1) {
     imgets(l_image//"[0]","AMPINTEG", >>& "dev$null")
     if(imgets.value=="1000")
     l_readmode="fast"
     else
     l_readmode="slow"

     # Set gainmode (high or low)  
     # If GAINORG is there use it, otherwise use GAIN
     keypar(l_image//"[1]","GAINORIG",silent+)
     if(!keypar.found) 
         imgets(l_image//"[1]","GAIN", >>& "dev$null")
     else
         imgets(l_image//"[1]","GAINORIG", >>& "dev$null")	 

     if( real(imgets.value)>3. )
         l_gainmode="high"
     else
         l_gainmode="low"

 }

 imgets(l_image//"["//n_i//"]","AMPNAME", >>& "dev$null" )
 l_gain=0.0 ; l_rdnoise=0.0
 match(l_readmode,l_gaindb,stop-) | \
 match(l_gainmode,"STDIN",stop-) | \
 match(imgets.value,"STDIN",stop-) | scan(dum,dum,l_gain,l_rdnoise,bias,amp)
 l_bias[n_i]=bias
} else {
    l_bias[n_i]=0.0
}

if((l_fl_paste || l_fl_bias) && !l_fl_sec) {
  print("WARNING - GDISPLAY: No data section info found, or image is already mosaiced")
  print("                    Displaying input image. Using fl_paste=no fl_bias=no")
  l_fl_paste=no ; l_fl_bias=no
}

if(!l_fl_paste) {

 if(l_fl_sec) {
# find x1, x2, y1, y2 and set display parameters
 n_j=stridx(":",l_dsec)
 x1=int(substr(l_dsec,2,n_j-1))
 l_dsec=substr(l_dsec,n_j+1,strlen(l_dsec))
 n_j=stridx(",",l_dsec)
 x2=int(substr(l_dsec,1,n_j-1))
 l_dsec=substr(l_dsec,n_j+1,strlen(l_dsec))
 n_j=stridx(":",l_dsec)
 y1=int(substr(l_dsec,1,n_j-1))
 l_dsec=substr(l_dsec,n_j+1,strlen(l_dsec))
 y2=int(substr(l_dsec,1,strlen(l_dsec)-1))
 } else {
    x1=1 ; y1=1
    imgets(l_image//"["//str(n_i)//"]","i_naxis1")
    x2=int(imgets.value)*Xbin
    imgets(l_image//"["//str(n_i)//"]","i_naxis2")
    y2=int(imgets.value)*Ybin
 }

 printf("%s %4d %4d %4d %4d\n",l_image//"["//str(n_i)//"]",x1,x2,y1,y2, >> "uparm$gdisplay"//str(l_frame))
# printf("%s %4d %4d %4d %4d\n",l_image//"["//str(n_i)//"]",x1,x2,y1,y2, >> "uparm$mscdisp"//str(l_frame))

if(n_k==1)
 fl_erase=yes
else
 fl_erase=no

l_xcenter=(x2-x1)/2.+x1
if(l_xcenter>2048)
  l_xcenter+=l_gap
if(l_xcenter>(2*2048.+l_gap))
  l_xcenter+=l_gap
l_xcenter+=Xoff

l_xcenter=l_xcenter/Xmax
l_ycenter=((y2-y1)/2.+y1+Yoff)/Ymax
#print(l_xcenter,l_ycenter)

# Reset xcenter,ycenter if image is binned
  l_xcenter = (l_xcenter-0.5)/Xbin + 0.5
  l_ycenter = (l_ycenter-0.5)/Ybin + 0.5

# Reset magnification if higher resolution can be displayed
# Never display more than 1pix to 1pix
  if(min(Xbin,Ybin)>1. && l_mag*min(Xbin,Ybin) <= 1. && l_fl_chmag) {
     l_mag=l_mag*min(Xbin,Ybin)
     l_fl_chmag=no  # don't check/change l_mag in the next rounds
     l_fl_mag=yes   # update xcenter,ycenter in all rounds
  }
# Adjust xcenter,ycenter if l_mag has been changed
  if(l_fl_mag) {
     l_xcenter = (l_xcenter-0.5)*min(Xbin,Ybin) + 0.5
     l_ycenter = (l_ycenter-0.5)*min(Xbin,Ybin) + 0.5
  } 

 printf("%s %4d %4d %4d %4d  %5.1f ",l_image//"["//str(n_i)//"]",x1,x2,y1,y2,l_mag)
 display(l_image//"["//str(n_i)//"]"//l_datsec,l_frame,erase=fl_erase,
  xcenter=l_xcenter,ycenter=l_ycenter,z1=l_z1,z2=l_z2,zrange=fl_zrange,
  zscale=fl_zscale,xmag=l_mag,ymag=l_mag,fill=no)

} else if (l_fl_sec) {
# Setup parameters for imtile
  if(n_k==1) {
      if (l_fl_bias) {
          copy(l_image,tmpim//".fits",verbose-)
          l_paste=tmpim
      } else {
          l_paste=l_image
      } 
  }
  if (l_fl_bias) {
      imarith(l_paste//"["//str(n_i)//"]","-",l_bias[n_i],l_paste//"["//str(n_i)//",overwrite+]"
  }

}

}
scanfile=""
#print(s_bias)

delete(tmpin,verify-, >>& "dev$null")

if(l_fl_paste) {

 print("GDISPLAY: Pasting images")
 gmosaic(l_paste,outimages=tmpout,fl_paste+,gap=l_gap,fl_fixpix-,
  verbose=l_verbose,fl_clean-,sci_ext=l_sci_ext,logfile="dev$null")
 display(tmpout//"[1]",l_frame,erase=yes,xcenter=0.5,ycenter=0.5,
    xmag=1,ymag=1,z1=l_z1,z2=l_z2,zrange=fl_zrange,zscale=fl_zscale,
    fill=no)
 print("GDISPLAY: Starting imexamine - quit with q")
 imexamine()
 if (l_output == "") {
   imdelete(tmpout,verify-, >>& "dev$null")
 } else {
   imrename(tmpout,l_output,verbose+)
 }
 if(l_fl_bias && l_fl_sec)
   imdelete(tmpim,verify-)

}

end
