# Copyright(c) 2002-2004 Association of Universities for Research in Astronomy, Inc.

procedure gfreduce (inimages)

# Reduce GMOS IFU images
# 
# Version   Sept 20, 2002 BM v1.4 release
#           Dec 4, 2003   BM generalize for GMOS-S IFU
#           Feb 29, 2004  BM added thresh param and gnsskysub option
#           Mar 19, 2004  BM added fl_novlap and perovlap params
#           Mar 24, 2004  BM fix bug in call to gscalibrate
#           Mar 25, 2004  BM Now exit gracefully if problem with input images
#           Apr 07, 2004  BM Fix print of INDEF, auto-select N&S MDF and
#                            set l_fl_skysub if N&S mask used

string inimages     {"",prompt="Input images"}
string outimages    {"",prompt="Output images"}
string outpref      {"default",prompt="Prefix for output images"}
string slits        {"both",enum="red|blue|both",prompt="Slit(s) used"}
bool   fl_nodshuffle {no,prompt="Nod & Shuffle output slit mask used"}
bool   fl_inter     {yes,prompt="Interactive use?"}
bool   fl_vardq     {no,prompt="Use variance and DQ planes?"}
bool   fl_addmdf    {yes,prompt="Add mask definition file (run GPREPARE)"}
bool   fl_over      {no,prompt="Apply overscan correction?"}
bool   fl_trim      {yes,prompt="Trim off the overscan section?"}
bool   fl_bias      {yes,prompt="Apply zero level (bias) correction?"}
bool   fl_gscrrej   {yes,prompt="Clean the cosmic rays?"}
bool   fl_gnsskysub {no,prompt="Run GNSSKYSUB?"}
bool   fl_extract   {yes,prompt="Extract spectra?"}
bool   fl_gsappwave {yes,prompt="Run gsappwave?"}
bool   fl_wavtran   {yes,prompt="Transform cube to apply wavelength solution?"}
bool   fl_skysub    {yes,prompt="Subtract sky?"}
bool   fl_fluxcal   {yes,prompt="Apply flux calibration?"}
string rawpath      {"",prompt="Path for input raw images"}
string key_mdf      {"",prompt="Header keyword for MDF"}
string mdffile      {"default",prompt="MDF file to use if keyword not found"}
string mdfdir       {"gmos$data/",prompt="MDF database directory"}
string key_biassec  {"BIASSEC",prompt="Header keyword for bias section"}
string key_datasec  {"DATASEC",prompt="Header keyword for data section"}
string bpmfile      {"gmos$data/chipgaps.dat",prompt="Bad pixel mask for column interpolation"}
string bias         {"",prompt="Bias image name"}
string reference    {"",prompt="Reference file"}
string response     {"",prompt="Fiber response file"}
string wavtraname   {"",prompt="Names of wavelength calibrations"}
string sfunction    {"",prompt="Sensitivity function from GSSTANDARD"}
string extinction   {"",prompt="Extinction file"}
bool   fl_fixnc     {no,prompt="Auto-correct for nod count mismatch?"}
bool   fl_novlap    {yes,prompt="Avoid spectral overlap?"}
real   perovlap     {10.,prompt="Percentage by which to shrink overlapping spectra"}
int    order        {1,min=1,prompt="Order of overscan fitting function"}
int    line         {INDEF,prompt="Line for finding peaks"}
int    nsum         {10,prompt="Number of columns to use for finding apertures"}
bool   trace        {yes,prompt="Trace spectra"}
bool   recenter     {yes,prompt="Recenter apertures?"}
real   thresh       {200.,prompt="Detection threshold for profile centering"}
string function     {"chebyshev",enum="chebyshev|spline1|spline3|legendre",prompt="Function for trace"}
int    t_order      {5,min=1,prompt="Order of trace fit"}
int    t_nsum       {10,min=1,prompt="Number of dispersion lines to be summed during trace"}
string weights      {"variance",enum="variance|none",prompt="Weighting during extraction"}
string gratingdb    {"gmos$data/GMOSgratings.dat",prompt="Gratings database file"}
string filterdb     {"gmos$data/GMOSfilters.dat",prompt="Filters database file"}
real   xoffset      {INDEF,prompt="X offset in wavelength [nm]"}
string expr         {"default",prompt="Expression for selection of sky spectra"}
string observatory  {"default",prompt="Observatory name"}
string  sci_ext     {"SCI",prompt="Name of science extension"}
string  var_ext     {"VAR",prompt="Name of variance extension"}
string  dq_ext      {"DQ",prompt="Name of data quality extension"}
string logfile      {"",prompt="Logfile name"}
bool   verbose      {yes,prompt="Verbose?"}
int    status       {0,prompt="Exit status (0=good)"}
struct *scanfile    {"",prompt="Internal use only"}

begin

    # Local variables for input parameters
    string l_inimages,l_outimages,l_prefix,l_logfile, l_rawpath
    string l_key_mdf,l_mdffile,l_mdfdir,l_slits
    string l_bias,l_key_biassec,l_key_datasec
    string l_response,l_gratingdb,l_filterdb
    string l_reference,l_bpmfile,l_function,l_weights
    string l_wavtraname,l_expr
    string l_sci_ext, l_var_ext, l_dq_ext
    string l_obs, l_sfunc, l_extinct
    real l_xoffset,l_thresh,l_perovlap
    int  l_order,l_t_order,l_line,l_nsum,l_t_nsum,inst
    bool l_verbose,l_fl_inter,l_fl_vardq
    bool l_fl_addmdf,l_fl_over, l_fl_bias, l_fl_trim
    bool l_fl_extract, l_recenter,l_trace,l_fl_gsappwave
    bool l_fl_wavtran, l_fl_skysub, l_fl_fluxcal, l_fl_ext, l_fl_gscrrej
    bool l_fl_gnsskysub,l_fl_fixnc,l_fl_novlap,l_fl_nodshuffle

    # Other local variables
    string inlist,infile[500],outlist,temp1,temp2
    string img,outimg,suf,root
    int i,j,nbad,nin,nout,nx,nslits
    bool useprefix,usenodsh
    struct sdate

    # Initialize exit status
    status=0

    # cache some parameter files
    cache("imgets","hedit","gimverify","fparse","gemlogname")

    # Initialize local variables
    l_inimages=inimages ; l_outimages=outimages ; l_prefix=outpref
    l_logfile=logfile ; l_verbose=verbose ; l_fl_inter=fl_inter
    l_rawpath=rawpath ; l_key_biassec=key_biassec 
    l_fl_addmdf=fl_addmdf ; l_key_mdf=key_mdf ; l_slits=slits
    l_mdffile=mdffile ; l_mdfdir=mdfdir
    l_fl_over=fl_over ; l_fl_bias=fl_bias ; l_order=order ; l_bias=bias
    l_fl_trim=fl_trim ; l_key_datasec=key_datasec
    l_response=response ; l_reference=reference ; l_bpmfile=bpmfile
    l_sci_ext=sci_ext ; l_var_ext=var_ext ; l_dq_ext=dq_ext 
    l_line=line
    l_fl_vardq=fl_vardq
    l_gratingdb=gratingdb
    l_filterdb=filterdb
    l_trace=trace
    l_nsum=nsum
    l_t_nsum=t_nsum
    l_t_order=t_order
    l_function=function
    l_weights=weights
    l_fl_gsappwave=fl_gsappwave
    l_recenter=recenter ; l_thresh=thresh
    l_fl_extract=fl_extract
    l_fl_wavtran=fl_wavtran ; l_wavtraname=wavtraname
    l_fl_skysub=fl_skysub ; l_expr=expr
    l_fl_fluxcal=fl_fluxcal ; l_sfunc=sfunction ; l_extinct=extinction
    l_obs=observatory
    l_fl_gscrrej=fl_gscrrej
    l_xoffset=xoffset
    l_fl_gnsskysub=fl_gnsskysub ; l_fl_fixnc=fl_fixnc
    l_fl_novlap=fl_novlap ; l_perovlap=perovlap
    l_fl_nodshuffle=fl_nodshuffle

    # Test the logfile:
    gemlogname(logpar=l_logfile,package="gmos")
    if (gemlogname.status != 0) {
        goto error
    }
    l_logfile=gemlogname.logname

    printlog("-------------------------------------------------------------------------------",l_logfile,l_verbose)
    date | scan(sdate)
    printlog("GFREDUCE -- "//sdate,l_logfile,l_verbose)
    printlog(" ",l_logfile,l_verbose)

    #The usual kind of checks to make sure everything we need is specified. 

    nbad=0

    if(l_inimages=="" || l_inimages==" ") {
      printlog("ERROR - GFREDUCE: Input spectra is an empty string",l_logfile,yes)
      nbad=nbad+1 
    }

    if(substr(l_inimages,1,1)=="@") {
       inlist=substr(l_inimages,2,strlen(l_inimages))
       if(!access(inlist)) {
          printlog("ERROR - GFREDUCE: Input list "//inlist//" not found",l_logfile,verbose+)
          nbad=nbad+1
       }
    }

    useprefix=yes
    if (l_outimages!="" && l_outimages!=" ") {
        useprefix=no
    } else if (l_prefix=="default") {
        printlog("Default prefixes will be used.",l_logfile,l_verbose)
    } else if (l_prefix == "" || l_prefix == " ") {
        printlog("ERROR - GFREDUCE: Neither outimages or outpref given.",l_logfile,yes)
        nbad=nbad+1
    }

    if(substr(l_outimages,1,1)=="@") {
       outlist=substr(l_outimages,2,strlen(l_outimages))
       if(!access(outlist)) {
          printlog("ERROR - GFREDUCE: Output list "//outlist//" not found",l_logfile,yes)
          nbad=nbad+1
       }
    }

#    if (l_fl_gnsskysub && l_fl_skysub) {  
#         printlog("ERROR - GFREDUCE: cannot use both fl_gnsskysub and fl_skysub options.",l_logfile,yes)
#          nbad=nbad+1
#    }

    temp1=mktemp("tmpin")
    temp2=mktemp("tmpout") 
    files(l_inimages,sort-, > temp1)
    count(temp1) | scan(nin)
    if(!useprefix) { 
        files(l_outimages,sort-, > temp2) 
        count(temp2) | scan(nout)
        if (nin != nout) {
            printlog("ERROR - GFREDUCE: Different number of input and output spectra",l_logfile,yes)
            nbad=nbad+1
        } 
    } else if (l_prefix != "default") {
        files(l_prefix//"//@"//temp1,sort-, > temp2)
    }

    i=0
    scanfile=temp1
    while(fscan(scanfile,img) !=EOF) { 
        gimverify(l_rawpath//img)
        if (gimverify.status>0) {
            printlog("ERROR - GFREDUCE: "//img//" not found or not a MEF",l_logfile,yes)
            nbad=nbad+1
        } else {
            i=i+1
            # name w/o suffix
            fparse(gimverify.outname,verbose-)
            infile[i]=fparse.root//".fits"
        }
    #end of while loop over images. 
    }
    scanfile=""

    # Check existence of output images
    if (access(temp2)) {
      scanfile=temp2
      while (fscan(scanfile,img) != EOF) {
        if (imaccess(img//"[0]")) {
            printlog("ERROR - GFREDUCE: "//img//" already exists",l_logfile,yes)
            nbad=nbad+1
        }
      }
      scanfile=""
    }

    #If anything was wrong then exit. 
    if (nbad > 0) { 
        printlog("ERROR - GFREDUCE: "//nbad//" error(s) found with input parameters. Exiting.",l_logfile,yes) 
        goto error
    } 

    # Determine which instrument
    imgets(l_rawpath//infile[1]//"[0]","INSTRUME", >>& "dev$null")
    if (imgets.value == "0") {
        printlog ("ERROR - GFREDUCE: Instrument keyword not found.",l_logfile, verbose+)
        goto error
    }
    inst=1  # Default is GMOS-N, support for old data
    if (imgets.value == "GMOS-S") {
        inst=2
    }

    # Determine if using nod&shuffle
    usenodsh=no
    imgets(l_rawpath//infile[1]//"[0]","NODPIX", >>& "dev$null")
    if (imgets.value != "0") {
       usenodsh=yes
    } 

    if (usenodsh || l_fl_nodshuffle) {
        if (l_fl_skysub) {
            printlog("WARNING - GFREDUCE: Nod&Shuffle mode on.  Resetting \
	        fl_skysub to 'no'", l_logfile, verbose+)
        }
        l_fl_skysub=no
    } else {
        if (l_fl_gnsskysub) {
	    printlog("WARNING - GFREDUCE: Nod&Shuffle mode off.  Resetting \
	         fl_gnsskysub to 'no'", l_logfile, verbose+)
	}
        l_fl_gnsskysub=no
    }

    # Default MDF files
    if (l_mdffile=="default") {
        if (inst==1) {
        if (l_slits=="red") {
            l_mdffile="gnifu_slitr_mdf.fits"
        } else if (l_slits=="blue") {
            l_mdffile="gnifu_slitb_mdf.fits"
        } else {
            l_mdffile="gnifu_slits_mdf.fits"
        }
        } else {
           if (usenodsh || l_fl_nodshuffle) {
               if (l_slits=="red") {
                   l_mdffile="gsifu_ns_slitr_mdf.fits"
               } else if (l_slits=="blue") {
                   l_mdffile="gsifu_ns_slitb_mdf.fits"
               } else {
                   l_mdffile="gsifu_ns_slits_mdf.fits"
               }
           } else {
               if (l_slits=="red") {
                   l_mdffile="gsifu_slitr_mdf.fits"
               } else if (l_slits=="blue") {
                   l_mdffile="gsifu_slitb_mdf.fits"
               } else {
                   l_mdffile="gsifu_slits_mdf.fits"
               }
           }
        }

    }

    # Default sky selection region
    if (l_expr=="default") {
        if (inst==1) {
            l_expr="XINST > 10."
        } else {
            l_expr="XINST < 10."
        }
    }

    # Default observatory
    if (l_obs=="default") {
        if (inst==1) {
            l_obs="Gemini-North"
        } else {
            l_obs="Gemini-South"
        }
    }

    # Number of slits
    nslits=1
    if (l_slits=="both") {
      nslits=2
    }

    #If we are here then everything should be OK. 
    #Write all the relevant info to the logfile:
    #
    printlog("",l_logfile,l_verbose)
    printlog("inimages   = "//l_inimages,l_logfile,l_verbose) 
    printlog("outimages  = "//l_outimages,l_logfile,l_verbose) 
    printlog("outpref    = "//l_prefix,l_logfile,l_verbose)
    printlog("slits      = "//l_slits,l_logfile,l_verbose)
    if (l_fl_nodshuffle || usenodsh) {
      printlog("nod&shuffle= yes",l_logfile,l_verbose)
    } else {
      printlog("nod&shuffle= no",l_logfile,l_verbose)
    }
    printlog("mdffile    = "//l_mdffile,l_logfile,l_verbose)
    printlog("mdfdir     = "//l_mdfdir,l_logfile,l_verbose)
    printlog("bias       = "//l_bias,l_logfile,l_verbose)
    printlog("reference  = "//l_reference,l_logfile,l_verbose)
    printlog("response   = "//l_response,l_logfile,l_verbose)
    printlog("wavtraname = "//l_wavtraname,l_logfile,l_verbose)
    printlog("sfunction  = "//l_sfunc,l_logfile,l_verbose)
    printlog("extinction = "//l_extinct,l_logfile,l_verbose)
    printlog("expr       = "//l_expr,l_logfile,l_verbose)
    printlog("gratingdb  = "//l_gratingdb,l_logfile,l_verbose)
    printlog("filterdb   = "//l_filterdb,l_logfile,l_verbose)
    if (l_xoffset != INDEF) {
      printlog("xoffset    = "//str(l_xoffset),l_logfile,verbose=l_verbose)
    } else {
      printlog("xoffset    = INDEF",l_logfile,verbose=l_verbose)
    }

    printlog(" ",l_logfile,l_verbose)

    # gprepare
    if (l_fl_addmdf) {
        gprepare("@"//temp1,rawpath=l_rawpath,outimages="",outpre="g",
            fl_addmdf=l_fl_addmdf,key_mdf=l_key_mdf, mdffile=l_mdffile,
            mdfdir=l_mdfdir,verbose=l_verbose,logfile=l_logfile,
            fl_vardq=l_fl_vardq,sci_ext=l_sci_ext,var_ext=l_var_ext,
            dq_ext=l_dq_ext)
        if (gprepare.status != 0)
            goto error
        for (i=1; i<=nin; i+=1) {
            infile[i]="g"//infile[i]
        }
    }

    # gireduce
    if (l_fl_over || l_fl_trim || l_fl_bias) {
        delete(temp1,verify-, >>& "dev$null")
        for (i=1; i<=nin; i+=1) {
            print(infile[i], >> temp1)
        }
        gireduce("@"//temp1,outimage="",outpre="r",fl_over=l_fl_over,
            fl_trim=fl_trim,fl_bias=fl_bias,fl_dark-,fl_flat-,order=l_order,
            key_biassec=l_key_biassec,key_datasec=l_key_datasec,bias=l_bias,
            verbose=l_verbose,fl_vardq=l_fl_vardq,fl_inter=l_fl_inter,
            logfile=l_logfile,sci_ext=l_sci_ext,var_ext=l_var_ext,
            dq_ext=l_dq_ext)
        if (gireduce.status != 0)
            goto error
        for (i=1; i<=nin; i+=1) {
            infile[i]="r"//infile[i]
        }
    }

    # gscrrej
    if (l_fl_gscrrej) {
        delete(temp1,verify-, >>& "dev$null")
        for (i=1; i<=nin; i+=1) {
            gscrrej(infile[i],"x"//infile[i],verbose=l_verbose,
                fl_inter=l_fl_inter,logfile=l_logfile)
            if (gscrrej.status != 0)
                goto error
            infile[i]="x"//infile[i]
        }
    }

    # Extract spectra
    if (l_fl_extract) {
        delete(temp1,verify-, >>& "dev$null")
        for (i=1; i<=nin; i+=1) {
            gfextract(infile[i],outimage="",outpref="e",reference=l_reference,
                response=l_response,bpmfile=l_bpmfile,line=l_line,nsum=l_nsum,
                trace=l_trace,recenter=l_recenter,function=l_function,
                order=l_t_order,t_nsum=l_nsum,weights=l_weights,thresh=l_thresh,
                gratingdb=l_gratingdb,filterdb=l_filterdb,xoffset=l_xoffset,
                fl_gsappwave=l_fl_gsappwave,fl_vardq=l_fl_vardq,
                fl_inter=l_fl_inter,verbose=l_verbose,logfile=l_logfile,
                sci_ext=l_sci_ext,var_ext=l_var_ext,dq_ext=l_dq_ext,
                fl_gnsskysub=l_fl_gnsskysub,fl_fixnc=l_fl_fixnc,
                fl_novlap=l_fl_novlap,perovlap=l_perovlap)
            if (gfextract.status != 0)
                goto error
            infile[i]="e"//infile[i]
        }
    }

    # Wavelength calibration
    if (l_fl_wavtran) {
        delete(temp1,verify-, >>& "dev$null")
        for (i=1; i<=nin; i+=1) {
            print(infile[i], >> temp1)
        }
        gftransform("@"//temp1,outimage="",outpre="t",wavtraname=l_wavtraname,
            database="database",fl_vardq=l_fl_vardq,fl_flux+,logfile=l_logfile,
            sci_ext=l_sci_ext,var_ext=l_var_ext,dq_ext=l_dq_ext,
            verbose=l_verbose)
        if (gftransform.status != 0)
            goto error
        for (i=1; i<=nin; i+=1) {
            infile[i]="t"//infile[i]
        }
    }

    # Sky subtraction
    if (l_fl_skysub) {
        delete(temp1,verify-, >>& "dev$null")
        for (i=1; i<=nin; i+=1) {
            print(infile[i], >> temp1)
        }
        gfskysub("@"//temp1,outimages="",outpre="s",expr=l_expr,
            logfile=l_logfile,verbose=l_verbose,fl_inter=l_fl_inter,
            sci_ext=l_sci_ext,var_ext=l_var_ext,dq_ext=l_dq_ext)
        if (gfskysub.status != 0)
            goto error
        for (i=1; i<=nin; i+=1) {
            infile[i]="s"//infile[i]
        }
    }

    # Flux calibration
    if (l_fl_fluxcal) {
        delete(temp1,verify-, >>& "dev$null")
        for (i=1; i<=nin; i+=1) {
            print(infile[i], >> temp1)
        }
        l_fl_ext=no
        if (access(l_extinct))
            l_fl_ext=yes
        gscalibrate("@"//temp1,output="",outpref="c",sfunction=l_sfunc,
            fl_vardq=l_fl_vardq,fl_flux=yes,fl_ext=l_fl_ext,
            extinction=l_extinct,logfile=l_logfile,verbose=l_verbose,
            sci_ext=l_sci_ext,var_ext=l_var_ext,
            dq_ext=l_dq_ext,observatory=l_obs)
        if (gscalibrate.status != 0)
            goto error
        for (i=1; i<=nin; i+=1) {
            infile[i]="c"//infile[i]
        }
    }

    # Rename output files
    if (access(temp2)) {
        printlog("",l_logfile,l_verbose)
        printlog("Renaming output files",l_logfile,l_verbose)
        i=0
        scanfile=temp2
        while(fscan(scanfile,img) !=EOF) { 
            i=i+1
            imrename (infile[i],img,verbose=l_verbose)
 
            # also need to rename aperture and other database files
            if (l_fl_extract) {
                fparse(img,verbose-)
                outimg=fparse.root
                j=stridx("e",infile[i])
                fparse(infile[i],verbose-)
                root=fparse.root
                root=substr(root,j,(strlen(root)))
                for (i=1; i<=nslits; i+=1) {
                    sed("-e","s/"//root//"/"//outimg//"/g","database/ap"//root//"_"//i, > "database/ap"//outimg//"_"//i)
                    delete("database/ap"//root//"_"//i,verify-)
                }
            }
        }
        scanfile=""
    }

    goto clean

error:
    status=1

clean:
    delete(temp1//","//temp2,verify-, >& "dev$null") 
    printlog("",l_logfile,l_verbose)
    scanfile=""
    if(status==0)
        printlog("GFREDUCE exit status: good",l_logfile,l_verbose)
    else
        printlog("GFREDUCE exit status: error",l_logfile,l_verbose)
    printlog("-------------------------------------------------------------------------------",
        l_logfile,l_verbose)

end
