# Copyright(c) 2003-2004 Association of Universities for Research in Astronomy, Inc.

procedure gifringe(input,output)

# Create fringe frames
# 
# Version Jul 24, 2003  BM  created
#         Jul 29, 2003  BM  add columns variable, subtract mean before fit
#         Aug 27, 2003  KL  IRAF2.12 - new parameters
#                             imstat: nclip, lsigma, usigma, cache
#         Dec 03, 2003  PLG took out imsurfit. The code now gemcombines the images after
#                           a zero point correction given by the keyword OFFINT (mean of CCD2).

string input        {"",prompt="Input MEF images"}                             # OLDP-1-input-primary-combine
string output       {"",prompt="Output MEF image"}                             # OLDP-1-output
string title        {"",prompt="Title for output SCI plane"}                   # OLDP-3
string logfile      {"",prompt="Log file"}                                     # OLDP-1
string sctype       {"mean",enum="mean|midpt",prompt="Operation to determine the zero point"} # OLDP-3
#gemcombine parameters
string combine      {"median",enum="average|median",prompt="Combination operation"} # OLDP-3
string reject       {"avsigclip",prompt="Rejection algorithm"}                 # OLDP-3
string scale        {"none",prompt="Image scaling"}                            # OLDP-3
string zero         {"mean",prompt="Image zeropoint offset"}                   # OLDP-3
string weight       {"none",prompt="Image weights"}                            # OLDP-3
string statsec      {"default",prompt="Statistics section"}                    # OLDP-3
string expname      {"EXPTIME",prompt="Exposure time header keyword"}          # OLDP-3
real   lthreshold   {INDEF,prompt="Lower threshold"}                           # OLDP-3
real   hthreshold   {INDEF,prompt="Upper threshold"}                           # OLDP-3
int    nlow         {1,min=0,prompt="minmax: Number of low pixels to reject"}  # OLDP-3
int    nhigh        {1,min=0,prompt="minmax: Number of high pixels to reject"} # OLDP-3
int    nkeep        {0,min=0,prompt="Minimum to keep or maximum to reject"}    # OLDP-3
bool   mclip        {yes,prompt="Use median in sigma clipping algorithms?"}    # OLDP-3
real   lsigma       {3.,prompt="Lower sigma clipping factor"}                  # OLDP-3
real   hsigma       {3.,prompt="Upper sigma clipping factor"}                  # OLDP-3
string key_ron      {"RDNOISE",prompt="Keyword for readout noise in e-"}       # OLDP-3
string key_gain     {"GAIN",prompt="Keyword for gain in electrons/ADU"}        # OLDP-3
real   ron          {0.0,min=0.,prompt="Readout noise rms in electrons"}       # OLDP-3
real   gain         {1.0,min=0.00001,prompt="Gain in e-/ADU"}                  # OLDP-3
string snoise       {"0.0",prompt="ccdclip: Sensitivity noise (electrons)"}    # OLDP-3
real   sigscale     {0.1,prompt="Tolerance for sigma clipping scaling correction"} # OLDP-3
real   pclip        {-0.5,prompt="pclip: Percentile clipping parameter"}       # OLDP-3
real   grow         {0.0,prompt="Radius (pixels) for neighbor rejection"}      # OLDP-3
string bpm          {"",prompt="Name of bad pixel mask file or image."}        # OLDP-2-input
string sci_ext      {"SCI",prompt="Name(s) or number(s) of science extension"} # OLDP-3
string var_ext      {"VAR",prompt="Name(s) or number(s) of variance extension"} # OLDP-3
string dq_ext       {"DQ",prompt="Name(s) or number(s) of data quality extension"} # OLDP-3
bool   fl_vardq     {no,prompt="Make variance and data quality planes?"}       # OLDP-2
bool   fl_dqprop    {no,prompt="Propagate all DQ values?"}                     # OLDP-2
bool   verbose      {yes,prompt="Verbose output?"}                             # OLDP-4
int    status       {0,prompt="Exit status (0=good)"}                          # OLDP-4
struct  *flist      {"",prompt="Internal use only"}                            # OLDP-4

begin

    string l_input,l_output,l_statsec, l_expname, l_bpm, l_logfile
    string l_combine, l_reject, l_scale, l_zero, l_weight, l_offsets
    string l_key_ron, l_key_gain, l_snoise
    string l_sgain, l_sron
    real   l_ron, l_gain, l_lthreshold, l_hthreshold, l_lsigma, l_hsigma
    real   l_grow, l_sigscale, l_pclip
    int    l_nlow, l_nhigh, l_nkeep, nfiles
    int    l_xorder,l_yorder,l_xmed,l_ymed
    bool   l_verbose, s_verbose, l_mclip, l_fl_dqprop, l_fl_vardq, badformat
    string l_sci_ext, l_var_ext, l_dq_ext, l_title, tsci,tvar,tdq
    string sextlist,vextlist,dextlist,extlist,sci[600],var[600],dq[600]
    string filelist,img,suf
    string plimg,sect,infile[500],mdf
    string l_function, s_empty,l_columns,tmpcolumns,sec_cols[6],ccdsum[6]
    string fccd[6]
    string exphu,phuimg
    char   scaletype, l_sctype
    real   meanval, firstval

    real   mean,epadu,rdnoise,gaineff,roneff
    int    ii,jj,kk,l,n,idum,nbad, len, nsci,nvar,ndq,n_sc
    int    nimg,nextnd,secpos1,secpos2,nsciext, dum
    bool   useextver,usemdf, l_cleanbpm
    struct sdate

    # Query parameters
    l_input=input ; l_output=output
    l_title=title
    l_logfile=logfile
    l_combine=combine ; l_reject=reject
    l_scale=scale ; l_zero=zero ; l_weight=weight
    l_statsec=statsec ; l_expname=expname
    l_lthreshold=lthreshold ; l_hthreshold=hthreshold
    l_nlow=nlow ; l_nhigh=nhigh ; l_nkeep=nkeep
    l_lsigma=lsigma ; l_hsigma=hsigma
    l_verbose=verbose ; l_mclip=mclip
    l_key_gain=key_gain ; l_key_ron=key_ron
    l_gain=gain ; l_ron=ron ;l_snoise=snoise 
    l_sgain=l_key_gain ;    l_sron=l_key_ron
    l_sigscale=sigscale ; l_pclip=pclip ; l_grow=grow
    l_sci_ext = sci_ext ; l_dq_ext = dq_ext ; l_var_ext=var_ext
    l_bpm=bpm ; l_fl_dqprop=fl_dqprop ; l_fl_vardq=fl_vardq
    l_statsec=statsec

    status=0
    useextver=no
    usemdf=no

    # Keep imgets parameters from changing by outside world
    cache("imgets","hedit","keypar")
    s_verbose = hedit.show
    hedit.show=l_verbose

    # Define the name of the logfile
    s_empty=""; print(l_logfile) | scan(s_empty); l_logfile=s_empty
    if (l_logfile == "") {
        l_logfile = gmos.logfile
        if (l_logfile == "") {
	    l_logfile = "gmos.log"
	    printlog("WARNING - GIFRINGE: both gifringe.logfile and gmos.logfile are empty.",
	        l_logfile,l_verbose)
	    printlog("                Using default file gmos.log.",
	        l_logfile,l_verbose)
        }
    }

    date | scan(sdate)
    printlog("--------------------------------------------------------------------------------",
        l_logfile,verbose=l_verbose)
    printlog("GIFRINGE -- "//sdate,l_logfile,verbose=l_verbose)
    printlog(" ",l_logfile,verbose=l_verbose)
    printlog("input = "//l_input,l_logfile,verbose=l_verbose)
    printlog("output = "//l_output,l_logfile,verbose=l_verbose)
    printlog("title = "//l_title,l_logfile,verbose=l_verbose)
    printlog("bpm = "//l_bpm,l_logfile,verbose=l_verbose)
    printlog("sci_ext = "//l_sci_ext,l_logfile,verbose=l_verbose)
    printlog("var_ext = "//l_var_ext,l_logfile,verbose=l_verbose)
    printlog("dq_ext = "//l_dq_ext,l_logfile,verbose=l_verbose)
    printlog(" ",l_logfile,verbose=l_verbose)

    # Verify that the extension names are not empty, otherwise exit gracefully
    if (l_sci_ext=="" || l_sci_ext==" ") {
	printlog("ERROR - GIFRINGE: Extension name sci_ext is missing",
	    l_logfile,yes)
	goto error
    }
    if (l_dq_ext=="" || l_dq_ext ==" ") {
	printlog("ERROR - GIFRINGE: Extension name dq_ext is missing",
	    l_logfile,yes)
	goto error
    } 

    # Define temporary files
    filelist=mktemp("tmpfilelist")
    sextlist=mktemp("tmpsextlist")
    vextlist=mktemp("tmpvextlist")
    dextlist=mktemp("tmpdextlist")
    extlist=mktemp("tmpextlist")

    #check that there are input files
    if (l_input == "" || l_input == " "){
        printlog("ERROR - GIFRINGE: input files not specified",l_logfile,yes)
        goto error
    }

    # check existence of list file
    if (substr(l_input,1,1) == "@") {
	len=strlen(l_input)
	if (!access(substr(l_input,2,len))) {
            printlog("ERROR - GIFRINGE: "//substr(l_input,2,len)//" does not exist.",
	        l_logfile,yes)
            goto error
	}
    }

    # Check for an image section
    sect=""
    secpos1 = stridx("[",l_input)
    secpos2 = stridx("]",l_input)
    if (secpos1>0 && secpos2>0) {
	sect = substr(l_input,secpos1,secpos2)
	l_input = substr(l_input,1,secpos1-1)
    } else if (secpos1>0 || secpos2>0) {
	printlog("ERROR - GIFRINGE: mismatched brackets in "//l_input,
	    l_logfile,yes)
	goto error
    }
    files(l_input,sort-, > filelist)

    #check that an output file is given
    if (l_output == "" || l_output == " "){
        printlog("ERROR - GIFRINGE: output files not specified",l_logfile,yes)
        goto error
    }

    # Check that the output file does not already exist. If so, exit.
    if (imaccess(l_output)) {  
        printlog("ERROR - GIFRINGE: Output file "//l_output//" already exists.",
	    l_logfile,yes)
        goto error
    }

    #put suffix on output image
    len=strlen(l_output)
    suf=substr(l_output,(len-4),len)
    if (suf != ".fits")
	l_output=l_output//".fits"

    # Make list of sci/var/dq extensions
    files(l_sci_ext,sort-, > sextlist)
    count(sextlist) | scan(nsci)
    files(l_var_ext,sort-, > vextlist)
    count(vextlist) | scan(nvar)
    files(l_dq_ext,sort-, > dextlist)
    count(dextlist) | scan(ndq)
    if ((nsci != nvar) || (nsci != ndq)) {
	printlog("ERROR - GIFRINGE: different numbers of SCI, VAR, or DQ extensions.",
	    l_logfile,yes)
	goto error
    }

    if (nsci > 600) {
	printlog("ERROR - GIFRINGE: the number of science extensions is larger than sci, var, and dq array lengths",
	    l_logfile,yes)
	goto error
    }

    join (sextlist//","//vextlist//","//dextlist,"",output=extlist)
    delete (sextlist//","//vextlist//","//dextlist,verify-, >& "dev$null")
    flist=extlist
    ii=0
    while (fscan(flist, tsci, tvar, tdq) != EOF) {
	ii=ii+1
	sci[ii]=tsci
	var[ii]=tvar
	dq[ii]=tdq
    }
    flist=""
    delete(extlist,verify-, >& "dev$null")

    # Check that all images in the list exist and are MEF
    nbad=0
    kk=0
    flist=filelist
    while (fscan(flist, img) != EOF) {
	secpos1 = stridx("[",img)
	secpos2 = stridx("]",img)
	if (secpos1>0 && secpos2>0) {
            sect = substr(img,secpos1,secpos2)
            img = substr(img,1,secpos1-1)
	} else if (secpos1>0 || secpos2>0) {
            printlog("ERROR - GIFRINGE: mismatched brackets in "//img,
	        l_logfile,yes)
            goto error
	}
	gimverify(img)
	if (gimverify.status>0)
            nbad=nbad+1
	else {
            kk=kk+1
            # name w/o suffix
            infile[kk]=gimverify.outname//".fits"
            if (kk==1) {
        	imgets(infile[kk]//"[0]","NSCIEXT", >>& "dev$null")
        	nsciext=int(imgets.value)
        	if (nsciext > 600) {
                    printlog("ERROR - GIFRINGE: the number of science extensions is larger than sci, var, and dq array lengths",
		        l_logfile,yes)
                    goto error
        	}

        	# check if EXTVER used
        	if (nsci==1 && nsciext > 0) {
                    if (imaccess(infile[kk]//"["//sci[kk]//","//nsciext//"]")) {
                	tsci=sci[1]
                	tvar=var[1]
                	tdq=dq[1]
                	for (ii=1; ii<=nsciext; ii+=1) {
                            sci[ii]=tsci//","//ii
                            var[ii]=tvar//","//ii
                            dq[ii]=tdq//","//ii
                	}
                	nsci=nsciext
                	useextver=yes
                    }
        	}
            }
	} 
    }
    
    #############             Checking the binning
    for (ii=1; ii<=nsci; ii+=1) {
	imgets(infile//"["//l_sci_ext//","//ii//"]","CCDSUM", >& "dev$null")
	ccdsum[ii]=imgets.value
    } 
    flist=""
    nimg=kk
    # Exit if problems found
    if (nbad > 0) {
	printlog("ERROR - GIFRINGE: "//nbad//" image(s) either do not exist or are not MEF files.",
	    l_logfile,yes)
	goto error
    }
    
    ###########                Define the statsec
    if (l_statsec=="default") {
	if (nsci == 6) l_statsec="["//l_sci_ext//",3][100:900,100:4500]"
	if (nsci == 3) l_statsec="["//l_sci_ext//",2][100:1900,100:4500]"
	if (nsci == 1) l_statsec="["//l_sci_ext//",1][100:1900,100:4500]"
	if (ccdsum[1]=="2 2" && ccdsum[2]=="2 2" && ccdsum[3]=="2 2") {
	    if(nsci == 6) l_statsec = "["//l_sci_ext//",3][50:450,50:2250]"
	    if(nsci == 3) l_statsec = "["//l_sci_ext//",2][50:850,50:2250]"
	    if(nsci == 1) l_statsec = "["//l_sci_ext//",1][50:850,50:2250]"
	} else if (ccdsum[1]=="4 4" && ccdsum[4]=="4 4" && ccdsum[3]=="4 4") { 
	    if(nsci == 6) l_statsec = "["//l_sci_ext//",3][25:225,25:1125]"
	    if(nsci == 3) l_statsec = "["//l_sci_ext//",2][25:425,25:1125]"
	    if(nsci == 1) l_statsec = "["//l_sci_ext//",1][25:425,25:1125]"
	}
    } else if (substr(l_statsec,2,4) != l_sci_ext) {
	 printlog("ERROR - GIFRINGE: illegal parameter STATSEC ("//l_statsec//") format is e.g [SCI,2][100:1900,100:4500]",
	     l_logfile,yes)
    }
    l_sctype=sctype
    l_zero="!OFFINT"
    
    #############             
    kk=0
    flist=filelist
    
    ##################          Computing the mean of CCD 2 for each image
    while (fscan(flist, img) != EOF) {
	imstatistics(img//l_statsec,field=l_sctype,lower=INDEF,upper=INDEF,
	    nclip=0,lsigma=INDEF,usigma=INDEF,binwidth=0.1,format-,cache-) | \
	    scan(meanval)
	for(ii=1; ii<=nsci; ii+=1)                  
	    hedit(img//"["//l_sci_ext//","//ii//"]","OFFINT",-meanval,add+,
	        addonly-,delete-,verify-,show-,update+)
    }
    ################             Gemcombine the zero-point corrected images
    # Add object removal/masking in future, wait for 2.12 port
    # Combine
    gemcombine("@"//filelist,l_output,title=l_title,logfile=l_logfile,
	combine=l_combine,reject=l_reject,offsets="none",scale=l_scale,
	zero=l_zero,weight=l_weight,statsec=l_statsec,expname=l_expname,
	lthreshold=l_lthreshold,hthreshold=l_hthreshold,nlow=l_nlow,
	nhigh=l_nhigh,nkeep=l_nkeep,mclip=l_mclip,lsigma=l_lsigma,
	hsigma=l_hsigma,key_ron=l_key_ron,key_gain=l_key_gain,ron=l_ron,
	gain=l_gain,snoise=l_snoise,sigscale=l_sigscale,pclip=l_pclip,
	sci_ext=l_sci_ext,var_ext=l_var_ext,dq_ext=l_dq_ext,
	fl_vardq=l_fl_vardq,fl_dqprop=l_fl_dqprop,verbose=l_verbose)
    if (gemcombine.status != 0) {
	printlog("ERROR - GIFRINGE: GEMCOMBINE did not end successfully",
	    l_logfile,yes)
	goto error
    } 

    #####################          Deleting the OFFINT keyword
    flist=""   
    flist=filelist
    while (fscan(flist, img) != EOF) {
	for(ii=1; ii<=nsci; ii+=1) {                 
	    hedit(img//"["//l_sci_ext//","//ii//"]","OFFINT","",add-,addonly-,
	        delete+,verify-,show-,update+)
	}
    }
    flist=""

    # Update PHU
    gemhedit(l_output//"[0]","GEM-TLM",sdate,"Last modification with GEMINI")
    gemhedit(l_output//"[0]","GIFRINGE",sdate,"Time stamp for GIFRINGE")

    # clean up
    goto clean

error:
    status=1

clean:
    delete(filelist,verify-, >& "dev$null")
    delete(sextlist//","//vextlist//","//dextlist//","//extlist,verify-,
        >& "dev$null")
    hedit.show=s_verbose
    printlog(" ",l_logfile,l_verbose)
    printlog("GIFRINGE done",l_logfile,l_verbose)
    printlog("--------------------------------------------------------------------------------",
        l_logfile,l_verbose )
    flpr

end
     
