# Copyright(c) 2003-2004 Association of Universities for Research in Astronomy, Inc.

procedure mireduce(inimages)

# Routine "mireduce"
#
# This script carries out the basic mid-ir pipeline on either (1) raw T-RECS or
# Michelle data files or (2) output files from TPREPARE or MPREPARE.
# It is supposed to apply all the steps needed to reduce an initial raw image 
# to a two-dimensional output image.
#
# This task uses most of the other tasks in the "midir" package 
#
# Version  Oct  17, 2003   KV made changes so it works with Michelle data
#          Oct  10, 2003   KV rewrote internal logic to clean up execution
#          Oct  05, 2003   KV cleaned up warning messages
#          Oct  03, 2003   KV syntax changes in parameter names, added bpm
#          Sept 30, 2003   KV added functionality for MIPBM, MIFLAT, MIREGISTER
#          Sept 11, 2003   KV writes first version

char  inimages        {"",prompt="Input T-ReCS or Michelle image(s)"}	# OLDP-1-input-primary-single-prefix=r
char  outimages       {"",prompt="Output image(s)"}			# OLDP-1-output
char  rawpath         {"",prompt="Path for input raw images"}		# OLDP-4
char  outpref         {"r",prompt="Prefix for output image name(s)"}	# OLDP-4
bool  fl_background   {no,prompt="Apply T/M/MIBACKGROUND to the image?"}	# OLDP-2
bool  fl_view         {no,prompt="Apply T/M/MIVIEW to the image?"}		# OLDP-2
bool  fl_mask         {no,prompt="Apply a pixel mask to the image?"}		# OLDP-2
char  bpm             {"",prompt="bad pixel mask file name"}			# OLDP-1-input-fl_mask-required
bool  fl_flat         {no,prompt="Apply flatfield frame to the image?"}		# OLDP-2
char  flatfieldfile   {"",prompt="Flat field file name(s)"}			# OLDP-1-input-fl_mask-required
char  stackoption     {"stack",prompt="Image combining option: stack|register"}	# OLDP-2
bool  fl_display      {no,prompt="Display each final image?"}			# OLDP-2
char  logfile         {"",prompt="Log file name"}				# OLDP-1
bool  verbose         {yes,prompt="Verbose logging yes/no?"}			# OLDP-4
int status            {"",prompt="Exit error status: (0=good, >0=bad)"}		# OLDP-4
struct* scanfile      {"",prompt="Internal use only"}				# OLDP-4

begin

    char   l_inputimages,l_prefix,l_image,l_rawpath,l_outputimages
    char   l_flatfieldfile,l_stackoption,l_logfile,l_filename,l_mask
    bool   l_verbose,l_doflat,l_doprepare,l_dobackground,l_domask
    bool   l_doview,l_display
    char   in[200],out[200],flat[200]
    char   in1,out1
    int    nimages,modeflag,maximages,noutputimages,nflats
    int    i,j,k,l,m,n,m1,l1
    int    instrument,ngood
    char   tmpfile,phu,interfile,tempprefix,interfile1,header,extname
    char   tmpstring
    int    naxis[4],iaxis
    struct l_struct

    l_inputimages=inimages
    l_outputimages=outimages
    l_rawpath=rawpath
    l_flatfieldfile=flatfieldfile
    l_prefix=outpref
    l_stackoption=stackoption
    l_verbose=verbose
    l_doflat=fl_flat
    l_doview=fl_view
    l_dobackground=fl_background
    l_domask=fl_mask
    l_logfile=logfile
    l_display=fl_display
    l_mask=bpm

    status=0
    ngood=0

    # Check for package log file or user-defined log file

    if ((l_logfile=="") || (l_logfile==" ")) {
	l_logfile=midir.logfile

	if((l_logfile=="") || (l_logfile==" ")) {
	    #if logfile not defined, default to mireduce.log
	    l_logfile="midir.log"
	    printlog("Warning - MIREDUCE:  Both mireduce.logfile and \
	        midir.logfile are empty.  Using default file midir.log.",
		logfile=l_logfile, verbose+)
	}
    }

    # Open log file
    date | scan(l_struct)
    printlog("-------------------------------------------------------------\
        ---------------",logfile=l_logfile, verbose=l_verbose)
    printlog("MIREDUCE -- "//l_struct, l_logfile, l_verbose)
    printlog(" ",logfile=l_logfile, verbose=l_verbose)

    # Check the rawpath name for a final /
    if (substr(l_rawpath,(strlen(l_rawpath)),(strlen(l_rawpath))) != "/")
	l_rawpath=l_rawpath//"/"
    if (l_rawpath=="/" || l_rawpath==" ")
	l_rawpath=""

    printlog("Input T-ReCS or Michelle image(s)     = "//l_inputimages,
        l_logfile,l_verbose)
    printlog("Path for input images                 = "//l_rawpath,
        l_logfile,l_verbose)
    printlog("Output image(s)                       = "//l_outputimages,
        l_logfile,l_verbose)
    printlog("Prefix for output file names(s)       = "//l_prefix,
        l_logfile,l_verbose)
    printlog("Apply T/M/MIBACKGROUND to the image   = "//l_dobackground,
        l_logfile,l_verbose)
    printlog("Apply T/M/MIVIEW to the image         = "//l_doview,
        l_logfile,l_verbose)
    printlog("Apply a pixel mask to the image       = "//l_domask,
        l_logfile,l_verbose)
    printlog("bad pixel mask file name              = "//l_mask,
        l_logfile,l_verbose)
    printlog("Apply flatfield frame to the image    = "//l_doflat,
        l_logfile,l_verbose)
    printlog("Flat field file name(s)               = "//l_flatfieldfile,
        l_logfile,l_verbose)
    printlog("Image combiing option: stack|register = "//l_stackoption,
        l_logfile,l_verbose)
    printlog("Display each final image?             = "//l_display,
        l_logfile,l_verbose)
    printlog(" ",logfile=l_logfile, verbose=l_verbose)

    nimages=0
    nflats=0
    maximages=200

    tmpfile=mktemp("tmplist")

    # Count the number of input images
    # First, generate the file list if needed

    if (stridx("*",l_inputimages) > 0) {
	files(l_rawpath//l_inputimages, > tmpfile)
	l_inputimages="@"//tmpfile
    }

    if (substr(l_inputimages,1,1)=="@")
	scanfile=substr(l_inputimages,2,strlen(l_inputimages))
    else {
	if (stridx(",",l_inputimages)==0) 
	    files(l_inputimages, > tmpfile)
	else {
	    j=9999
	    while (j!=0) {
		j=stridx(",",l_inputimages)
		if (j>0)
		    files(substr(l_inputimages,1,j-1), >> tmpfile)
		else
		    files(l_inputimages, >> tmpfile)
		l_inputimages=substr(l_inputimages,j+1,strlen(l_inputimages))
	    }
	}
	scanfile=tmpfile
    }

    i=0
    while (fscan(scanfile,l_filename) != EOF) {
	i=i+1

	if (substr(l_filename,strlen(l_filename)-4,strlen(l_filename)) == ".fits")
	    l_filename=substr(l_filename,1,strlen(l_filename)-5)

	j=0
	if (stridx("/",l_rawpath) > 0 && stridx("/",l_filename) > 0) {
	    j=stridx("/",l_filename)
	    if (j > 0) {
		for (k=1; k < 100 && j > 0; k+=1) {
		    l_filename=substr(l_filename,j+1,strlen(l_filename))
		    j=stridx("/",l_filename)
		}
	    }
	}

	if (!imaccess(l_rawpath//l_filename)) {
	    printlog("ERROR - MIREDUCE: Input image "//l_rawpath//l_filename//\
	        " was not found.",logfile=l_logfile,verbose+)
	    status=status+1
	} else {
	    nimages=nimages+1
	    if (nimages > maximages) {
		printlog("ERROR - MIREDUCE: Maximum number of input \
		    images ["//str(maximages)//"] has been exceeded.",
		    logfile=l_logfile,verbose+)
		status=1
		goto exit
	    }
	    in[nimages]=l_filename
	    out[nimages]=l_filename
	    j=stridx("/",out[nimages])
	    if (j > 0) {
		for (k=1; k < 100 && j > 0; k+=1) {
		    out[nimages]=substr(out[nimages],j+1,strlen(out[nimages]))
		    j=stridx("/",out[nimages])
		}
	    }
	}
    }

    scanfile=""

    delete(tmpfile,ver-,>>& "dev$null")

    if (nimages == 0) {
	printlog("ERROR - MIREDUCE: no input images were defined.",
	    logfile=l_logfile,verbose+)
	status=1
	goto exit
    } else
	printlog("MIREDUCE: Processing "//str(nimages)//" image(s).",
	    logfile=l_logfile,verbose=l_verbose)

    tmpfile=mktemp("tmplist")

    if (l_doflat) {
	# Count the number of flat field images
	# First, generate the file list if needed

	if (stridx("*",l_flatfieldfile) > 0) {
	    files(l_flatfieldfile, > tmpfile)
	    l_flatfieldfile="@"//tmpfile
	}

	if (substr(l_flatfieldfile,1,1)=="@") {
	    scanfile=substr(l_flatfieldfile,2,strlen(l_flatfieldfile))
	} else {
	    if (stridx(",",l_flatfieldfile)==0) 
	        files(l_flatfieldfile, > tmpfile)
	    else {
		j=9999
		while (j!=0) {
		    j=stridx(",",l_flatfieldfile)
		    if (j>0)
		        files(substr(l_flatfieldfile,1,j-1), >> tmpfile)
		    else
			files(l_flatfieldfile, >> tmpfile)
		    l_flatfieldfile=substr(l_flatfieldfile,j+1,
		        strlen(l_flatfieldfile))
		}
	    }
	    scanfile=tmpfile
	}

	i=0
	while (fscan(scanfile,l_filename) != EOF) {
	    i=i+1

	    if (substr(l_filename,strlen(l_filename)-4,strlen(l_filename)) == ".fits")
		l_filename=substr(l_filename,1,strlen(l_filename)-5)

	    j=0
	    if (stridx("/",l_rawpath) > 0 && stridx("/",l_filename) > 0) {
		j=stridx("/",l_filename)
		if (j > 0) {
		    for (k=1; k < 100 && j > 0; k+=1) {
			l_filename=substr(l_filename,j+1,strlen(l_filename))
			j=stridx("/",l_filename)
		    }
		}
	    }

	    if (!imaccess(l_filename)) {
		printlog("ERROR - MIREDUCE: Flat field image "//l_filename//\
		    " was not found.",logfile=l_logfile,verbose+)
		status=status+1
	    } else {
		nflats=nflats+1
		if (nflats > maximages) {
		    printlog("ERROR - MIREDUCE: Maximum number of flat \
		        field images ["//str(maximages)//"] has been exceeded.",
			logfile=l_logfile,verbose+)
		    status=1
		    goto exit
		}
		flat[nflats]=l_filename
	    }
	}

	scanfile=""

	delete(tmpfile,ver-,>>& "dev$null")

	if (nflats == 0) {
	    printlog("ERROR - MIREDUCE: no flat field images were defined.",
	        logfile=l_logfile,verbose+)
	    status=1
	    goto exit
	}

	if (nflats == 1 && nimages > 1) {
	    for (j=2; j <= nimages; j+=1) {
		flat[j]=flat[1]
	    }
	    nflats=nimages
	}

	if (nflats != nimages) {
	    printlog("ERROR - MIREDUCE: The number flat field images ("//\
	        str(nflats)//") does not match the number of input images ("//\
		str(nimages)//").",logfile=l_logfile,verbose+)
	    status=1
	    goto exit
	}
    }

    # Now, do the same counting for the output file

    tmpfile=mktemp("tmplist")

    noutputimages=0
    if (l_outputimages != "" && l_outputimages != " ") {
	if (substr(l_outputimages,1,1) == "@")
	    scanfile=substr(l_outputimages,2,strlen(l_outputimages))
	else {
	    if (stridx("*",l_outputimages) > 0) {
		files(l_outputimages,sort-) | \
		    match(".hhd",stop+,print-,metach-, > tmpfile)
		scanfile=tmpfile
	    } else {
		files(l_outputimages,sort-, > tmpfile)
		scanfile=tmpfile
	    }
	}

	while (fscan(scanfile,l_filename) != EOF) {
	    if (substr(l_filename,strlen(l_filename)-4,strlen(l_filename)) == ".fits")
		l_filename=substr(l_filename,1,strlen(l_filename)-5)

	    noutputimages=noutputimages+1
	    if (noutputimages > maximages) {
	        printlog("ERROR - MIREDUCE: Maximum number of output \
		    images "//str(maximages)//" exceeded.",
		    logfile=l_logfile,verbose+)
		status=1
		goto exit
	    }
	    out[noutputimages]=l_filename
	    if (imaccess(out[noutputimages])) {
		printlog("ERROR - MIREDUCE: Output image "//l_filename//\
		    " already exists.",logfile=l_logfile,verbose+)
		status=1
		goto exit
	    }
	}
	if (noutputimages != nimages) {
	    printlog("ERROR - MIREDUCE: different number of input ("//\
	        str(nimages)//") and output ("//str(noutputimages)//\
		" image names have been specified.",logfile=l_logfile,verbose+)
	    status=1
	    goto exit
	}

	scanfile=""
	delete(tmpfile,ver-, >>& "dev$null")
    } else {
	if (l_prefix == "" || l_prefix == " ")
	    l_prefix="r"
	for (i=1; i <= nimages; i+=1) {
	    out[i]=l_prefix//out[i]
	    if (imaccess(out[i])) {
		printlog("ERROR - MIREDUCE: Output image "//out[i]//\
		    " already exists.",logfile=l_logfile,verbose+)
		status=1
		goto exit
	    }
	}
    }

    l=1
    while (l <= nimages) {

	j=0
	instrument=0
	l_image=in[l]
	if (!imaccess(l_rawpath//l_image)) {
	    printlog("ERROR - MIREDUCE: Input image "//l_rawpath//l_image//\
	        " could not be accessed.",logfile=l_logfile,verbose+)
	    status=status+1
	    goto nextimage
	}
	phu=l_rawpath//l_image//"[0]"
	imgets(phu,"INSTRUME",>& "dev$null")
	if (imgets.value == "TReCS") {
	    instrument=1
	    imgets(phu,"TPREPARE",>& "dev$null")
	    if (imgets.value == "0")
		l_doprepare=yes
	    else
		l_doprepare=no
	}
	if (imgets.value == "michelle") {
	    instrument=2
	    imgets(phu,"MPREPARE",>& "dev$null")
	    if (imgets.value == "0")
		l_doprepare=yes
	    else
	  l_doprepare=no
	}
	if (instrument == 0) {
	    printlog("ERROR - MIREDUCE: Input image "//in[l]//" is not a \
	        T-ReCS or Michelle image file.",logfile=l_logfile,verbose+)
	    status=status+1
	    goto nextimage
	}

	interfile=in[l]

	if (l_dobackground) {
	    interfile1=mktemp("reducetmpfile")
	    if (l_doprepare == no) {
		mibackground(interfile,outimages=interfile1,rawpath=l_rawpath)
	        if (mibackground.status == 0) 
		    interfile=interfile1
	        else {
                    printlog("ERROR - MIREDUCE: Task MIBACKGROUND failed for \
		        image "//in[l]//".",logfile=l_logfile,verbose+)
                    status=status+1
                    goto nextimage
	        }
	    } else {
	        if (instrument == 1) {
		    tbackground(interfile,outimages=interfile1,
		        rawpath=l_rawpath)
		    if (tbackground.status == 0)
		        interfile=interfile1
                    else {
		        printlog("ERROR - MIREDUCE: Task TBACKGROUND failed \
			    for image "//in[l]//".",logfile=l_logfile,verbose+)
		        status=status+1
		        goto nextimage      
		    }
		} else {
		    mbackground(interfile,outimages=interfile1,
		        rawpath=l_rawpath)
		    if (mbackground.status == 0)
			interfile=interfile1
		    else {
			printlog("ERROR - MIREDUCE: Task MBACKGROUND failed \
			    for image "//in[l]//".",logfile=l_logfile,verbose+)
		        status=status+1
		        goto nextimage
		    }
	        }
	    }

        }

	if (l_doview) {
	    interfile1=mktemp("reducetmpfile")
	    if (l_doprepare == no) {
	        if (interfile == l_image) 
                    miview(interfile,outimages=interfile1,rawpath=l_rawpath)
	  	else 
            	    miview(interfile,outimages=interfile1)
	  
		if (miview.status == 0) {
		    if (imaccess(interfile1)) 
			interfile=interfile1
            
		} else {
		    printlog("ERROR - MIREDUCE: Task MIVIEW failed for \
		        image "//in[l]//".",logfile=l_logfile,verbose+)
		    status=status+1
		    goto nextimage
		}
	    } else {
		if (instrument == 1) {
		    if (interfile == l_image)
		        tview(interfile,outimages=interfile1,rawpath=l_rawpath)
            	    else 
			tview(interfile,outimages=interfile1)
                    if (tview.status == 0) {
		        if (imaccess(interfile1)) 
			    interfile=interfile1
                    } else {
		        printlog("ERROR - MIREDUCE: Task TVIEW failed for \
		            image "//in[l]//".",logfile=l_logfile,verbose+)
		        status=status+1
		        goto nextimage
		    }
	        } else {
                    if (interfile == l_image) 
			mview(interfile,outimages=interfile1,rawpath=l_rawpath)
		    else 
			mview(interfile,outimages=interfile1)
            	    if (miview.status == 0) {
			if (imaccess(interfile1)) 
			    interfile=interfile1
                    } else {
			printlog("ERROR - MIREDUCE: Task MVIEW failed for \
			    image "//in[l]//".",logfile=l_logfile,verbose+)
			status=status+1
			goto nextimage
		    }
		}
	    }
	}

	if (l_stackoption == "stack") {
	    if (instrument == 1) {
		imgets(phu,"TPREPARE",>>& "dev$null")
		if (imgets.value == "0") {
		    tempprefix="t"
		    interfile1="t"//l_image
                    if (interfile1 == out[l] || imaccess(interfile1)) {
			k=0
			while (k < 20) {
			    tempprefix="t"//tempprefix
			    interfile1=tempprefix//l_image
			    if (!imaccess(interfile1)) k=50
			    k=k+1
			}
			if (k < 40) {
			    printlog("ERROR - MIREDUCE: Could not use the \
			        prefix "//"'t'"//" with file "//in[l]//\
				" because the output file already exists.",
				logfile=l_logfile,verbose+)
			    status=status+1
			    goto nextimage
			}
		    }
		    if (interfile == l_image)
			tprepare(interfile,rawpath=l_rawpath,
			    outimage=interfile1,stackoption=l_stackoption,
			    outpref="",verbose=l_verbose,logfile=l_logfile)
		    else
			tprepare(interfile,rawpath="",outimage=interfile1,
			    stackoption=l_stackoption,outpref="",
			    verbose=l_verbose,logfile=l_logfile)
		    j=tprepare.status
		    if (j != 0) {
			printlog("ERROR - MIREDUCE: Task TPREPARE failed for \
			    image "//in[l]//".",logfile=l_logfile,verbose+)
			status=status+1
			goto nextimage
		    }
		} else {
		    printlog("MIREDUCE: file "//l_image//" has already been \
		        'prepared'",logfile=l_logfile,verbose=l_verbose)
		    interfile1=interfile
		}
	    }
	    if (instrument == 2) {
		imgets(phu,"MPREPARE",>>& "dev$null")
		if (imgets.value == "0") {
		    tempprefix="m"
		    interfile1="m"//l_image
		    if (interfile1 == out[l] || imaccess(interfile1)) {
			k=0
			while (k < 20) {
			    tempprefix="m"//tempprefix
			    interfile1=tempprefix//l_image
			    if (!imaccess(interfile1)) k=50
			    k=k+1
			}
			if (k < 40) {
			    printlog("ERROR - MIREDUCE: Could not use the \
			        prefix "//"'m'"//" with file "//in[l]//\
				" because the output file already exists.",
				logfile=l_logfile,verbose+)
			    status=status+1
			    goto nextimage
			}
		    }
		    if (interfile == l_image) 
			mprepare(interfile,rawpath=l_rawpath,
			    outimage=interfile1,outpref="",verbose=l_verbose,
			    logfile=l_logfile)
		    else 
			mprepare(interfile,rawpath="",outimage=interfile1,
			    outpref="",verbose=l_verbose,logfile=l_logfile)
            
		    j=mprepare.status
		    if (j != 0) {
			printlog("ERROR - MIREDUCE: Task MPREPARE failed for \
			    image "//in[l]//".",logfile=l_logfile,verbose+)
			status=status+1
			goto nextimage
		    }
		} else {
		    printlog("MIREDUCE: file "//l_image//" has already been \
		        'prepared'.",logfile=l_logfile,verbose=l_verbose)
		    interfile1=interfile
		}
	    }
	    if (imaccess(interfile1))
		interfile=interfile1

	    imgets(phu,"MISTACK", >>& "dev$null")
	    if (imgets.value == "0") {
		interfile1=mktemp("reducetmpfile")
		if (interfile == l_image) 
		    mistack(interfile,rawpath=l_rawpath,outimages=interfile1,
		        outpref=l_prefix,logfile=l_logfile,verbose=l_verbose)
		else
		    mistack(interfile,outimages=interfile1,outpref=l_prefix,
		        logfile=l_logfile,verbose=l_verbose)
	  
		if (mistack.status > 0) {
		    printlog("ERROR - MIREDUCE: Task MISTACK failed for \
		        image "//in[l]//".",logfile=l_logfile,verbose+)
		    status=status+1
		    goto nextimage
		}
		if (imaccess(interfile1)) 
		    interfile=interfile1
	    } else {
		printlog("MIREDUCE: file "//l_image//" has already been \
		    'stacked'.",logfile=l_logfile,verbose=l_verbose)
		interfile1=interfile
	    }
	    if (imaccess(interfile1)) 
		interfile=interfile1
	}

	if (l_domask) {
	    if (l_mask == "") {
		if (instrument == 1) 
		    l_mask="midir$data/trecs.badpixels"
	  	else 
		    l_mask="midir$data/michelle.badpixels"
	    }
	    if (!access(l_mask)) {
		printlog("Error - MIREDUCE: the bad pixel mask file "//\
		    l_mask//" was not found.",l_logfile,l_verbose)
		status=status+1
		goto nextimage
	    }

	    tmpfile=mktemp("tmplist")
	    interfile1=mktemp("reducetmpfile")

	    if (interfile == l_image) 
		in1=l_rawpath//interfile
	    else 
		in1=interfile

	    fparse(in1)
	    in1 = fparse.directory//fparse.root//".fits"
	    fparse(interfile1)
	    out1 = fparse.directory//fparse.root//".fits"

	    copy(in1,out1,verbose-)
	    printlog("MIREDUCE: Copying image "//interfile//" to "//interfile1,
	        l_logfile,l_verbose)
	    header=interfile1//"[0]"
	    if (!imaccess(header)) {
		printlog("ERROR - MIREDUCE: Output file "//interfile1//" was \
		    not created properly.",logfile=l_logfile,verbose+)
		status=1
		goto nextimage
	    }

	    l1=0
	    for (m=1; m < 100000 && l1 == 0; m+=1) {
		extname=interfile1//"["//str(m)//"]"
		if (imaccess(extname)) {
		    imgets(extname,"i_naxis")
		    iaxis=int(imgets.value)
		    if (iaxis < 2 || iaxis > 4) {
			printlog("ERROR - MIREDUCE: Input image "//l_image//\
			    " has a bad number of dimensions ("//str(iaxis)//\
			    ") for extension "//str(m)//".",
			    logfile=l_logfile,verbose+)
		        status=status+1
			goto nextimage
		    }
		    naxis[1]=0
		    naxis[2]=0
		    naxis[3]=0
		    naxis[4]=0

		    for (n=1; n <= iaxis; n+=1) {
			tmpstring="naxis"//str(n)
			imgets(extname,tmpstring)
			naxis[n]=int(imgets.value)
			if (naxis[n] <= 0) {
			    printlog("ERROR - MIREDUCE: Image "//l_image//\
			        " has zero or negative image dimension(s) in \
				extension "//str(m)//".",
				logfile=l_logfile,verbose+)
			    status=status+1
			    goto nextimage
			}
		    }

		    if (naxis[1] != 320 || naxis[2] != 240) {
			printlog("ERROR - MIREDUCE: Image "//l_image//\
			    " does not have the correct image dimensions in \
			    extension "//str(m)//".",logfile=l_logfile,verbose+)
			status=status+1
			goto nextimage
		    }
		    if (naxis[3] == 0) naxis[3]=1
		    if (naxis[4] == 0) naxis[4]=1

		    for (m1=1; m1 <= naxis[3]; m1+=1) {
			for (n=1; n <= naxis[4]; n+=1) {
			    tmpfile=mktemp("tmplist")
			    imedit(extname,tmpfile,cursor=l_mask,display-)
			    if (iaxis == 2)
				imcopy(tmpfile,extname//"[1:320,1:240]",
				    verbose-)
			    if (iaxis == 3) 
				imcopy(tmpfile,extname//"[*,*,"//str(m1)//"]",
				    verbose-)
			    if (iaxis == 4) 
				imcopy(tmpfile,
				    extname//"[*,*,"//str(m1)//","//str(n)//"]",
				    verbose-)
        	
			    delete(tmpfile,ver-,>>& "dev$null")
			}
		    }
		} else 
		   l1=1
	    }

	    if (imaccess(interfile1)) 
		interfile=interfile1
	}

	if (l_doflat) {
	    if (!imaccess(flat[l])) {
		printlog("Error - MIREDUCE: the flat field file file "//\
		    flat[l]//" was not found.",l_logfile,l_verbose)
		status=status+1
		goto nextimage
	    }

	    tmpfile=mktemp("tmplist")
	    interfile1=mktemp("reducetmpfile")

	    if (interfile == l_image) 
		in1=l_rawpath//interfile
	    else 
		in1=interfile

	    m=stridx(".",in1)
	    n=stridx(".",interfile1)
	    if (m == 0) 
		in1=in1//".fits"
	    if (n == 0)
		out1=interfile1//".fits"
	    else
		out1=interfile1
	    copy(in1,out1,verbose-)
	    printlog("MIREDUCE: Copying image "//interfile//" to "//interfile1,
	        l_logfile,l_verbose)
	    header=interfile1//"[0]"
	    if (!imaccess(header)) {
		printlog("ERROR - MIREDUCE: Output file "//interfile1//" was \
		    not created properly.",logfile=l_logfile,verbose+)
		status=1
		goto nextimage
	    }

	    l1=0
	    for (m=1; m < 100000 && l1 == 0; m+=1) {
		extname=interfile1//"["//str(m)//"]"
		if (imaccess(extname)) {
		    imgets(extname,"i_naxis")
		    iaxis=int(imgets.value)
		    if (iaxis < 2 || iaxis > 4) {
		        printlog("ERROR - MIREDUCE: Input image "//l_image//\
			    " has a bad number of dimensions ("//str(iaxis)//\
			    ") for extension "//str(m)//".",
			    logfile=l_logfile,verbose+)
			status=status+1
			goto nextimage
		    }
		    naxis[1]=0
		    naxis[2]=0
		    naxis[3]=0
		    naxis[4]=0

		    for (n=1; n <= iaxis; n+=1) {
			tmpstring="naxis"//str(n)
			imgets(extname,tmpstring)
			naxis[n]=int(imgets.value)
			if (naxis[n] <= 0) {
			    printlog("ERROR - MIREDUCE: Image "//l_image//\
			        " has zero or negative image dimension(s) in \
				extension "//str(m)//".",
				logfile=l_logfile,verbose+)
			    status=status+1
			    goto nextimage
			}
		    }

		    if (naxis[1] != 320 || naxis[2] != 240) {
			printlog("ERROR - MIREDUCE: Image "//l_image//\
			    " does not have the correct image dimensions in \
			    extension "//str(m)//".",logfile=l_logfile,verbose+)
			status=status+1
			goto nextimage
		    }
		    if (naxis[3] == 0) naxis[3]=1
		    if (naxis[4] == 0) naxis[4]=1

		    for (m1=1; m1 <= naxis[3]; m1+=1) {
			for (n=1; n <= naxis[4]; n+=1) {
			    tmpfile=mktemp("tmplist")
			    imarith(extname,"*",flat[l],tmpfile)
			    if (iaxis == 2) 
				imcopy(tmpfile,extname//"[1:320,1:240]",
				    verbose-)
			    if (iaxis == 3) 
				imcopy(tmpfile,extname//"[*,*,"//str(m1)//"]",
				    verbose-)
			    if (iaxis == 4)
				imcopy(tmpfile,
				    extname//"[*,*,"//str(m1)//","//str(n)//"]",
				    verbose-)
			    delete(tmpfile,ver-,>>& "dev$null")
			}
		    }
		} else
		    l1=1
	    }

	    if (imaccess(interfile1)) 
		interfile=interfile1
	}

	if (l_stackoption == "register") {
	    if (instrument == 1) {
		imgets(phu,"TPREPARE",>>& "dev$null")
		if (imgets.value == "0") {
		    tempprefix="t"
		    interfile1="t"//l_image
		    if (interfile1 == out[l] || imaccess(interfile1)) {
		        k=0
			while (k < 20) {
			    tempprefix="t"//tempprefix
			    interfile1=tempprefix//l_image
			    if (!imaccess(interfile1)) k=50
			    k=k+1
			}
			if (k < 40) {
			    printlog("ERROR - MIREDUCE: Could not use the \
			        prefix "//"'t'"//" with file "//in[l]//\
				" because the output file already exists.",
				logfile=l_logfile,verbose+)
			    status=status+1
			    goto nextimage
			}
		    }
		    if (interfile == l_image)
			tprepare(interfile,rawpath=l_rawpath,
			    outimage=interfile1,stackoption=l_stackoption,
			    outpref="",verbose=l_verbose,logfile=l_logfile)
		    else
			tprepare(interfile,rawpath="",outimage=interfile1,
			    stackoption=l_stackoption,outpref="",
			    verbose=l_verbose,logfile=l_logfile)
		    j=tprepare.status
		    if (j != 0) {
			printlog("ERROR - MIREDUCE: Task TPREPARE failed for \
			    image "//in[l]//".",logfile=l_logfile,verbose+)
			status=status+1
			goto nextimage
		    }
		} else {
		    printlog("MIREDUCE: file "//l_image//" has already been \
		        'prepared'",logfile=l_logfile,verbose=l_verbose)
		    interfile1=interfile
		}
	    }
	    if (instrument == 2) {
		imgets(phu,"MPREPARE",>>& "dev$null")
		if (imgets.value == "0") {
		    tempprefix="m"
		    interfile1="m"//l_image
		    if (interfile1 == out[l] || imaccess(interfile1)) {
			k=0
			while (k < 20) {
			    tempprefix="m"//tempprefix
			    interfile1=tempprefix//l_image
			    if (!imaccess(interfile1)) k=50
			    k=k+1
			}
			if (k < 40) {
			    printlog("ERROR - MIREDUCE: Could not use the \
			        prefix "//"'m'"//" with file "//in[l]//\
				" because the output file already exists.",
				logfile=l_logfile,verbose+)
			    status=status+1
			    goto nextimage
			}
		    }
		    if (interfile == l_image)
			mprepare(interfile,rawpath=l_rawpath,
			    outimage=interfile1,outpref="",verbose=l_verbose,
			    logfile=l_logfile)
		    else 
			mprepare(interfile,rawpath="",outimage=interfile1,
			    outpref="",verbose=l_verbose,logfile=l_logfile)
		    j=mprepare.status
		    if (j != 0) {
			printlog("ERROR - MIREDUCE: Task MPREPARE failed for \
			    image "//in[l]//".",logfile=l_logfile,verbose+)
			status=status+1
			goto nextimage
		    }
		} else {
		    printlog("MIREDUCE: file "//l_image//" has already been \
		        'prepared'.",logfile=l_logfile,verbose=l_verbose)
		    interfile1=interfile
		}
	    }
	    if (imaccess(interfile1))
		interfile=interfile1

	    imgets(phu,"MIREGIST", >>& "dev$null")
	    if (imgets.value == "0") {
	        interfile1=mktemp("reducetmpfile")
		if (interfile == l_image)
		    miregister(interfile,rawpath=l_rawpath,outimages=interfile1,
		        outpref=l_prefix,logfile=l_logfile,verbose=l_verbose)
		else
		    miregister(interfile,outimages=interfile1,outpref=l_prefix,
		        logfile=l_logfile,verbose=l_verbose)
		if (miregister.status > 0) {
		    printlog("ERROR - MIREDUCE: Task MIREGISTER failed for \
		        image "//in[l]//".",logfile=l_logfile,verbose+)
		    status=status+1
		    goto nextimage
		}
		if (imaccess(interfile1))
		    interfile=interfile1
	    } else {
		printlog("MIREDUCE: file "//l_image//" has already been \
		    'stacked'.",logfile=l_logfile,verbose=l_verbose)
		interfile1=interfile
	    }
	}

	k=stridx(".",interfile1)
	j=stridx(".",out[l])
	if (k == 0)
	    in1=interfile1//".fits"
	else
	    in1=interfile1
	if (j == 0)
	    out1=out[l]//".fits"
	else
	    out1=out[l]
	copy(in1,out1,verbose-)
	printlog("MIREDUCE: Renaming image "//interfile1//" to "//out[l],
	    l_logfile,l_verbose)

	if (l_display)
	    display(out[l]//"[1]")

	# Time stamp the primary header
	#
	date | scan(l_struct)
	gemhedit(out[l]//"[0]","MIREDUCE",l_struct,"Time stamp for MIREDUCE")

	ngood=ngood+1

nextimage:

	delete("reducetmpfile*.fits",ver-,>>& "dev$null")
	delete("tmplist*",ver-,>>& "dev$null")
	l=l+1
    }

exit:

    date | scan(l_struct)
    if (status == 0) {
	printlog("MIREDUCE: exit status => good.  "//l_struct,
	    logfile=l_logfile, verbose=l_verbose)
	printlog("MIREDUCE: all "//str(nimages)//" successfully reduced.",
	    logfile=l_logfile, verbose+)
    } else {
	printlog("MIREDUCE: exit status => error.  "//l_struct,
	    logfile=l_logfile, verbose=l_verbose)
	printlog("MIREDUCE: "//str(ngood)//" of "//str(nimages)//\
	    " successfully processed.",logfile=l_logfile, verbose+)
    }
    printlog("--------------------------------------------------------------\
        --------------", logfile=l_logfile, verbose=l_verbose)
    printlog(" ",logfile=l_logfile, verbose=l_verbose)

end
