# Copyright(c) 2003-2004 Association of Universities for Research in Astronomy, Inc.

procedure mistack(inimages)

# This routine stacks the individual frames of a "prepared" T-ReCS or
# Michelle file.
#
# Version:  June 23, 2003  KV wrote original script from BRs prototype 
#           Sept  5, 2003  TB made update for 3D file structure, added outprefix 
#                             option, omitted all TReCS "saveset" info.
#           Oct   1, 2003  KV changed the routine to handle all observing modes
#           Oct   3, 2003  KV made small changes to the parameter syntax
#           Oct   5, 2003  KV made changes to allowed MODE FITS keywords for Michelle
#           Oct  29, 2003  KL IRAF 2.12 - new/modified parameters
#                              imcombine: rejmask->rejmasks, plfile->nrejmasks
#                                         headers,bpmasks,expmasks,outlimits
#			       hedit: addonly
#           Dec   3, 2003  KV removed a 'delete("tmp*")' statement, went back 
#                              to specific deletes of tmp working files.

char  inimages     {"",prompt="Input T-ReCS or Michelle image(s)"}	# OLDP-1-input-primary-single-prefix=s
char  rawpath      {"",prompt="Path for in raw images"}      		# OLDP-4
char  outimages    {"",prompt="Output image(s)"}             		# OLDP-1-output
char  outpref      {"s",prompt="Prefix for out image(s)"}    		# OLDP-4
char  logfile      {"",prompt="Logfile"}                        	# OLDP-1
bool  verbose      {yes,prompt="Verbose"}            			# OLDP-4
int status         {"",prompt="Exit error status: (0=good, >0=bad)"}	# OLDP-4
struct* scanfile   {"",prompt="Internal use only"}   			# OLDP-4

begin

    char   l_inputimages, l_outputimages, l_filename, l_prefix, l_logfile
    char   l_rawpath
    int    i,j,k,l,l_nodset,l_saveset,itotal,n_sig
    int    n_ref,nimages,maximages,noutimages,l_frames,l_extensions
    int    source,reference,nbadsets,badsetnumber[100],badflag
    int    aframe, nbad
    char   in[100],out[100],header,exheader,instrument, l_temp
    char   tmpon,tmpfile,tmpfinal,tmpfile1,tmpfile2,tmphead
    real   exptime,norm,ave1,ave2,diff1
    int    modeflag
    bool   l_verbose
    struct l_struct

    tmpfile=mktemp("tmpin")
    tmpon=mktemp("tmpon")
    tmphead=mktemp("tmphead")

    l_verbose=verbose 
    l_inputimages=inimages
    l_outputimages=outimages
    l_logfile=logfile
    l_prefix=outpref
    l_rawpath=rawpath

    nimages=0
    maximages=100
    status=0

    # Test that name of logfile makes sense
    #cache("midir")
    print(l_logfile) | scan(l_logfile)
    if (l_logfile=="" || l_logfile==" ") {
	l_logfile=midir.logfile
	print(l_logfile) | scan(l_logfile)
	if ((l_logfile=="") || (l_logfile==" ")) {
	    l_logfile="midir.log"
	    printlog("MISTACK: Both mistack.logfile and midir.logfile are \
	        empty.",l_logfile,l_verbose)
	    printlog("                  Using default file midir.log.",
	        l_logfile, verbose+)
	}
    }

    # Open log file
    date | scan(l_struct)
    printlog("-------------------------------------------------------------\
        ---------------",l_logfile,l_verbose)
    printlog("MISTACK -- "//l_struct,l_logfile,l_verbose)
    printlog(" ",l_logfile,l_verbose)

    #----------------------------------------------------------------------
    # Check the rawpath name for a final /
    if (substr(l_rawpath,(strlen(l_rawpath)),(strlen(l_rawpath))) != "/")
	l_rawpath=l_rawpath//"/"
    if (l_rawpath=="/" || l_rawpath==" ")
	l_rawpath=""

    printlog("Input mid-ir image(s)                 = "//l_inputimages,
        l_logfile,l_verbose)
    printlog("Path for input images                 = "//l_rawpath,
        l_logfile,l_verbose)
    printlog("Output image(s)                       = "//l_outputimages,
        l_logfile,l_verbose)
    printlog("Prefix for output file names(s)       = "//l_prefix,
        l_logfile,l_verbose)
    printlog(" ",logfile=l_logfile, verbose=l_verbose)

    #-----------------------------------------------------------------------
    # Load up arrays of input name lists
    # This version handles both *s, / and commas in l_inputimages

    # check that list file exists
    if (substr(l_inputimages,1,1)=="@") {
	l_temp=substr(l_inputimages,2,strlen(l_inputimages))
	if (!access(l_temp) && !access(l_rawpath//l_temp)) {
	    printlog("ERROR - MISTACK:  Input file "//l_temp//" not found.",
	        l_logfile,l_verbose)
	    status=1
	    goto clean
	}
    }


    # Count the number of in images
    # First, generate the file list if needed

    if (stridx("*",l_inputimages) > 0) {
	files(l_inputimages, > tmpfile)
	l_inputimages="@"//tmpfile
    }

    if (substr(l_inputimages,1,1)=="@")
	scanfile=substr(l_inputimages,2,strlen(l_inputimages))
    else {
	files(l_inputimages,sort-, > tmpfile)
	scanfile=tmpfile
    }

    i=0
    while (fscan(scanfile,l_filename) != EOF && i <= 10) {
	i=i+1
	if (substr(l_filename,strlen(l_filename)-4,strlen(l_filename)) == ".fits")
	    l_filename=substr(l_filename,1,strlen(l_filename)-5)

	if (!imaccess(l_filename) && !imaccess(l_rawpath//l_filename)) {
	    printlog("ERROR - MISTACK: Input image"//l_filename//" was not \
	        found.",l_logfile, l_verbose)
	    status=1
	    goto clean
	} else {
	    nimages=nimages+1
	    if (nimages > maximages) {
		printlog("ERROR - MISTACK: Maximum number of input images \
		    exceeded:"//maximages,l_logfile,l_verbose)
		status=1 
		goto clean
	    }
	    if (l_rawpath=="" || l_rawpath==" ")
	        in[nimages]=l_filename
	    else
	        in[nimages]=l_rawpath//l_filename
	}
    }

    scanfile=""
    delete(tmpfile,ver-,>>& "/dev/null")

    if (nimages == 0) {
	printlog("ERROR - MISTACK: No input images defined.",l_logfile,verbose+)
	status=1
	goto clean
    }

    # Now, do the same counting for the out file

    nbad=0
    noutimages=0
    if ((l_outputimages != "") && (l_outputimages != " ")) {
	if (substr(l_outputimages,1,1) == "@")
	    scanfile=substr(l_outputimages,2,strlen(l_outputimages))
	else {
	    if (stridx("*",l_outputimages) > 0) {
		files(l_outputimages,sort-) | match(".hhd",stop+,print-,
		    metach-, > tmpfile)
		scanfile=tmpfile
	    } else {
		files(l_outputimages,sort-, > tmpfile)
		scanfile=tmpfile
	    }
	}

	while (fscan(scanfile,l_filename) != EOF) {
	    noutimages=noutimages+1
	    if (noutimage > maximages) {
		printlog("ERROR - MISTACK: Maximum number of output images \
		    exceeded:"//maximages,l_logfile, verbose+)
		status=1
		goto clean
	    }
	    out[noutimages]=l_filename
	    if (imaccess(out[noutimages])) {
		printlog("ERROR - MISTACK: Output image "//l_filename//\
		    " already exists.",l_logfile,l_verbose)
		nbad+=1
	    }
	}
	if (noutimages != nimages) {
	    printlog("ERROR - MISTACK: Different number of in images \
	        ("//nimages//") and out images ("//noutimages//")",
		l_logfile, verbose+)
	    status =1
	    goto clean
	}

	scanfile=""
	delete(tmpfile,ver-, >>& "/dev/null",>>& "/dev/null")

    } else {  # If prefix is to be used instead of filename
    
	print(l_prefix) | scan(l_prefix)
	if (l_prefix=="" || l_prefix==" ") {
	    printlog("ERROR - MISTACK: Neither output image name nor output \
	        prefix is defined.",l_logfile,verbose+)
	    status=1
	    goto clean
	}
	
	i=1
	while (i<=nimages) {
	    fparse(in[i])
	    out[i]=l_prefix//fparse.root//".fits"
	    if (imaccess(out[i])) {
		printlog("ERROR - MISTACK: Output image "//out[i]//" already \
		    exists.",l_logfile,verbose+)
		nbad+=1
	    }
	    i=i+1
	}
    }
    if (nbad > 0) {
	printlog("ERROR - MISTACK: "//nbad//" image(s) already exist.",
	    l_logfile,verbose+)
	status=1
	goto clean
    }

    #
    # The main loop: accumulate source and reference images
    #

    nbad=0
    i=1
    while (i <= nimages) {
        # Create tmp FITS file name used within this loop
	tmpfinal=mktemp("tmpfinal")

	imgets(in[i]//"[0]","MISTACK", >>& "dev$null")
	if (imgets.value != "0") {
	    printlog("MISTACK: file "//in[i]//" has already been stacked.",
	        l_logfile,l_verbose)
	    goto nextimage
	}

	printlog("MISTACK:  "//in[i]//" --> "//out[i],l_logfile,l_verbose)

	# check the primary FITS header
	header=in[i]//"[0]"

	imgets(header,"INSTRUMENT")
	instrument=imgets.value
	printlog("Instrument is:"//instrument,logfile=l_logfile, verbose+)

	if (instrument == "michelle") {
	    imgets(header,"MPREPARE", >& "dev$null")
	    if (imgets.value == "0") {
		printlog("ERROR - MISTACK: Image "//in[i]//" not MPREPAREd.",
		    l_logfile,l_verbose)
		status=1
		goto clean
	    }
	}


	if (instrument == "TReCS") {
	    imgets(header,"TPREPARE", >& "dev$null")
	    if (imgets.value == "0") {
		printlog("ERROR - MISTACK: Image "//in[i]//" not TPREPAREd.",
		    l_logfile,l_verbose)
		status=1
		goto clean
	    }
	}

	# find the observation mode
	#
	if (instrument == "TReCS") {
	    imgets(header,"OBSMODE", >& "dev$null")
	    if (imgets.value == "0") {
	  	printlog("ERROR - MISTACK: could not find the OBSMODE from \
		    the primary header.",l_logfile,l_verbose)
		status=status+1
		goto nextimage
	    }
	    modeflag=0
	    if (imgets.value == "chop-nod") modeflag=1
	    if (imgets.value == "chop") modeflag=2
	    if (imgets.value == "nod") modeflag=3
	    if (imgets.value == "stare") modeflag=4
	    if (modeflag == 0) {
		printlog("ERROR - MISTACK: Unrecognised OBSMODE \
		    ("//imgets.value//") in the primary header.",
		    l_logfile,l_verbose)
		status=status+1
		goto nextimage
	    }
	} else {
	    imgets(header,"MODE", >& "dev$null")
	    if (imgets.value == "0") {
        	printlog("ERROR - MISTACK: could not find the MODE from the \
		    primary header.",l_logfile,l_verbose)
        	status=status+1
        	goto nextimage
	    }
	    # Change these according to the Michelle "MODE" keywords.
	    # I am not sure whether there are other "non-destructive" (nd) modes,
	    # or whether these keywords are all correct.
	    modeflag=0
	    if (imgets.value == "chop-nod") modeflag=1
	    if (imgets.value == "ndchop") modeflag=1
	    if (imgets.value == "chop") modeflag=2
	    if (imgets.value == "nod") modeflag=3
	    if (imgets.value == "ndstare") modeflag=4
	    if (imgets.value == "stare") modeflag=4
	    if (modeflag == 0) {
		printlog("ERROR - MISTACK: Unrecognised MODE \
		    ("//imgets.value//") in the primary header.",
		    l_logfile,l_verbose)
		status=status+1
		goto nextimage
	    }
	}


	# Count the number of extensions

	l_extensions=1
	while (imaccess(in[i]//"["//l_extensions//"]")) {
	    imgets(in[i]//"["//l_extensions//"]","i_naxis")
	    if (modeflag == 1 || modeflag == 2) {
		if (imgets.value != "3" && l_extensions > 0) {
		    printlog("ERROR - MISTACK: extension "//l_extension//\
		        " has "//imgets.value//" dimensions.  It should be 3.",
			l_logfile,l_verbose)
		    status=status+1
		    goto nextimage
		}
	    }
	    if (modeflag == 3 || modeflag == 4) {
		if (imgets.value != "2" && l_extensions > 0) {
		    printlog("ERROR - MISTACK: extension "//l_extension//\
		        " has "//imgets.value//" dimensions.  It should be 2.",
			l_logfile,l_verbose)
		    status=status+1
		    goto nextimage
		}
	    }
	    l_extensions=l_extensions+1
	}

	j=l_extensions-1
	printlog("MISTACK: Number of extensions is "//j,l_logfile, l_verbose)

	if (modeflag == 1) {
	    if (2*(j/2) != j) {
		printlog("ERROR - MISTACK: number of extensions for input \
		    file "//in[i]//" does not correspond to complete nodsets.  \
		    Going to next image.",logfile=l_logfile, verbose+)
		status=status+1
		goto nextimage
	    }

	    for (j=1; j < l_extensions; j=j+2) {
	        # Create tmp FITS file names created within this loop
		# ** DO NOT DELETE tmpfile1 at the bottom of the loop:
		# ** the images are used later on.
		tmpfile1=mktemp("tmpfile1")
		tmpfile2=mktemp("tmpfile2")

		imgets(in[i]//"["//j//"]","BADNOD",>>&"/dev/null")
		if (imgets.value != "0") {
		    imgets(in[i]//"["//j+1//"]","BADNOD",>>& "/dev/null")
		    if (imgets.value != "0")
		        printlog("Warning - MISTACK:  Nods "//j//\
			    " and "//j+1//" are both bad.)",l_logfile,l_verbose)
		    else
		        printlog("Warning - MISTACK:  Nod "//j//" is bad.  \
			    (Omitting both Nod "//j//" and "//j+1//")",
			    l_logfile,l_verbose)
		} else {
		    imgets(in[i]//"["//j+1//"]","BADNOD",>>& "/dev/null")
		    if (imgets.value != "0") {
			printlog("Warning - MISTACK:  Nod "//j+1//" is bad.  \
			    (Omitting both Nod "//j//" and "//j+1//")",
			    l_logfile,l_verbose)
		    } else {
			imarith(in[i]//"["//j//"]"//"[*,*,3].fits","+",
			    in[i]//"["//j+1//"]"//"[*,*,3].fits",tmpfile2,ver-)
			imarith(tmpfile2,"/","2.0",tmpfile1,ver-)
			print(tmpfile1, >> tmpon)
			
			# Delete tmp files not needed any longer
			# ** DO NOT DELETE tmpfile1
			imdelete(tmpfile2,ver-,>& "dev$null")
		    }
		}
	    }

	    imcombine("@"//tmpon,tmpfinal,headers="",bpmasks="",rejmasks="",
                nrejmasks="",expmasks="",sigmas="",logfile=l_logfile,
                combine="average",reject="none",project=no,outtype="double",
	        outlimits="")

	    wmef(tmpfinal,out[i],extname="SCI",phu=header,verb-,>>& "dev$null")

	    # Time stamps
	    date | scan(l_struct)
	    printf("%-8s= \'%-18s\' / %-s\n","MISTACK",l_struct,"Time stamp for MISTACK", >> tmphead)

	    mkheader(out[i]//"[0]",tmphead,append+,verbose-)
	    delete(tmphead,verify-, >>& "dev$null")

	    hedit(out[i]//"[1]","AXISLAB4","",add-,addonly-,delete+,verify-,
	        show-,update+)
	    hedit(out[i]//"[1]","AXISLAB3","",add-,addonly-,delete+,verify-,
	        show-,update+)
	}

	if (modeflag == 2) {
	    if (j > 1) {
		printlog("ERROR - MISTACK: More than one extension \
		    ("//j//") in CHOP mode.",l_logfile,l_verbose)
		status=status+1
		goto nextimage
	    }

	    imgets(in[i]//"[1]","BADNOD",>>& "/dev/null")
	    if (imgets.value != "0") {
		printlog("Error - MISTACK:  The only NOD is marked as BAD.",
		    l_logfile,l_verbose)
		status=status+1
		goto nextimage
	    }

	    tmpfile1=mktemp("tmpfile1")
	    imcopy(in[i]//"[1][*,*,3].fits",tmpfile1,ver-)
	    print(tmpfile1, >> tmpon)

	    imcombine("@"//tmpon,tmpfinal,headers="",bpmasks="",rejmasks="",
		nrejmasks="",expmasks="",sigmas="",logfile=l_logfile,
		combine="average",reject="none",project=no,outtype="double",
		outlimits="")

	    wmef(tmpfinal,out[i],extname="SCI",phu=header,verb-,>>& "dev$null")


	    # Time stamps
	    date | scan(l_struct)
	    printf("%-8s= \'%-18s\' / %-s\n","MISTACK",l_struct,
	        "Time stamp for MISTACK", >> tmphead)

	    mkheader(out[i]//"[0]",tmphead,append+,verbose-)
	    delete(tmphead,verify-, >>& "dev$null")
	    hedit(out[i]//"[1]","AXISLAB4","",add-,addonly-,delete+,verify-,
	        show-,update+)
	    hedit(out[i]//"[1]","AXISLAB3","",add-,addonly-,delete+,verify-,
	        show-,update+)
	}

	if (modeflag == 3) {

	    if (2*(j/2) != j) {
		printlog("ERROR - MISTACK: number of extensions for input \
		    file "//in[i]//" does not correspond to complete nodsets.  \
		    Going to next image.",logfile=l_logfile, verbose+)
		status=status+1
		goto nextimage
	    }

	    for (j=1; j < l_extensions; j=j+2) {
	        # Create tmp FITS file names created within this loop
                # ** DO NOT DELETE tmpfile1 at the bottom of the loop: 
                # ** the images are used later on.
                tmpfile1=mktemp("tmpfile1")
                tmpfile2=mktemp("tmpfile2")

		imgets(in[i]//"["//j//"]","BADNOD",>>& "/dev/null")
		if (imgets.value != "0") {
		    imgets(in[i]//"["//j+1//"]","BADNOD",>>& "/dev/null")
		    if (imgets.value != "0")
			printlog("Warning - MISTACK:  Nods "//j//\
			    " and "//j+1//" are both bad.)",l_logfile,l_verbose)
		    else
			printlog("Warning - MISTACK:  Nod "//j//" is bad.  \
			    (Omitting both Nod "//j//" and "//j+1//")",
			    l_logfile,l_verbose)
		} else {
		    imgets(in[i]//"["//j+1//"]","BADNOD",>>& "/dev/null")
		    if (imgets.value != "0") {
			printlog("Warning - MISTACK:  Nod "//j+1//" is bad.  \
			    (Omitting both Nod "//j//" and "//j+1//")",
			    l_logfile,l_verbose)
		    } else {
			imarith(in[i]//"["//j//"]"//"[*,*].fits","+",
			    in[i]//"["//j+1//"]"//"[*,*].fits",tmpfile2,ver-)
			imarith(tmpfile2,"/","2.0",tmpfile1,ver-)
			print(tmpfile1, >> tmpon)
			
			# Delete tmp files not needed any longer
                        # ** DO NOT DELETE tmpfile1, tmprefimage
                        imdelete(tmpfile2,ver-,>& "dev$null")
		    }
		}
	    }

	    imcombine("@"//tmpon,tmpfinal,headers="",bpmasks="",rejmasks="",
		nrejmasks="",expmasks="",sigmas="",logfile=l_logfile,
		combine="average",reject="none",project=no,outtype="double",
		outlimits="")

	    wmef(tmpfinal,out[i],extname="SCI",phu=header,verb-,>>& "dev$null")

	    # Time stamps
	    date | scan(l_struct)
	    printf("%-8s= \'%-18s\' / %-s\n","MISTACK",l_struct,
	        "Time stamp for MISTACK", >> tmphead)

	    mkheader(out[i]//"[0]",tmphead,append+,verbose-)
	    delete(tmphead,verify-, >>& "dev$null")

	    hedit(out[i]//"[1]","AXISLAB4","",add-,addonly-,delete+,verify-,
	        show-,update+)
	    hedit(out[i]//"[1]","AXISLAB3","",add-,addonly-,delete+,verify-,
	        show-,update+)
	}

	if (modeflag == 4) {
	    if (j > 1) {
		printlog("ERROR - MISTACK: More than one extension ("//j//") \
		    in STARE mode.",l_logfile,l_verbose)
		status=status+1
		goto nextimage
	    }

	    imgets(in[i]//"[1]","BADNOD",>>& "/dev/null")
	    if (imgets.value != "0") {
		printlog("Error - MISTACK:  The only NOD is marked as BAD.",
		    l_logfile,l_verbose)
		status=status+1
		goto nextimage
	    }

	    tmpfile1=mktemp("tmpfile1")
	    imcopy(in[i]//"[1][*,*].fits",tmpfile1,ver-)
	    print(tmpfile1, >> tmpon)

	    imcombine("@"//tmpon,tmpfinal,headers="",bpmasks="",rejmasks="",
		nrejmasks="",expmasks="",sigmas="",logfile=l_logfile,
		combine="average",reject="none",project=no,outtype="double",
		outlimits="")

	    wmef(tmpfinal,out[i],extname="SCI",phu=header,verb-,>>& "dev$null")


	    # Time stamps
	    date | scan(l_struct)
	    printf("%-8s= \'%-18s\' / %-s\n","MISTACK",l_struct,
	        "Time stamp for MISTACK", >> tmphead)

	    mkheader(out[i]//"[0]",tmphead,append+,verbose-)
	    delete(tmphead,verify-, >>& "dev$null")

	    hedit(out[i]//"[1]","AXISLAB4","",add-,addonly-,delete+,verify-,
	        show-,update+)
	    hedit(out[i]//"[1]","AXISLAB3","",add-,addonly-,delete+,verify-,
	        show-,update+)
	}
      
	# jump to here if there is a problem

nextimage:
	i=i+1
	delete(tmpfinal//".fits",ver-,>& "dev$null")
    }


clean:
    if (status==0) {
	printlog(" ",l_logfile,l_verbose)
	printlog("MISTACK exit status:  good.",l_logfile,l_verbose)
    }
    if (status!=0) {
	printlog(" ",l_logfile,l_verbose)
	printlog("MISTACK: Exited with errors. ",l_logfile,l_verbose)
    }

    scanfile="" 
    delete(tmpon,ver-, >>& "dev$null")
    delete(tmpfile//","//tmphead,ver-, >>& "dev$null")
    delete("tmpin*",ver-, >>& "dev$null")
    delete("tmpon*",ver-, >>& "dev$null")
    delete("tmpfinal*",ver-, >>& "dev$null")
    delete("tmpfile1*",ver-, >>& "dev$null")
    delete("tmpfile*",ver-, >>& "dev$null")
    printlog("-------------------------------------------------------------\
        ---------------",l_logfile,l_verbose)
    printlog(" ",logfile=l_logfile, verbose=l_verbose)


end
