# Copyright(c) 2003-2004 Association of Universities for Research in Astronomy, Inc.

procedure tprepare(inimages)

#
# This routine converts a raw T-ReCS file to the standard "midir" format, 
# combining savesets for each nod.
#
# Version:  Sept  5, 2003  KV started writing the script from "trgather.cl"
#           Sept  9, 2003  KV added "outpref" based on TBs example
#           Sept 26, 2003  KV register option added, logging changed a bit
#	    Oct   3, 2003  KV syntax changes to the parameters, changed to 
#                             using imstack
#	    Oct  29, 2003  KL IRAF 2.12 - new/modified parameters
#				imcombine: headers,bpmasks,expmask,outlimits
#				           rejmask->rejmasks,plfile->nrejmasks
#				hedit: addonly
#           Nov  12, 2003  KV delete "AXISLAB4" from extension headers
#                             add WCS values to extension headers
#           Dec  02, 2003  KV took out the WCS values since these crash 
#                             imcombine for subsequent steps (i.e. mistack)

char  inimages     {"",prompt="Input T-ReCS image(s)"}
char  rawpath      {"",prompt="Path for input raw images"}
char  outimages    {"",prompt="Output image(s)"}
char  outpref      {"t",prompt="Prefix for output image name(s)"}
char  stackoption  {"stack",prompt="Image combining option: stack|register"}
char  logfile      {"",prompt="Logfile name"}
bool  verbose      {yes,prompt="Verbose logging yes/no?"}
int status         {"",prompt="Exit error status: (0=good, >0=bad)"}
struct* scanfile   {"",prompt="Internal use only"}

begin

    char   l_inputimages,l_outputimages,l_filename,l_logfile
    char   l_rawpath,l_prefix,l_stack
    int    i,j,k,l,i_nodset,i_saveset,itotal,n_sig
    int    n_ref,nimages,maximages,noutputimages,i_frames,i_extensions
    int    source,reference,nodcount,nbadsets,badsetnumber[100],badflag,modeflag
    int    filetype,iregister
    bool   l_verbose
    char   in[100],out[100],header,exheader
    char   tmpin,tmpout,tmpon,tmpoff,tmpfile,tmpfinal,tmpwork
    char   tmplist,tmpshift,tmplog
    char   refimage
    real   exptime,norm,ave1,ave2,diff1,airmass
    struct l_struct

    tmpfile=mktemp("tmpin")
    tmpin=mktemp("tmpin")
    tmplist=mktemp("tmplist")
    tmpshift=mktemp("tmpshift")
    tmplog=mktemp("tmpregister")

    l_inputimages=inimages
    l_outputimages=outimages
    l_prefix=outpref
    l_rawpath=rawpath
    l_logfile=logfile
    l_verbose=verbose
    l_stack=stackoption

    status=0

    # Check the rawpath name for a final /
    if (substr(l_rawpath,(strlen(l_rawpath)),(strlen(l_rawpath))) != "/")
	l_rawpath=l_rawpath//"/"
    if (l_rawpath=="/" || l_rawpath==" ")
	l_rawpath=""

    # Set up the logfile
    #
    # First, check for package logfile if the logfile parameter is blank; 
    # if it is empty use tprepare.logfile as the default.
    #
    if ((l_logfile=="") || (l_logfile==" ")) {
	l_logfile=midir.logfile

	if ((l_logfile=="") || (l_logfile==" ")) {
	    l_logfile="midir.log"
	    printlog("Warning - TPREPARE:  Both tprepare.logfile and \
	        midir.logfile are undefined.  Using default file midir.log.",
		logfile=l_logfile, verbose+)
	}
    }

    # Open log file
    date | scan(l_struct)
    printlog("--------------------------------------------------------------\
        --------------",logfile=l_logfile, verbose=l_verbose)
    printlog("Starting TPREPARE -- "//l_struct,
        logfile=l_logfile,verbose=l_verbose)
    printlog("",logfile=l_logfile,verbose=l_verbose)
    printlog("Input T-ReCS image(s)           = "//l_inputimages,
        l_logfile,l_verbose)
    printlog("Path for input raw image(s)     = "//l_rawpath,
        l_logfile,l_verbose)
    printlog("Output image(s)                 = "//l_outputimages,
        l_logfile,l_verbose)
    printlog("Prefix for output file names(s) = "//l_prefix,l_logfile,l_verbose)
    printlog("Stack option                    = "//l_stack,l_logfile,l_verbose)
    printlog(" ",logfile=l_logfile, verbose=l_verbose)

    nimages=0
    maximages=100

    # Count the number of input images
    # First, generate the file list if needed

    if (stridx("*",l_inputimages) > 0) {
	files(l_rawpath//l_inputimages, > tmpfile)
	l_inputimages="@"//tmpfile
    }

    # The following is copied from the "gprepare.cl" script...but it does 
    # not work properly on aguila.  It complains about not being able to 
    # change the type from string.  
    #
    # Parse wildcard, list file, and comma-separated lists of input images
    #
    if (substr(l_inputimages,1,1)=="@")
	scanfile=substr(l_inputimages,2,strlen(l_inputimages))
    else {
	if (stridx(",",l_inputimages)==0) 
	    files(l_inputimages, > tmpfile)
	else {
	    j=9999
	    while (j!=0) {
		j=stridx(",",l_inputimages)
		if (j>0)
		    files(substr(l_inputimages,1,j-1), >> tmpfile)
		else
		    files(l_inputimages, >> tmpfile)
		l_inputimages=substr(l_inputimages,j+1,strlen(l_inputimages))
	    }
	}
	scanfile=tmpfile
    }

    i=0
    while (fscan(scanfile,l_filename) != EOF) {
	i=i+1

	if (substr(l_filename,strlen(l_filename)-4,strlen(l_filename)) == ".fits")
	    l_filename=substr(l_filename,1,strlen(l_filename)-5)

	j=0
	if (stridx("/",l_rawpath) > 0 && stridx("/",l_filename) > 0) {
	    j=stridx("/",l_filename)
	    if (j > 0) {
		for (k=1; k < 100 && j > 0; k+=1) {
		    l_filename=substr(l_filename,j+1,strlen(l_filename))
		    j=stridx("/",l_filename)
		}
	    }
	}

	if (!imaccess(l_rawpath//l_filename))
	    printlog("ERROR - TPREPARE: Input image "//l_rawpath//l_filename//\
	        " was not found.",logfile=l_logfile,verbose+)
	else {
	    nimages=nimages+1
	    if (nimages > maximages) {
		printlog("ERROR - TPREPARE: Maximum number of input \
		    images ["//str(maximages)//"] has been exceeded.",
		    logfile=l_logfile,verbose+)
		status=1
		goto exit
	    }
	    in[nimages]=l_rawpath//l_filename
	    out[nimages]=l_filename
	    j=stridx("/",out[nimages])
	    if (j > 0) {
		for (k=1; k < 100 && j > 0; k+=1) {
		    out[nimages]=substr(out[nimages],j+1,strlen(out[nimages]))
		    j=stridx("/",out[nimages])
		}
	    }
	}
    } #end while-loop

    scanfile=""
    delete(tmpfile//","//tmpin,ver-,>>& "/dev/null")

    if (nimages == 0) {
	printlog("ERROR - TPREPARE: no input images were defined.",
	    logfile=l_logfile,verbose+)
	status=1
	goto exit
    } else 
	printlog("TPREPARE: Processing "//str(nimages)//" image(s).",
	    logfile=l_logfile,verbose=l_verbose)

    # Now, do the same counting for the output file
    tmpfile=mktemp("tmpfile")

    noutputimages=0
    if (l_outputimages != "" && l_outputimages != " ") {
	if (substr(l_outputimages,1,1) == "@")
	    scanfile=substr(l_outputimages,2,strlen(l_outputimages))
	else {
	    if (stridx("*",l_outputimages) > 0) {
		files(l_outputimages,sort-) | \
	            match(".hhd",stop+,print-,metach-, > tmpfile)
		scanfile=tmpfile
	    } else {
		files(l_outputimages,sort-, > tmpfile)
		scanfile=tmpfile
	    }
	}

	while (fscan(scanfile,l_filename) != EOF) {
	    if (substr(l_filename,strlen(l_filename)-4,strlen(l_filename)) == ".fits")
		l_filename=substr(l_filename,1,strlen(l_filename)-5)

	    noutputimages=noutputimages+1
	    if (noutputimages > maximages) {
		printlog("ERROR - TPREPARE: Maximum number of output \
		    images "//str(maximages)//" exceeded.",
		    logfile=l_logfile,verbose+)
		status=1
		goto exit
	    }
	    out[noutputimages]=l_filename
	    if (imaccess(out[noutputimages])) {
		printlog("ERROR - TPREPARE: Output image "//l_filename//\
		    " already exists.",logfile=l_logfile,verbose+)
		status=1
		goto exit
	    }
	}
	if (noutputimages != nimages) {
	    printlog("ERROR - TPREPARE: different number of input \
	        ("//str(nimages)//") and output ("//str(noutputimages)//\
		" image names have been specified.",logfile=l_logfile,verbose+)
	    status=1
	    goto exit
	}

	scanfile=""
	delete(tmpfile,ver-, >>& "/dev/null")
    } else {
	if (l_prefix == "" || l_prefix == " ") 
	    l_prefix="t"
	for (i=1; i <= nimages; i+=1) {
	    out[i]=l_prefix//out[i]
	    if (imaccess(out[i])) {
		printlog("ERROR - TPREPARE: Output image "//out[i]//\
		    " already exists.",logfile=l_logfile,verbose+)
		status=1
		goto exit
	    }
	}
    }

    #
    # The main loop: accumulate source and reference images
    #

    i=1
    while (i <= nimages) {
	printlog(" ",logfile=l_logfile,verbose=l_verbose)
	imgets(in[i]//"[0]","TPREPARE", >>& "dev$null")
	if (imgets.value != "0") {
	    printlog("TPREPARE: file "//in[i]//" has already been prepared.",
	        l_logfile,l_verbose)
	    goto nextimage
	}

	printlog("TPREPARE: Input image "//in[i]//" output image "//out[i]//".",
	    logfile=l_logfile,verbose=l_verbose)

	# check the file structure

	# The following allows the routine to work with older raw data files.  
	# It should give an error if the value is not "complete" and these 
	# should be handled by another routine once I figure out what need to 
	# be done.
	#
	# For handling older files, I will not check this value.
	#

	imgets(in[i]//"[0]","COMPSTAT", >& "dev$null")
	if (imgets.value == "0") {
	    printlog("ERROR - TPREPARE: Image "//in[i]//" does not have the \
	        COMPSTAT header value.",l_logfile,l_verbose)
	    status=1
	    goto nextimage
	} else {
	    if (imgets.value != "COMPLETE") {
		printlog("ERROR - TPREPARE: Image "//in[i]//\
		    " has COMPSTAT = "//imgets.value//".",l_logfile,l_verbose)
		status=1
		goto nextimage
	    }
	}

	tcheckstructure(in[i],logfile=l_logfile,verbose=l_verbose)
	j=tcheckstructure.status
	if (j != 0) {
	    printlog("ERROR - TPREPARE: Image "//in[i]//" does not have the \
	        expected structure.",l_logfile,l_verbose)
	    status=1
	    goto nextimage
	}

	modeflag=tcheckstructure.modeflag
	if (modeflag > 4) {
	    modeflag=modeflag-10
	    filetype=2
	} else
	    filetype=1

	if (filetype != 1) {
	    printlog("ERROR - TPREPARE: Image "//in[i]//" appears to be a \
	        'prepared' T-ReCS file.",l_logfile,l_verbose)
	    status=1
	    goto nextimage
	}

	# check the primary FITS header
	header=in[i]//"[0]"

	if (modeflag == 1 || modeflag == 3) {
	    imgets(header,"NNODS", >& "dev$null")
	    if (imgets.value != "2") {
		printlog("ERROR - TPREPARE: Image "//in[i]//" has "//\
		    imgets.value//" nod phases (must be 2).",
		    logfile=l_logfile,verbose+)
		status=1
		goto nextimage
	    }
	    imgets(header,"NNODSETS", >& "dev$null")
	    i_nodset=int(imgets.value)
	    imgets(header,"SAVESETS", >& "dev$null")
	    i_saveset=int(imgets.value)
	    if (modeflag == 1)
		i_frames=i_saveset*4*i_nodset
	    else
		i_frames=i_saveset*2*i_nodset
	}
	if (modeflag == 4 || modeflag == 2) {
	    imgets(header,"SAVESETS", >& "dev$null")
	    i_saveset=int(imgets.value)
	    if (modeflag == 2)
		i_frames=i_saveset*2
	    else
		i_frames=i_saveset*2
	}

	if (i_frames == 0) {
	    printlog("ERROR - TPREPARE: Either zero nod sets or zero save \
	        sets in image "//in[i]//".",logfile=l_logfile,verbose+)
	    status=1
	    goto nextimage
	}

	imgets(header,"NEXTEND", >& "dev$null")
	if (imgets.value != "0")
	    i_extensions=int(imgets.value)
	else {
	    if (modeflag == 1 || modeflag == 3)
		i_extensions=i_nodset*2
	    else
		i_extensions=1
	}

	for (j=1; j <= i_extensions; j=j+1) {
	    # Create tmp FITS file names used within this loop
	    tmpout=mktemp("tmpout")
	    tmpon=mktemp("tmpon")
	    tmpoff=mktemp("tmpoff")
	    tmpfinal=mktemp("tmpfinal")

	    exheader=in[i]//"["//j//"]"
	    if (modeflag == 1 || modeflag == 3) {
		imgets(exheader,"NOD", >& "dev$null")
		if (imgets.value == "A") {
		    source=1
		    reference=2
		} else {
		    if (imgets.value == "B") {
			source=2
			reference=1
		    } else {
			printlog("ERROR - TPREPARE: Bad nod position \
			    ("//imgets.value//") in image "//in[i]//".",
			    logfile=l_logfile,verbose+)
			status=1
			goto nextimage
		    }
		}
		imgets(exheader,"NODSET", >& "dev$null")
		nodcount=int(imgets.value)
		if (imgets.value == "0") {
		    printlog("ERROR - TPREPARE: Nod number recorded as "//\
		        imgets.value//".",logfile=l_logfile,verbose+)
		    status=1
		    goto nextimage
		} else {
		    i_nodset=int(imgets.value)
		    if (i_nodset < 1) {
			printlog("ERROR - TPREPARE: Nod number recorded as "//\
			    imgets.value//".",logfile=l_logfile,verbose+)
			status=1
			goto nextimage
		    }
		}
	    } else {
		# For chop or stare observations, assume it is in nod 
		# position A unless the header says otherwise.
		#
		source=1
		reference=2
		imgets(exheader,"NOD", >& "dev$null")
		if (imgets.value == "B") {
		    source=2
		    reference=1
		}
	    }
	    #
	    # Check for bad savesets in the nodset...
	    #
	    imgets(exheader,"NBADSET", >& "dev$null")
	    nbadsets=int(imgets.value)
	    if (nbadsets < 0 || nbadsets >= i_saveset) {
		printlog("ERROR - TPREPARE:  header records "//\
		    str(nbadsets)//" (should be less than "//\
		    str(i_saveset)//").",logfile=l_logfile, verbose+)
		goto nextimage
	    }
	    if (nbadsets > 0) {
		for (k=1; k <= nbadsets; k=k+1) {
		    if (k < 10)
			imgets(exheader,"BADSET0"//str(k), >& "dev$null")
		    else
			imgets(exheader,"BADSET"//str(k), >& "dev$null")
		    badsetnumber[k]=int(imgets.value)
		    if (badsetnumber[k] < 1 || badsetnumber[k] > i_saveset) {
			printlog("ERROR - TPREPARE: header records bad save \
			    set "//imgets.value//" (range 1 to "//\
			    str(i_savesets)//").",logfile=l_logfile, verbose+)
			goto nextimage
		    }
		}
	    }

	    iregister=0
	    for (k=1; k <= i_saveset; k=k+1) {
		badflag=0
		for (l=1; l <= nbadsets; l=l+1) {
		    if (badsetnumber[l] == k) {
			badflag=1
			printlog("TPREPARE: header records bad saveset "//\
			    str(k)//" in nod set; will be omitted from the \
			    combined frame.",logfile=l_logfile, verbose+)
		    }
		}
		if (badflag == 0) {
		    print(exheader//"[*,*,"//str(source)//","//str(k)//"]",
		        >> tmpin)
		    print(exheader//"[*,*,"//str(reference)//","//str(k)//"]",
		        >> tmpfile)
		    unlearn("imexpr")
		    if (modeflag != 4) {
		        # Create tmp FITS file names used within this loop
			# DO NOT DELETE tmpwork at the end of this for loop
			# as the image are used throught @tmplist
			tmpwork=mktemp("tmpwork")
			
			if (iregister == 0) {
			    refimage=tmpwork
			    iregister=1
			}
			print(tmpwork, >> tmplist)
			imexpr("(a-b)",tmpwork,exheader//"[*,*,"//\
			    str(source)//","//str(k)//"]",exheader//\
			    "[*,*,"//str(reference)//","//str(k)//"]",verbose-)
			print("tmp"//out[i]//"shift"//str(j)//str(k),
			    >> tmpshift)
		    }
		}
	    }
	    imcombine("@"//tmpin,tmpon,headers="",bpmasks="",rejmasks="",
	        nrejmasks="",expmasks="",sigmas="",logfile="",combine="average",
		reject="none",project=no,outtype="double",outlimits="")
	    if (modeflag != 4) {
		imcombine("@"//tmpfile,tmpoff,headers="",bpmasks="",rejmasks="",
		    nrejmasks="",expmasks="",sigmas="",logfile="",
		    combine="average",reject="none",project=no,outtype="double",
		    outlimits="")
		if (l_stack == "register" && iregister != 0) {
		    xregister("@"//tmplist,refimage,"[*,*]",tmplog,
		        output="@"//tmpshift,interac-,verbose-)
		    imcombine("@"//tmpshift,tmpout,headers="",bpmasks="",
		        rejmasks="",nrejmasks="",expmasks="",sigmas="",
			logfile="",combine="average",reject="sigclip",
			project=no,outtype="double",outlimits="")
		} else {
		    imcombine("@"//tmplist,tmpout,headers="",bpmasks="",
		        rejmasks="",nrejmasks="",expmasks="",sigmas="",
			logfile="",combine="average",reject="none",project=no,
			outtype="double",outlimits="")
		}
	    }
	    count(tmpin) | scan(itotal)
	    delete(tmpin//","//tmpfile//","//tmplist,verify-, >& "dev$null")
	    if (l_stack == "register") {
		delete(tmpshift//","//tmplog,verify-, >& "dev$null")
		delete("tmp*shift*.fits",verify-, >& "dev$null")
	    }
	    delete("tmpshift*",verify-, >& "dev$null")
	    itotal=itotal/2

	    # Write out the combined image(s)
	    if (modeflag == 4)
		imcopy(tmpon,tmpfinal,verb-)
	    else
		imstack(tmpon//","//tmpoff//","//tmpout,tmpfinal)

	    # If j is one, write the primary header to the output file.  
	    # Also write "NEXTEND" if it is not there already...it is useful 
	    # for later tasks.
	    #
	    # Write the image extensions with "fxinsert".
	    #
	    if (j == 1) {
		imcopy(header,out[i])
		# KL shouldn't it be add+, or addonly+???
		hedit(out[i]//".fits","NEXTEND",i_extension,add-,addonly-,
		    delete-,verify-,show-,update+)
	    }
	    l=j-1
	    fxinsert(tmpfinal//".fits",out[i]//".fits["//str(l)//"]","",verb-)

	    if (modeflag == 1 || modeflag == 3) {
		if (source == 1)
		    hedit(out[i]//".fits["//str(j)//"]","NOD","A",add+,addonly-,
		        delete-,verify-,show-,update+)
		else
		    hedit(out[i]//".fits["//str(j)//"]","NOD","B",add+,addonly-,
		        delete-,verify-,show-,update+)
		hedit(out[i]//".fits["//str(j)//"]","NODSET",nodcount,add+,
		    addonly-,delete-,verify-,show-,update+)
	    }
	    hedit(out[i]//".fits["//str(j)//"]","SAVESETS",i_saveset,add+,
	        addonly-,delete-,verify-,show-,update+)
	    imgets(exheader,"UTSTART", >& "dev$null")
	    if (imgets.value != "0")
		hedit(out[i]//".fits["//str(j)//"]","UTSTART",imgets.value,
		    add+,addonly-,delete-,verify-,show-,update+)
	    imgets(exheader,"UTEND", >& "dev$null")
	    if (imgets.value != "0")
		hedit(out[i]//".fits["//str(j)//"]","UTEND",imgets.value,add+,
		    addonly-,delete-,verify-,show-,update+)
	    imgets(exheader,"AMSTART", >& "dev$null")
	    if (imgets.value != "0") {
		airmass=real(imgets.value)
		hedit(out[i]//".fits["//str(j)//"]","AMSTART",airmass,add+,
		    addonly-,delete-,verify-,show-,update+)
	    }

	    # Copy the WCS values from the primary header to the extension 
	    # headers

	    #imgets(header,"CD1_1", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    airmass=real(imgets.value)
	    #    hedit(out[i]//".fits["//str(j)//"]","CD1_1",airmass,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}
	    #imgets(header,"CD1_2", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    airmass=real(imgets.value)
	    #    hedit(out[i]//".fits["//str(j)//"]","CD1_2",airmass,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}
	    #imgets(header,"CD2_1", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    airmass=real(imgets.value)
	    #    hedit(out[i]//".fits["//str(j)//"]","CD2_1",airmass,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}
	    #imgets(header,"CD2_2", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    airmass=real(imgets.value)
	    #    hedit(out[i]//".fits["//str(j)//"]","CD2_2",airmass,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}
	    #imgets(header,"WCSDIM", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    airmass=real(imgets.value)
	    #    hedit(out[i]//".fits["//str(j)//"]","WCSDIM",airmass,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}
	    #imgets(header,"LTM1_1", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    airmass=real(imgets.value)
	    #    hedit(out[i]//".fits["//str(j)//"]","LTM1_1",airmass,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}
	    #imgets(header,"LTM1_2", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    airmass=real(imgets.value)
	    #    hedit(out[i]//".fits["//str(j)//"]","LTM1_2",airmass,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}
	    #imgets(header,"LTM2_1", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    airmass=real(imgets.value)
	    #    hedit(out[i]//".fits["//str(j)//"]","LTM2_1",airmass,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}
	    #imgets(header,"LTM2_2", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    airmass=real(imgets.value)
	    #    hedit(out[i]//".fits["//str(j)//"]","LTM2_2",airmass,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}
	    #imgets(header,"CRPIX1", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    airmass=real(imgets.value)
	    #    hedit(out[i]//".fits["//str(j)//"]","CRPIX1",airmass,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}
	    #imgets(header,"CRPIX2", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    airmass=real(imgets.value)
	    #    hedit(out[i]//".fits["//str(j)//"]","CRPIX2",airmass,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}
	    #imgets(header,"CRVAL1", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    airmass=real(imgets.value)
	    #    hedit(out[i]//".fits["//str(j)//"]","CRVAL1",airmass,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}
	    #imgets(header,"CRVAL2", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    airmass=real(imgets.value)
	    #    hedit(out[i]//".fits["//str(j)//"]","CRVAL2",airmass,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}
	    #imgets(header,"CTYPE1", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    hedit(out[i]//".fits["//str(j)//"]","CTYPE1",imgets.value,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}
	    #imgets(header,"CTYPE2", >& "dev$null")
	    #if (imgets.value != "0") {
	    #    hedit(out[i]//".fits["//str(j)//"]","CTYPE2",imgets.value,add+,
	    #        addonly-,delete-,verify-,show-,update+)
	    #}

	    hedit(out[i]//".fits["//str(j)//"]","AXISLAB4","",add-,addonly-,
	        delete+,update+,verify-,show-)

	    if (modeflag == 2 || modeflag == 4)
		printlog("TPREPARE: Image written to output file extension "//\
		    str(j)//".",logfile=l_logfile,verbose=l_verbose)
	    else {
		if (source == 1)
		    printlog("TPREPARE: Nod A image for set "//\
		        str(nodcount)//" written to output file extension "//\
			str(j)//".",logfile=l_logfile,verbose=l_verbose)
		else
		    printlog("TPREPARE: Nod B image for set "//\
		        str(nodcount)//" written to output file extension "//\
			str(j)//".",logfile=l_logfile,verbose=l_verbose)
	    }
	    
	    # Delete temporary files and images
	    imdelete(tmpon//","//tmpoff//","//tmpout//","//tmpfinal//","//\
	        tmpshift,verify-, >& "dev$null")
	    delete("tmpwork*.fits",verify-, >& "dev$null")
	    
	} #end for-loop on extensions

	# Time stamp the primary header
	#
	date | scan(l_struct)
	gemhedit(out[i]//"[0]","TPREPARE",l_struct,"Time stamp for TPREPARE")

nextimage:
	# jump to here if there is a problem
	i=i+1
    }

exit:
    date | scan(l_struct)
    if (status == 0)
	printlog("TPREPARE: exit status => good.  "//l_struct,
	    logfile=l_logfile, verbose=l_verbose)
    else
	printlog("TPREPARE: exit status => error.  "//l_struct,
	    logfile=l_logfile, verbose=l_verbose)
    printlog("-------------------------------------------------------------\
        ---------------", logfile=l_logfile, verbose=l_verbose)
    printlog(" ",logfile=l_logfile, verbose=l_verbose)

end
