# Copyright(c) 2001-2004 Association of Universities for Research in Astronomy, Inc.

procedure nscombine(inimages)

# Version Sept 20, 2002 JJ v1.4 release
#         Aug 19, 2003  KL IRAF2.12 - new parameter, addonly, in hedit
#                          Fixed even though the hedit calls are commented out.

char   inimages     {prompt="Input images or spectra to shift and combine"}
real   tolerance    {0.5,min=0.,prompt="Maximum offset for grouping the frames (arcsec)"}
char   output       {"default",prompt="File name for output shifted and combined spectra"}
char   output_suffix {"_add",prompt="Suffix for output shifted and combined image"}
char   bpm          {"",prompt="Name of bad pixel mask."}
char   key_pixscale {"PIXSCALE",prompt="Image header keyword of pixel scale [arcsec/pix]"}
real   pixscale     {1.,prompt="Pixel scale in arcsec/pixel"}
char   key_xoff     {"GXOFF",prompt="Header keyword for instrument x offset (in arcsec)"}
char   key_yoff     {"GYOFF",prompt="Header keyword for instrument y offset (in arcsec)"}
bool   fl_keepshift {no,prompt="Keep shifted images?"}             
bool   fl_shiftint  {yes,prompt="Shift frames by integer pixels?"}
char   interptype   {"linear",prompt="Interpolation type for shifting",\
                    enum="nearest|linear|poly3|poly5|spline3|sinc|drizzle"}
char   boundary	    {"nearest",prompt="Boundary type for shifting",\
                    enum="constant|nearest|reflect|wrap"}
real   constant	    {0.,prompt="Constant value for boundary extension when shifting"}
char   combtype     {"average",prompt="Combination operation",\
                    enum="average|median"}
char   rejtype      {"avsigclip",prompt="Rejection algorithm when combining",\
                    enum="none|minmax|ccdclip|crreject|sigclip|avsigclip|pclip"}
char   key_exptime  {"EXPTIME",prompt="Exposure time header keyword"}
char   statsec      {"[*,*]",prompt="Image section to be used for statistics"}
char   scale        {"none",prompt="Image scaling"}
char   zero         {"none",prompt="Image zeropoint offset"}
char   weight       {"none",prompt="Image weights"}
real   lthreshold   {INDEF,prompt="Lower threshold"}
real   hthreshold   {INDEF,prompt="Upper threshold"}
int    nlow         {1,min=0,prompt="minmax: Number of low pixels to reject"}
int    nhigh        {1,min=0,prompt="minmax: Number of high pixels to reject"}
int    nkeep        {0,min=0,prompt="Minimum to keep or maximum to reject"}  
bool   mclip        {yes,prompt="Use median in sigma clipping algorithms?"}
real   lsigma       {3.,prompt="Lower sigma clipping factor"}
real   hsigma       {3.,prompt="Upper sigma clipping factor"}
char   key_ron      {"RDNOISE",prompt="Keyword for readout noise in e-"}
char   key_gain     {"GAIN",prompt="Keyword for gain in electrons/ADU"}
real   ron          {0.0,min=0.,prompt="Readout noise rms in electrons"}
real   gain         {1.0,min=0.00001,prompt="Gain in e-/ADU"}
char   snoise       {"0.0",prompt="ccdclip: Sensitivity noise (electrons)"}
real   sigscale     {0.1,prompt="Tolerance for sigma clipping scaling correction"}
real   pclip        {-0.5,prompt="pclip: Percentile clipping parameter"}
real   grow         {0.0,prompt="Radius (pixels) for neighbor rejection"}
char   sci_ext      {"SCI",prompt="Name or number of science extension"}
char   var_ext      {"VAR",prompt="Name or number of variance extension"}
char   dq_ext       {"DQ",prompt="Name or number of data quality extension"}
bool   fl_vardq     {yes,prompt="Create output variance and data quality frames?"}
char   logfile      {"",prompt="Logfile"}
bool   verbose      {yes,prompt="Verbose output?"}
int    status       {0,prompt="Exit status (0=good)"}
struct *scanfile    {"",prompt="Internal use only"} 

begin

char l_inimages, l_logfile, l_output, l_sci_ext, l_var_ext, l_dq_ext
char l_key_exptime, l_statsec, l_bpm, l_interptype,l_boundary, l_key_pixscale
char l_key_xoff, l_key_yoff
char l_combtype, l_rejtype, l_shift_suffix, l_output_suffix
bool l_fl_vardq, l_verbose, l_fl_keepshift
char l_scale, l_zero, l_weight
char l_key_ron, l_key_gain, l_snoise
char l_sgain, l_sron
real   l_ron, l_gain, l_lthreshold, l_hthreshold, l_lsigma, l_hsigma
real   l_grow, l_sigscale, l_pclip, l_tolerance, l_constant, l_pixscale
int    l_nlow, l_nhigh, l_nkeep, varext, dqext, jj
bool   l_mclip, l_fl_shiftint

struct sdate
char tmpimages, tmpokimages, tmptarget, tmpoffset, tmpshiftcoo,tmpshift, img
char tmpshiftvar, tmpshiftvarout, tmpshiftdq, tmpshiftdqout, dummy, tmphead
int  nbad, nfiles, nn, ntarg, noff, nshift
real xoff, yoff, xrefoff, yrefoff

cache("imgets","keypar")

l_inimages = inimages; l_logfile=logfile; l_output=output
l_combtype=combtype; l_rejtype=rejtype; l_scale=scale; l_zero=zero
l_weight=weight; l_lthreshold=lthreshold; l_hthreshold=hthreshold
l_nlow=nlow; l_nhigh=nhigh; l_mclip=mclip; l_lsigma=lsigma; l_hsigma=hsigma
l_key_ron=key_ron; l_key_gain=key_gain; l_ron=ron; l_gain=gain; l_snoise=snoise
l_sigscale=sigscale; l_pclip=pclip; l_grow=grow; l_bpm=bpm
l_tolerance=tolerance; l_output_suffix=output_suffix
l_sci_ext=sci_ext; l_var_ext=var_ext; l_dq_ext=dq_ext
l_key_exptime=key_exptime; l_statsec=statsec; l_nkeep=nkeep
l_fl_vardq=fl_vardq; l_verbose=verbose; l_fl_shiftint=fl_shiftint
l_interptype=interptype; l_boundary=boundary; l_constant=constant
l_fl_keepshift=fl_keepshift
l_key_pixscale=key_pixscale
l_pixscale=pixscale; l_key_xoff=key_xoff; l_key_yoff=key_yoff

l_shift_suffix="_shift" 

status=0

#----------------------------------------------------------------------------
# Make temp files
tmpimages = mktemp("tmpimages")
tmpokimages = mktemp("tmpokimages")
tmptarget = mktemp("tmptarget")
tmpoffset = mktemp("tmpoffset")
tmpshiftcoo = mktemp("tmpshiftcoo")
tmpshift = mktemp("tmpshift")
tmpshiftvar = mktemp("tmpshiftvar")
tmpshiftvarout = mktemp("tmpshiftvarout")
tmpshiftdq = mktemp("tmpshiftdq")
tmpshiftdqout = mktemp("tmpshiftdqout")
tmphead = mktemp("tmphead")

# Start logging to file
cache("niri")
print(l_logfile) | scan(l_logfile)
if(l_logfile=="") {
  l_logfile=niri.logfile
  print(l_logfile) | scan(l_logfile)
  if(l_logfile=="") {
    l_logfile="niri.log"
    printlog("WARNING - NSCOMBINE: Both nscombine.logfile and niri.logfile \
    	are empty.",l_logfile,verbose+)
    printlog("                    Using default file niri.log.",l_logfile,
    	verbose+)
  }
}
date | scan(sdate)
printlog("----------------------------------------------------------------\
	----------------",l_logfile,verbose=l_verbose)
printlog("NSCOMBINE -- "//sdate,l_logfile,verbose=l_verbose)
printlog(" ",l_logfile,verbose=l_verbose)
printlog("Input images = "//l_inimages,l_logfile,l_verbose)
if(l_verbose) files(l_inimages)
files(l_inimages, >> l_logfile)
printlog("Bad pixel mask = "//l_bpm,l_logfile,l_verbose)
printlog("Offset tolerance = "//l_tolerance,l_logfile,l_verbose)
printlog("Statistics section = "//l_statsec,l_logfile,l_verbose)
printlog("Combine algorithm = "//l_combtype,l_logfile,l_verbose)
printlog("Rejection algorithm = "//l_rejtype,l_logfile,l_verbose)
printlog("Use integer pixel shifts? "//l_fl_shiftint,l_logfile,l_verbose)
printlog("Boundary extension  = "//l_boundary,l_logfile,l_verbose)

#-------------------------------------------------------------------------
# If mef extension names are empty, exit
print(l_sci_ext) | scan(l_sci_ext)
if(l_sci_ext=="") {
  printlog("ERROR - NSCOMBINE: Empty sci_ext extension",l_logfile,verbose+)
  status=1
  goto clean
}
if(l_fl_vardq) {
  print(l_var_ext) | scan(l_var_ext)
  if(l_var_ext=="") {
    printlog("ERROR - NSCOMBINE: Empty variance extension name", l_logfile,
      verbose+)
    status=1
    goto clean
  }
  print(l_dq_ext) | scan(l_dq_ext)
  if(l_dq_ext=="") {
    printlog("ERROR - NSCOMBINE: Empty data quality extension name",l_logfile,
      verbose+)
    status=1
    goto clean
  }
}

# Check if key_exptime is set
print(l_key_exptime) | scan(l_key_exptime)
if(l_key_exptime=="") {
  printlog("WARNING - NSCOMBINE: key_exptime not set, cannot check exposure \
    times.",l_logfile,verbose+)
}

# Check for empty statistics section
print(l_statsec) | scan(l_statsec)
if(l_statsec=="") {
  l_statsec="[*,*]"
  printlog("WARNING - NSCOMBINE: No statsec defined.  Using entire array.",
    l_logfile,verbose+)
}

# Pixel scale
print(l_key_pixscale) | scan(l_key_pixscale)
if(l_key_pixscale=="" && l_pixscale>0) {
  printlog("WARNING - NSCOMBINE: key_pixscale not set, using pixscale \
    value = "//l_pixscale,l_logfile,verbose+)
}

#--------------------------------------------------------------------------------
# Check the input images - put all images in file tmpokimages
files(l_inimages,sort-, > tmpimages)
count(tmpimages) | scan(nfiles)

nbad=0
scanfile=tmpimages
while(fscan(scanfile,img) != EOF) {
  gimverify(img); img = gimverify.outname
  if(gimverify.status!=0) {
    printlog("WARNING - NSCOMBINE: Input image: "//img//".fits does not exist.",
       l_logfile,verbose+)
    nbad+=1
  } else print(img, >> tmpokimages)
					
}
if(nbad==nfiles) {
  printlog("ERROR - NSCOMBINE: No images to combine.",l_logfile,verbose+)
  goto clean
}
delete (tmpimages,ver-)

# If only one file to combine - exit.
count(tmpokimages) | scan(nfiles)
if(nfiles==1) {
  printlog("ERROR - NSCOMBINE: Only one image to combine",l_logfile,verbose+)
  goto clean
}

#--------------------------------------------------------------------------------
# check output image 
print(l_output) | lcase | scan(dummy)
if(dummy=="default" || dummy=="") {
  head(tmpokimages,nlines=1) | scan(img)
  l_output = img//l_output_suffix
} 
gimverify(l_output); l_output=gimverify.outname
if(l_output=="") {
  printlog("ERROR - NSCOMBINE: No output image was specified.",l_logfile,verbose+)
  status=1
  goto clean
} else if (gimverify.status!=1) {
  printlog("ERROR - NSCOMBINE: Output image "//l_output//" already exists", 
    l_logfile,verbose+)
  status=1
  goto clean
}
printlog("output = "//l_output,l_logfile,l_verbose)

#--------------------------------------------------------------------------------

# Write the offsets from the WCS using gemoffsetlist
head(tmpokimages,nlines=1) | scan(img)
imgets(img//"[0]","GXOFF")
xrefoff=real(imgets.value)
imgets(img//"[0]","GYOFF")
yrefoff=real(imgets.value)
gemoffsetlist("@"//tmpokimages,img,l_tolerance,tmptarget,tmpoffset,
   logfile=l_logfile,verbose=l_verbose,offsettype="distance",
   key_xoff=l_key_xoff,key_yoff=l_key_yoff,key_time="", >& "dev$null")

if(gemoffsetlist.status!=0) {
  printlog("ERROR - NSCOMBINE: Error in GEMOFFSETLIST.  Can't determine offsets.",
     l_logfile,verbose+)
  status=1
  goto clean
}

noff=0
if(access(tmpoffset)) count(tmpoffset) | scan(noff)

# Shift images in tmpoffset by the offsets in the headers that
# gemoffsetlist calculated: GXOFF, and GYOFF
if(noff>0) {
  # Put the shifts in a temp file
  scanfile=tmpoffset
  nshift=0
  while(fscan(scanfile,img) != EOF) {

    if(access(img//l_shift_suffix//".fits")) {
      printlog("WARNING - NSCOMBINE: shifted image "//img//l_shift_suffix//".fits \
         already exists.",l_logfile,verbose+)
      printlog("                     Existing image will be used, but offsets may be wrong!",
         l_logfile,verbose+)
    } else {
      print(img//l_shift_suffix//"["//l_sci_ext//"]", >> tmpimages)
      print(img//l_shift_suffix//"["//l_sci_ext//",overwrite]", >> tmpshift)
      copy(img//".fits",img//l_shift_suffix//".fits",verbose=l_verbose)

      imgets(img//"[0]","GXOFF")
      xoff=real(imgets.value)-xrefoff
      imgets(img//"[0]","GYOFF")
      yoff=real(imgets.value)-yrefoff

      keypar(img//"[0]",l_key_pixscale,silen+)
      pixsc = l_pixscale
      if(keypar.found) pixsc = real(keypar.value)
      xoff=xoff/real(keypar.value)
      yoff=yoff/real(keypar.value)
      print(xoff,yoff, >> tmpshiftcoo)
      nshift+=1
    }
    print(img//l_shift_suffix, >> tmptarget)
  }	

  if(noff==nshift) {

    # If user wants integer shifts, change them accordingly
    if(l_fl_shiftint) {
       tcalc(tmpshiftcoo,"c1","int(c1+0.5)",datatype="int",
          colunits="arcsec",colfmt="%d")
       tcalc(tmpshiftcoo,"c2","int(c2+0.5)",datatype="int",
          colunits="arcsec",colfmt="%d")
    }

    # print the offsets in the logfile
    jj=2
    printlog(" ",l_logfile,l_verbose)
    printlog("Offsets used to shift images, in pixels:",l_logfile,l_verbose)
    scanfile=tmpshiftcoo
    while(fscan(scanfile,xoff,yoff) != EOF) {
      printlog("   Image "//jj//"  "//xoff//"  "//yoff,l_logfile,l_verbose)
      jj+=1
    }

    imshift("@"//tmpimages,"@"//tmpshift,shifts_file=tmpshiftcoo,
       interp_type=l_interptype,boundary=l_boundary,constant=l_constant)
			
    if(l_fl_vardq) {
      scanfile=tmpoffset
      while(fscan(scanfile,img) != EOF) {
        varext=0; dqext=0
        fxheader(img//l_shift_suffix,format="",long-,count-) | match(l_var_ext,stop-,
           print+,meta+) | count | scan(varext)			
        if(varext==1) {
          print(img//l_shift_suffix//"["//l_var_ext//"]", >> tmpshiftvar)
          print(img//l_shift_suffix//"["//l_var_ext//",overwrite]", >> tmpshiftvarout)
        }

        fxheader(img//l_shift_suffix,format="",long-,count-) | match(l_dq_ext,stop-,
           print+,meta+) | count | scan(dqext)			
        if(dqext==1) {
          print(img//l_shift_suffix//"["//l_dq_ext//"]",  >> tmpshiftdq)
          print(img//l_shift_suffix//"["//l_dq_ext//",overwrite]", >> tmpshiftdqout)
        }					
      }		

      if(access(tmpshiftvar) && access(tmpshiftvarout))
      imshift("@"//tmpshiftvar,"@"//tmpshiftvarout,shifts_file=tmpshiftcoo,
         interp_type=l_interptype,boundary=l_boundary,constant=l_constant)
      if(access(tmpshiftdq) && access(tmpshiftdqout))
      imshift("@"//tmpshiftdq,"@"//tmpshiftdqout,shifts_file=tmpshiftcoo,
         interp_type=l_interptype,boundary=l_boundary,constant=l_constant)
    }
  } # end if(noff==nshift)	
} else {
  printlog("WARNING - NSCOMBINE: No images to shift.",l_logfile,verbose+)
}

# Combine images in tmptarget and tmpshift
delete(tmpokimages,ver-)

gemcombine("@"//tmptarget,l_output,title="",logfile=l_logfile,combine=l_combtype,
   reject=l_rejtype,scale=l_scale,zero=l_zero,weight=l_weight,statsec=l_statsec,
   expname=l_key_exptime,lthreshold=l_lthreshold,hthreshold=l_hthreshold,
   nlow=l_nlow,nhigh=l_nhigh,nkeep=l_nkeep,mclip=l_mclip,lsigma=l_lsigma,
   hsigma=l_hsigma,key_ron=l_key_ron,key_gain=l_key_gain,ron=l_ron,gain=l_gain,
   snoise=l_snoise,sigscal=l_sigscale,pclip=l_pclip,grow=l_grow,bpm=l_bpm,
   sci_ext=l_sci_ext,var_ext=l_var_ext,dq_ext=l_dq_ext,fl_vardq=l_fl_vardq,
   offsets="none",verbose=l_verbose)

#----------------------------------------------------------------------
# update headers of output image
	
gemhedit(l_output//"["//l_sci_ext//"]","extver",1,"Extension version")
if(l_fl_vardq) {
  gemhedit(l_output//"["//l_var_ext//"]","extver",1,"Extension version")
  gemhedit(l_output//"["//l_dq_ext//"]","extver",1,"Extension version")
}
l_output=l_output//"[0]"
gemhedit(l_output,"GEM-TLM",sdate,"Last modification with GEMINI")

# hedit(l_output,"GXOFF",add-,addonly-,delete+,verify-,show-,update+)
# hedit(l_output,"GYOFF",add-,addonly-,delete+,verify-,show-,update+)
# hedit(l_output,"GOFFREF",add-,addonly-,delete+,verify-,show-,update+)

printf("%-8s= \'%-18s\' / %-s\n","NSCOMBIN",sdate,"Time stamp for NSCOMBINE", >> tmphead)
printf("%-8s= %20.3f / %-s\n","NSCOMTOL",l_tolerance,"Spatial tolerance for NSCOMBINE", >> tmphead)
printf("%-8s= \'%-18s\' / %-s\n","NSCOMSTA",l_statsec,"Statistics section used by NSCOMBINE", >> tmphead)
printf("%-8s= \'%-18b\' / %-s\n","NSCOMSFT",l_fl_shiftint,"Use integer pixel shifts?", >> tmphead)
if(!l_fl_shiftint) {
  printf("%-8s= \'%-18s\' / %-s\n","NSCOMINT",l_interptype,"Pixel interpolation method", >> tmphead)
}
printf("%-8s= \'%-18s\' / %-s\n","NSCOMBND",l_boundary,"Boundary type for shifting", >> tmphead)
if(l_boundary=="constant") {
  printf("%-8s= %20.3f / %-s\n","NSCOMCON",l_constant,"Constant for boundary extension for shifting", >> tmphead)
}
printf("%-8s= \'%-18s\' / %-s\n","NSCOMCOM",l_combtype,"Combine method used by NSCOMBINE", >> tmphead)
if(l_zero!="none") {
  printf("%-8s= \'%-18s\' / %-s\n","NSCOMZER",l_zero,"Statistic used to compute additive offsets", >> tmphead)
}
if(l_scale!="none") {
  printf("%-8s= \'%-18s\' / %-s\n","NSCOMSCL",l_scale,"Statistic used to compute scale factors", >> tmphead)
}
if(l_weight!="none") {
  printf("%-8s= \'%-18s\' / %-s\n","NSCOMWEI",l_weight,"Statistic used to compute relative weights", >> tmphead)
}
if(l_rejtype!="none") {
  printf("%-8s= \'%-18s\' / %-s\n","NSCOMREJ",l_rejtype,"Rejection algorithm used by NSCOMBINE", >> tmphead)
  printf("%-8s= %20.5f / %-s\n","NSCOMLTH",l_lthreshold,"Lower threshold before combining", >> tmphead)
  printf("%-8s= %20.5f / %-s\n","NSCOMHTH",l_hthreshold,"Upper threshold before combining", >> tmphead)
  printf("%-8s= %20.2f / %-s\n","NSCOMGRW",l_grow,"Radius for additional pixel rejection", >> tmphead)
}
if(l_rejtype=="minmax") {
  printf("%-8s= %20.0f / %-s\n","NSCOMNLO",l_nlow,"Low pixels rejected (minmax)", >> tmphead)
  printf("%-8s= %20.0f / %-s\n","NSCOMNHI",l_nhigh,"High pixels rejected (minmax)", >> tmphead)
}
if(l_rejtype=="sigclip" || l_rejtype=="avsigclip" || l_rejtype=="ccdclip" \
    || l_rejtype=="crreject" || l_rejtype=="pclip") {
  printf("%-8s= %20.2f / %-s\n","NSCOMLSI",l_lsig,"Lower sigma for rejection", >> tmphead)
  printf("%-8s= %20.2f / %-s\n","NSCOMHSI",l_hsig,"Upper sigma for rejection", >> tmphead)
  printf("%-8s= %20.0f / %-s\n","NSCOMNKE",l_nkeep,"Min(max) number of pixels to keep(reject)", >> tmphead)
  printf("%-8s= \'%-18b\' / %-s\n","NSCOMMCL",l_mclip,"Use median in clipping algorithms", >> tmphead)
} 
if(l_rejtype=="sigclip" || l_rejtype=="avsigclip" || l_rejtype=="ccdclip" \
    || l_rejtype=="crreject") {
  printf("%-8s= %20.2f / %-s\n","NSCOMSSC",l_sigscale,"Tolerance for sigma clip scaling correction", >> tmphead)
}
if(l_rejtype=="pclip") {
  printf("%-8s= %20.2f / %-s\n","NSCOMPCL",l_pclip,"Percentile clipping factor used by pclip", >> tmphead)
}     
if(l_rejtype=="ccdclip") {
  printf("%-8s= \'%-18s\' / %-s\n","NSCOMSNO",l_snoise,"Sensitivity noise (e) used by ccdclip", >> tmphead)
}     
 
jj=1
scanfile=tmptarget
while(fscan(scanfile,img) != EOF) {
  if(substr(img,strlen(img)-5,strlen(img))=="_shift") 
    img=substr(img,1,strlen(img)-6)
  printf("%-8s= \'%-18s\' / %-s\n","NSCOM000"+jj,img,"Image "//jj//" combined with NSCOMBINE", >> tmphead)
  jj+=1
}
scanfile=""
jj=2
if(access(tmpshiftcoo)) {
scanfile=tmpshiftcoo
while(fscan(scanfile,xoff,yoff) != EOF) {
  printf("%-8s= %20.2f / %-s\n","NSCOMX00"+jj,xoff,"x shift (pix) for input image "//jj, >> tmphead)
  printf("%-8s= %20.2f / %-s\n","NSCOMY00"+jj,yoff,"y shift (pix) for input iamge "//jj, >> tmphead)
  jj+=1
}
scanfile=""
}
      
# put all the new stuff in the header
mkheader(l_output,tmphead,append+,verbose-)


#----------------------------------------------------------------------
clean:
if(!l_fl_keepshift && access(tmpshift)) {
  scanfile=tmpshift
  while(fscan(scanfile,img)!=EOF) {
    img=substr(img,1,strlen(img)-15)
    imdelete(img,verify-, >& "dev$null")
  }
  scanfile=""
}
delete(tmpimages//","//tmpokimages//","//tmptarget//","//tmpoffset,verify-, >>& "dev$null")
delete(tmpshift//","//tmpshiftcoo,verify-, >>& "dev$null")
delete(tmpshiftvar//","//tmpshiftvarout//","//tmpshiftdq//","//tmpshiftdqout,verify-, >>& "dev$null")
delete(tmphead,verify-, >>& "dev$null")

if(status==0) 
  printlog("NSCOMBINE: Exit status good.",l_logfile,l_verbose)
else 
  printlog("NSCOMBINE: Exit status error.",l_logfile,l_verbose)
printlog("--------------------------------------------------------------------------------",
   l_logfile,l_verbose)	
      
end
