# Copyright(c) 2001-2004 Association of Universities for Research in Astronomy, Inc.

procedure nscut(inimages)

# Version Sept 20, 2002 JJ v1.4 release
#         Aug 19, 2003  KL IRAF2.12 - new parameter, addonly, in hedit

char   inimages     {prompt="Input files"}
char   outspectra   {"",prompt="Output spectra"}
char   outprefix    {"default",prompt="Prefix for output spectra"}
char   key_section  {"default",prompt="Header keyword containing section to chop"}
char   section      {"",prompt="Section to chop if not in header"}
char   sci_ext      {"SCI",prompt="(Root) name or number of science extension"}
char   var_ext      {"VAR",prompt="(Root) name or number of variance extension"}
char   dq_ext       {"DQ",prompt="(Root) name or number of data quality extension"}
char   logfile      {"",prompt="Logfile"}
bool   fl_nsappwave {yes,prompt="Put approximate wavelength solution in header?"}
char   nsappwavedb  {"niri$nsappwave.dat", prompt="NSAPPWAVE Database"}
real   crval        {INDEF,prompt="Central wavelength"}
real   cdelt        {INDEF,prompt="Resolution in wavelength per pixel"}
char   key_filter   {"FILTER",prompt="Keyword for filter"}
char   key_fpmask   {"FPMASK",prompt="Keyword for filter"}
char   key_camera   {"CAMERA",prompt="Keyword for filter"}
char   key_dispaxis {"DISPAXIS",prompt="Header keyword for dispersion axis"}
char   key_xoff     {"XOFFSET",prompt="Header keyword for instrument x offset (in arcsec)"}
char   key_yoff     {"YOFFSET",prompt="Header keyword for instrument y offset (in arcsec)"}
bool   verbose      {yes,prompt="Verbose output?"}
int    status       {0,prompt="Exit status (0=good)"}
struct *scanfile    {"",prompt="Internal use only"} 

begin

    char   l_inimages,l_logfile,l_key_section,l_section,l_outspectra,l_outprefix
    char   l_sci_ext, l_var_ext, l_dq_ext, l_key_xoff, l_key_yoff
    bool   l_verbose, l_fl_nsappwave
    char   l_nsappwavedb, l_key_filter,l_key_fpmask,l_key_camera,l_key_dispaxis
    real   l_crval, l_cdelt
    char   tmpinimages, tmpokin, tmpoutfiles, tmpokout, tmpok, tmpout
    char   img, outimg, sec, subsec, dummy, tmptarg, tmpoffset 
    int    nx, ny, sx, sy, nbad, nin, nout, numout
    struct sdate

    cache("imgets")

    l_inimages=inimages; l_logfile=logfile; l_verbose=verbose
    l_key_section=key_section; l_section=section
    l_sci_ext=sci_ext;l_var_ext=var_ext;l_dq_ext=dq_ext
    l_key_xoff=key_xoff; l_key_yoff=key_yoff
    l_outspectra=outspectra; l_outprefix=outprefix
    l_fl_nsappwave=fl_nsappwave; l_crval=crval; l_cdelt=cdelt
    l_key_filter=key_filter; l_key_fpmask=key_fpmask
    l_key_camera=key_camera; l_key_dispaxis=key_dispaxis
    l_nsappwavedb=nsappwavedb

    dummy=""

    # Start logging to file
    cache("niri")
    print(l_logfile) | scan(l_logfile)
    if (l_logfile=="") {
        l_logfile=niri.logfile
        print(l_logfile) | scan(l_logfile)
        if (l_logfile=="") {
	    l_logfile="niri.log"
	    printlog("WARNING - NSCUT: Both nscut.logfile and niri.logfile \
    	        are empty.",l_logfile,verbose+)
	    printlog("                    Using default file niri.log.",
    	        l_logfile,verbose+)
        }
    }
    date | scan(sdate)
    printlog("------------------------------------------------------------------------------",
        l_logfile,verbose=l_verbose)
    printlog("NSCUT -- "//sdate,l_logfile,verbose=l_verbose)
    printlog(" ",l_logfile,verbose=l_verbose)

    # Make temp files:
    tmpinimages = mktemp("tmpinimages")
    tmpoutfiles = mktemp("tmpoutfiles")
    tmpokin = mktemp("tmpokin")
    tmpokout = mktemp("tmpokout")
    tmpok = mktemp("tmpok")

    #----------------------------------------------------------------------
    # Check that input files exist
    status = 0
    if (substr(l_inimages,1,1) =="@" && !access(substr(l_inimages,2,strlen(l_inimages)))) {
        printlog("ERROR - NSCUT: input file \
	    ("//substr(l_inimages,2,strlen(l_inimages))//") does not exist.",
	    l_logfile,verb+)
        status = 1
        goto clean
    }	

    files(l_inimages,sort-, > tmpinimages)
    scanfile = tmpinimages
    while (fscan(scanfile,img) != EOF) {
	gimverify(img); img = gimverify.outname
	if (gimverify.status!=0) {
	    printlog("WARNING - NSCUT: input image ("//img//") does not exist",
		l_logfile,verbose+)
        } else {
	    keypar(img//"[0]","NSCUT",silent+)
	    if (!keypar.found) 
		print(img, >> tmpokin)
	    else
	        printlog("WARNING - NSCUT: File "//img//" already run through NSCUT.",
                    l_logfile,verbose+)
        }				
    } 

    if (!access(tmpokin)) {
	printlog("ERROR - NSCUT: no input files to process.",l_logfile,verb+)
	status = 1
	goto clean
    }

    # Check if l_key_section is set to default in which case it will be 
    # SPECSEC1
    print(l_key_section) | lcase | scan(dummy)
    if (dummy=="default") l_key_section="SPECSEC1"

    # Check if l_outprefix is set to default in which case it will be s1, s2, 
    # or s3 depending on the value of l_key_section
    dummy=""
    print(l_outprefix) | lcase | scan(dummy)
    if (dummy=="default") {
	if (l_key_section=="SPECSEC1") l_outprefix="s1"
	else if (l_key_section=="SPECSEC2") l_outprefix="s2"
	else if (l_key_section=="SPECSEC3") l_outprefix="s3"
	else l_outprefix="s"
    }	

    #----------------------------------------------------------------------
    # Get the section from the header and if not there, from the parameter 
    # section
    print(l_key_section) | scan(l_key_section)
    print(l_section) | scan(l_section)
    sec=l_section
    delete(tmpinimages,ver-, >& "dev$null")
    if (l_key_section=="" && l_section!="") {
	rename(tmpokin,tmpinimages,field="all")
    } else if (l_key_section!="" && l_section!="") {
	scanfile = tmpokin
	while (fscan(scanfile,img) != EOF) {
	    keypar(img//"[0]", l_key_section, silent+) 
	    if (keypar.found)
	        print(img//" "//keypar.value, >> tmpinimages)
	    else {
		print(img//" "//l_section, >> tmpinimages)
		printlog("WARNING - NSCUT: section for "//img//" = "//l_section//".",
		    l_logfile,verbose+)
	    }			
	}
    } else if (l_key_section!="" && l_section=="") {
	scanfile = tmpokin
	while (fscan(scanfile,img) != EOF) {
	    keypar(img//"[0]", l_key_section, silent+) 
	    if (keypar.found)
	        print(img//" "//keypar.value, >> tmpinimages)
	    else {
		printlog("WARNING - NSCUT: can't read "//l_key_section//" to determine",
		    l_logfile,verbose+)
		printlog("                 section to chop in "//img//".",
		    l_logfile,verbose+)
	    }
	}
    } else if (l_section=="") {
	printlog("ERROR - NSCUT: Unknown section.",l_logfile,verbose+)
	status = 1
	goto clean
    }
    if (!access(tmpinimages)) {
	printlog("ERROR - NSCUT: No images will be chopped.",l_logfile,verbose+)
	status = 1
	goto clean
    }

    # Check that section is not bigger than the image itself
    scanfile = tmpinimages
    delete(tmpokin,ver-, >& "dev$null")
    while (fscan(scanfile,img,sec) != EOF) {
	imgets(img//"["//l_sci_ext//"]","i_naxis1")
	nx = int(imgets.value)
	imgets(img//"["//l_sci_ext//"]","i_naxis2")
	ny = int(imgets.value)
	sx = int(substr(sec,stridx(":",sec)+1,stridx(",",sec)-1))
	subsec = substr(sec,stridx(",",sec)+1,strlen(sec)-1)
	sy = int(substr(subsec,stridx(":",subsec)+1,strlen(subsec)))
	if (nx<sx || ny<sy)
	    printlog("WARNING - NSCUT: section of "//img//" is larger than image.",
		l_logfile,verbose+)
	else
	    print(img//" "//sec, >> tmpokin)
    }
    if (!access(tmpokin)) {
	printlog("ERROR - NSCUT: no images to cut.",l_logfile,verbose+)
	status = 1
	goto clean
    }

    printlog("Using input files:",l_logfile,l_verbose)
    if (l_verbose) type(tmpokin)
    type(tmpokin, >> l_logfile)

    #----------------------------------------------------------------------
    # Add XOFFSET and YOFFSET to header, if not already present

    if (l_fl_nsappwave) {
	head(tmpokin,nlines=1) | scan(img)
	keypar(img//"[0]",l_key_xoff,silent+)
	if (!keypar.found) {
	    tmptarg=mktemp("tmptarg")
	    tmpoffset=mktemp("tmpoffset")
	    gemoffsetlist("@"//tmpokin,img,9999999.,tmptarg,tmpoffset,
		offsettype="distance",key_xoff=l_key_xoff,key_yoff=l_key_yoff,
		key_time="",logfile=l_logfile,verbose-, >& "dev$null") 
	    if (gemoffsetlist.status!=0) {
	        printlog("WARNING - NSCUT: Problem determining offsets with GEMOFFSETLIST.",
		    l_logfile,verbose+)
	        printlog("                 Headers will not have x and y offsets.",
		    l_logfile,verbose+)
	    }
	    delete(tmptarg//","//tmpoffset,ver-, >& "dev$null")
        }
    }
    #----------------------------------------------------------------------
    # Check the output images
    # outspectra could contain legal * if it is of a form like %st%stX%*.fits
    if (substr(l_outspectra,1,1)=="@") {
        if(!access(substr(l_outspectra,2,strlen(l_outspectra)))) {
	   printlog("ERROR - NSCUT:  Input file "//\
	       substr(l_outspectra,2,strlen(l_outspectra))//" not found.",
	       l_logfile,verbose+)
	   status=1
	   goto clean
        }
    }
    files(l_outspectra,sort-, > tmpoutfiles)

    # If tmpoutfiles is empty, the output files names should be created with 
    # the prefix parameter
    numout = 0
    count(tmpoutfiles) | scan(numout)
    if (numout==0) {
	scanfile=tmpokin
	while (fscan(scanfile,img) != EOF) {
	    print(l_outprefix//img, >> tmpoutfiles)
	}
    } 

    # Verify that the output images do not exist
    nbad=0
    scanfile = tmpoutfiles
    while (fscan(scanfile,img) != EOF) {
	gimverify(img); img = gimverify.outname
	if (gimverify.status!=1) {
	    printlog("ERROR - NSCUT: File "//img//" already exists.",
	        l_logfile,verbose+)
	    nbad+=1
	} 
	print(img, >> tmpokout)
    } # end while loop

    # Exit if problems found with output files
    if (nbad > 0) {
	status=1
	goto clean
    } # end if (nbad > 0)

    printlog(" ",l_logfile,l_verbose)
    printlog("Using output files:",l_logfile,l_verbose)
    if (l_verbose) type(tmpokout)
    type(tmpokout, >> l_logfile)

    # Check that output and input files have the same number of entries.
    count(tmpokin) | scan(nin)
    count(tmpokout) | scan(nout)
    if (nin!=nout) {
	printlog("ERROR - NSCUT: Number of input and output images differ.",
	    l_logfile,verbose+)
	status = 1
	goto clean
    } 

    #----------------------------------------------------------------------
    # Make the output file, call nsappwave if requested

    joinlines(tmpokin,tmpokout,output=tmpok,delim=" ",missing="INDEF",
	maxchars=200,short+,verbose-) 

    if (!l_fl_nsappwave) {
	scanfile = tmpok
	while (fscan(scanfile,img,sec,outimg) != EOF) {
	    imcopy(img//".fits[0]",outimg//".fits",verbose-)
	    imcopy(img//"["//l_sci_ext//"]"//sec,
	        outimg//"["//l_sci_ext//",append]",verbose-)
	    if (imaccess(img//"["//l_var_ext//"]"))
		imcopy(img//"["//l_var_ext//"]"//sec,
		    outimg//"["//l_var_ext//",append]",verbose-)
	    if (imaccess(img//"["//l_dq_ext//"]"))
		imcopy(img//"["//l_dq_ext//"]"//sec,
		    outimg//"["//l_dq_ext//",append]",verbose-)
	}
    } else {
	scanfile = tmpok
	while (fscan(scanfile,img,sec,outimg) != EOF) {
	    # Create tmp FITS file name used within this loop
	    tmpout = mktemp("tmpout")

	    # copy sec to tempfile and add header keywords needed by nsappwave
	    imcopy(img//".fits[0]",tmpout,verbose-)
	    imcopy(img//"["//l_sci_ext//"]"//sec,
	        tmpout//"["//l_sci_ext//",append]",verbose-)
	    if (imaccess(img//"["//l_var_ext//"]"))
		imcopy(img//"["//l_var_ext//"]"//sec,
		    tmpout//"["//l_var_ext//",append]",verbose-)
	    if (imaccess(img//"["//l_dq_ext//"]"))
		imcopy(img//"["//l_dq_ext//"]"//sec,
		    tmpout//"["//l_dq_ext//",append]",verbose-)
	    gemhedit(tmpout//"[0]","NSCUT",sdate,"Time stamp for NSCUT")
	    #sometimes NSCUT doesn't appear...
	    keypar(tmpout//"[0]","NSCUT",silent+)
	    if (!keypar.found)
	        gemhedit(tmpout//"[0]","NSCUT",sdate,"Time stamp for NSCUT")
	    if (l_key_section!="")
	        gemhedit(tmpout//"[0]","NSCUTSPC",l_key_section,
		    "SPECSEC extracted by NSCUT")
	    gemhedit(tmpout//"[0]","NSCUTSEC",sec,"Region extracted by NSCUT")

	    # call nsappwave
	    keypar(tmpout//"[0]",l_key_filter,silent+)
	    if (keypar.value!="blank") {
	        nsappwave(tmpout,outspectra=outimg,prefix="",
		    nsappwavedb=l_nsappwavedb,crval=l_crval,cdelt=l_cdelt,
		    key_filter=l_key_filter,key_fpmask=l_key_fpmask,
		    key_camera=l_key_camera,key_dispaxis=l_key_dispaxis,
		    sci_ext=l_sci_ext,var_ext=l_var_ext,dq_ext=l_dq_ext,
		    key_xoff=l_key_xoff,key_yoff=l_key_yoff,logfile=l_logfile,
		    verbose=l_verbose)
	        if (nsappwave.status!=0) {
                    printlog("WARNING - NSCUT: Error in NSAPPWAVE.  No approximate wavelength",
                        l_logfile,verbose+)
                    printlog("                 information written to header.",
		        l_logfile,verbose+)
                    imrename(tmpout,outimg)
	        }
	    } else
	        imrename(tmpout,outimg)
	    imdelete(tmpout,ver-, >& "dev$null")
        }
    }

    #----------------------------------------------------------------------
    # fix up the headers:
    scanfile=tmpok
    while (fscan(scanfile,img,sec,outimg) != EOF) {		
        imgets(img//"["//l_sci_ext//"]","i_naxis1")
        gemhedit(outimg//"[0]","ORIGXSIZ",int(imgets.value),
	    "Original size in X")
        imgets(img//"["//l_sci_ext//"]","i_naxis2")
        gemhedit(outimg//"[0]","ORIGYSIZ",int(imgets.value),
	    "Original size in Y")
        gemhedit(outimg//"[0]","NSCUT",sdate,"Time stamp for NSCUT")
        #sometimes NSCUT doesn't appear...
        keypar(outimg//"[0]","NSCUT",silent+)
        if (!keypar.found)
	    gemhedit(outimg//"[0]","NSCUT",sdate,"Time stamp for NSCUT")
        if (l_key_section!="")
	    gemhedit(outimg//"[0]","NSCUTSPC",l_key_section,
	        "SPECSEC extracted by NSCUT")
        gemhedit(outimg//"[0]","NSCUTSEC",sec,"Region extracted by NSCUT")
        hedit(outimg//"[0]","GEM-TLM",sdate,add+,addonly-,delete-,verify-,
	    show-,update+)
    }

    #----------------------------------------------------------------------

clean:
    scanfile = ""
    delete(tmpokin//","//tmpokout//","//tmpok,verify-, >>& "dev$null")
    delete(tmpinimages//","//tmpoutfiles, verify-, >>& "dev$null")
    if (status==0) {
	printlog(" ",l_logfile,l_verbose)
	printlog("NSCUT exit status:  good.",l_logfile,l_verbose)
    }
    printlog("----------------------------------------------------------------------------",
	l_logfile,l_verbose)

end
