# Copyright(c) 2001-2004 Association of Universities for Research in Astronomy, Inc.

procedure nsextract (inimages)

# Script to extact NIRI spectra
# 
# Version  Sept 20, 2002 MT,JJ v1.4 release
#          Aug 19,  2003 KL  IRAF2.12 new/modified parameters
#                              hedit: addonly
#                              imcombine: headers,bpmasks,expmasks,outlimits
#                                  rejmask->rejmasks, plfile->nrejmasks

char    inimages     {"",prompt="Input images"}
char    outspectra   {"",prompt="Output images"}
char    outprefix    {"x",prompt="Prefix for output spectra"}
char    logfile      {"",prompt="Logfile name"}
char    key_dispaxis {"DISPAXIS",prompt="Header keywordd for dispersion axis"}
int     dispaxis     {1,min=1,max=2,prompt="Dispersion axis"}
char    database     {"database",prompt="Directory for calibration files"}
int     line         {500,prompt="Disperson line"}
int     nsum         {10,prompt="Number of dispersion lines to sum or median"}
char    background   {"none",prompt="Background to subtract",\
                     enum="none|average|fit"}
bool    fl_inter     {no,prompt="Fit interactively?"}
bool    fl_apall     {no,prompt="Use apall for extraction?"}
bool    fl_project   {no,prompt="Project spectra to 1D if not using apall?"}
bool    fl_findneg   {no,prompt="Find and extract a negative spectrum?"}
real    upper        {10,prompt="Upper aperture limit relative to center"}
real    lower        {-10,prompt="Lower aperture limit relative to center"}
char    sci_ext      {"SCI",prompt="Name or number of science extension"}
char    var_ext      {"VAR",prompt="Name or number of variance extension"}
char    dq_ext       {"DQ",prompt="Name or number of data quality extension"}
bool    verbose      {yes,prompt="Verbose output?"}
int     status       {0,prompt="Exit status (0=good)"}
struct* scanfile     {"",prompt="For internal use only"}


begin

    char   l_inimages, l_outspectra, l_prefix, l_logfile, l_background
    char   l_sci_ext, l_var_ext, l_dq_ext, l_database, l_key_dispaxis
    bool   l_fl_inter, l_fl_findneg, l_fl_apall, l_fl_project, l_verbose
    int    l_line,l_nsum, l_dispaxis
    real   l_lower,l_upper

    char   img, speca, specb, imlog, dqspec, imgin, imgout
    char   tsci, tvar, tdq, tmplog, sdum
    char   scispec,varspec,varb, dqb
    char   l_temp, l_logfilesave, l_databasesave
    bool   l_verbosesave
    int    l_dispaxissave
    char   tmpfile1, tmpinlist, tmpoutlist
    char   sec, imcombtdq
    int    nin, nout, nbad, iy1, iy2, varpl
    int    count, nfiles, oldcount, lcount
    real   ycp, ycn, y1, y2, npixr
    bool   fl_runbefore
    struct sdate, expr

    # Set local variables
    l_inimages=inimages ; l_outspectra=outspectra ; l_logfile=logfile
    l_line=line ; l_nsum=nsum ; l_background=background 
    l_sci_ext = sci_ext ; l_dq_ext = dq_ext ; l_var_ext=var_ext
    l_verbose=verbose ; l_fl_inter=fl_inter ; l_fl_findneg=fl_findneg
    l_fl_apall=fl_apall ; l_fl_project=fl_project 
    l_prefix = outprefix ; l_database=database ; l_dispaxis=dispaxis
    l_key_dispaxis=key_dispaxis

    l_logfilesave=specred.logfile
    l_databasesave=specred.database ; specred.database=l_database
    l_dispaxissave=specred.dispaxis
    l_verbosesave=specred.verbose ; specred.verbose=l_verbose

    if (!l_fl_apall) {
	l_lower=lower
	l_upper=upper
    }     
    varpl=4
    if (l_background=="none") 
        varpl=3

    # Initialize variables
    status=0

    cache("imgets","hedit","niri","specred")

    # logfile name
    print(l_logfile) | scan(l_logfile)
    if (l_logfile=="") {   
        l_logfile=niri.logfile
        if (l_logfile=="") {  
	    l_logfile="niri.log"
	    printlog("WARNING - NSEXTRACT:  Both nsextract.logfile and \
	        niri.logfile are",l_logfile,verbose+) 
	    printlog("                      empty.  Using niri.log.",
	        l_logfile,verbose+)
        }
    }
    specred.logfile=l_logfile

    # Make temporary files
    tmpfile1 = mktemp("tmpfile1")
    tmpinlist = mktemp("tmpinl")
    tmpoutlist = mktemp("tmpout")
    speca=mktemp("tmpspeca")
    specb=mktemp("tmpspecb")
    tmplog=mktemp("tmplog")
    tdq=mktemp("tmpdq")
    imcombtdq=mktemp("tmpimcombdq")
    
    # Assign dummy name to tmp file name variables
    # (Just needs to be given some value. The real name is created later.)
    tsci = "dummy"

    # Write to the log
    date | scan(sdate)
    printlog("-----------------------------------------------------------\
        ---------------------",l_logfile,verbose=l_verbose)
    printlog("NSEXTRACT -- "//sdate,l_logfile,verbose=l_verbose)
    printlog(" ",l_logfile,verbose=l_verbose)

    #----------------------------------------------------------------------
    # Make lists of input spectra
    # Put all input images in a temporary file: tmpfile1

    if (substr(l_inimages,1,1)=="@") {
        l_temp=substr(l_inimages,2,strlen(l_inimages))
        if (!access(l_temp)) {
	    printlog("ERROR - NSEXTRACT:  Input file "//l_temp//" not found.",
	        l_logfile,verbose+)
	    status=1
	    goto clean
        }
    }

    files(l_inimages,sort-, > tmpfile1)

    nbad=0; nfiles=0
    scanfile = tmpfile1
    while (fscan(scanfile,img) != EOF) {
	fl_runbefore=no
	gimverify(img); img=gimverify.outname
	if (gimverify.status==1) {
	    printlog("ERROR - NSEXTRACT: File "//img//" not found.",
	        l_logfile,verbose+)
	    nbad+=1
	} else if (gimverify.status>1) {
	    printlog("ERROR - NSEXTRACT: File "//img//" not a MEF FITS image.",
		l_logfile,verbose+)
	    nbad+=1
	} else {
	    imgets(img//"[0]","NPREPARE", >& "dev$null")
	    if (imgets.value == "0") {
		printlog("ERROR - NSEXTRACT: Image "//img//" not NPREPAREd.",
		    l_logfile,verbose+)
		nbad+=1
	    }

	    imgets(img//"[0]","NSEXTRAC", >& "dev$null")
	    if (imgets.value != "0") {
		printlog("WARNING - NSEXTRACT: Image "//img//" has been run \
		    through nsextract before.",l_logfile,verbose+)
		fl_runbefore=yes

	    }
	}

	if (!fl_runbefore) {
	    nfiles+=1
	    print(img, >> tmpinlist)
	}      
    } # end while loop

    # Check for empty file list
    if (nfiles==0) {
	printlog("ERROR - NSEXTRACT:  No input images to process.",
	    l_logfile,verbose+)
	status=1
	goto clean
    }

    # Exit if problems found with input files
    if (nbad > 0) {
	printlog("ERROR - NSEXTRACT: "//nbad//" image(s) either do not exist, \
	    are not MEF files, or",l_logfile, verbose+)
	printlog("                 have not been run through NPREPARE.",
	    l_logfile,verbose+)
	status=1
	goto clean
    } # end if (nbad > 0)

    printlog("Using input files:",l_logfile,l_verbose)
    if (l_verbose) type(tmpinlist)
    type(tmpinlist, >> l_logfile)
    delete(tmpfile1, ver-)

    #----------------------------------------------------------------------
    # Check output images
    if (substr(l_outspectra,1,1)=="@") {
	l_temp=substr(l_outspectra,2,strlen(l_outspectra))
	if (!access(l_temp)) {
	    printlog("ERROR - NSEXTRACT: list of output file \
	        names "//l_temp//" not found.",l_logfile,verbose+)
	    status=1
	    goto clean
	}
    }
    files(l_outspectra,sort-, > tmpfile1)

    # If tmpfile1 is empty, the output files names should be created with the
    # prefix parameter
    count=0; count(tmpfile1) | scan(count)
    if (count==0) {
	delete(tmpfile1,ver-)
	scanfile=tmpinlist
	while (fscan(scanfile,img) != EOF) {
	    print(l_prefix//img, >> tmpfile1)
	}
    }               

    # Verify that the output images do not exist
    nbad=0
    nfiles=0
    scanfile = tmpfile1
    while (fscan(scanfile,img) != EOF) {
	gimverify(img)
	img = gimverify.outname
	if (gimverify.status!=1) {
	    printlog("ERROR - NSEXTRACT: File "//img//" already exists.",
	        l_logfile,verbose+)
	    nbad+=1
	} 
	nfiles+=1
	print(img, >> tmpoutlist)
    } # end while loop

    # Exit if problems found with output files
    if (nbad > 0) {
	status=1
	goto clean
    } # end if (nbad > 0)

    printlog("Using output files:",l_logfile,l_verbose)
    if (l_verbose) type(tmpoutlist)
    type(tmpoutlist, >> l_logfile)
    delete(tmpfile1, ver-)

    # Check number of input output images
    count(tmpinlist) | scan(nin)
    count(tmpoutlist) | scan(nout)
    if (nin!=nout) {
	printlog("ERROR - NSEXTRACT: Different number of input and output \
	    spectra",l_logfile,verbose+)
	status=1
	goto clean
    }

    ######################################################################
    # Paste input and output files and put it in tmpfile1
    # Input images are in tmpinlist and output images in tmpoutlist
    join(tmpinlist,tmpoutlist,output=tmpfile1,delim=" ",missing="INDEF",
	maxchars=161,shortest+,verbose-)

    # Run apall or extract a line 
    print(l_key_dispaxis) | scan(l_key_dispaxis)
    scanfile=tmpfile1
    while (fscan(scanfile,imgin,imgout) != EOF) {
        # Create tmp FITS file name used within this loop
	varb=mktemp("tmpvarb")
	dqb=mktemp("tmpdqb")
	tsci=mktemp("tmpsci")
	scispec=mktemp("tmpscispec")
	varspec=mktemp("tmpvarspec")
	dqspec=mktemp("tmpdqspec")
	
	sec = "["//l_sci_ext//"]"
	imgets(imgin//sec,l_key_dispaxis, >& "dev$null")
	if (imgets.value!="0")
	    l_dispaxis=int(imgets.value)
	else
	    l_dispaxis=dispaxis
	specred.dispaxis=l_dispaxis

	imcopy(imgin//"[0]",imgout,verbose-)     


	if (l_fl_apall) {
	    printlog("",l_logfile,l_verbose)
	    imlog=l_database//"/ap"//imgin//"_"//l_sci_ext//"_"
	    if (access(imlog)) delete(imlog,ver-, >& "dev$null")
	    specred.apall(imgin//sec,1,output=speca,format="multispec",
		interactive=l_fl_inter,find+,recenter-,resize-,edit+,trace+,
		fittrace+, extract+,extras+,review+,line=l_line,nsum=l_nsum,
		background=l_background,weights="variance",pfit="fit1d",
		clean-,skybox=1,readnoise=0.,gain=1.,saturation=INDEF,
		lsigma=4.,usigma=4.,nsubaps=1)

	    if (l_fl_findneg) {
		printlog("",l_logfile,l_verbose)
		imdelete("neg_"//imgin,ver-, >& "dev$null")
		imarith(imgin//sec,"*",-1.0,"neg_"//imgin,verbose-)
		specred.apall("neg_"//imgin,1,output=specb,format="multispec",
		    interactive=l_fl_inter,find+,recenter-,resize-,edit+,
		    trace+,fittrace+,extract+,extras+,review+,line=l_line,
		    nsum=l_nsum,background=l_background,weights="variance",
		    pfit="fit1d",clean-,skybox=1,readnoise=0.,gain=1.,
		    saturation=INDEF,lsigma=4.,usigma=4.,nsubaps=1)
		imdelete("neg_"//imgin,ver-)

		sarith(speca,"+",specb,scispec,apertures=1,bands=1,reverse-,
		    format="multispec",verb-)

		# Now get the variance plane from the variance band.            
		imarith(speca//"[*,1,"//varpl//"]","*",
		    speca//"[*,1,"//varpl//"]",speca,verb-)  
		imarith(specb//"[*,1,"//varpl//"]","*",
		    specb//"[*,1,"//varpl//"]",specb,verb-)
		imarith(speca,"+",specb,varspec,verb-,noact-)
		imdelete(specb,ver-)
	    } else {
		imcopy(speca//"[*,1,1]",scispec,verb-)
		imarith(speca//"[*,1,"//varpl//"]","*",
		    speca//"[*,1,"//varpl//"]",varspec,verb-,noact-)
	    }

	    imarith(varspec,"*",0,dqspec,pixtype="int",verb-,noact-)
	    imdelete(speca,ver-)
	} else {
	    # Find positive spectra
	    printlog("",l_logfile,l_verbose)
	    imlog=l_database//"/ap"//imgin//"_"//l_sci_ext//"_"
	    if (access(imlog)) delete(imlog,ver-, >& "dev$null")
	    specred.apfind(imgin//sec,1,apertures="",references="",
		interact=l_fl_inter,find+,recenter-,resize-,edit+,line=l_line,
		nsum=l_nsum,minsep=5.,maxsep=1000.,order="increasing")
	    if (access(imlog)) {
		match("center",imlog,stop-) | scan(sdum, sdum, ycp)
	    } else {
		printlog("ERROR - NSEXTRACT: Cannot find database file "//imlog,
		    l_logfile,verbose+)
		status=1
		goto clean
	    }
	    if (l_lower>0) l_lower=-l_lower
	    iy1=nint(ycp+l_lower)+1
	    iy2=iy1+nint(l_upper-l_lower)-1
	    if (iy1 < 0) {
		printlog("ERROR - NSEXTRACT: image section off image",
		    l_logfile,yes)
		status = 1
		goto clean
	    }

	    imcopy(imgin//sec//"[*,"//iy1//":"//iy2//"]",scispec,verb-)
	    sec = "["//l_var_ext//"]"
	    imcopy(imgin//sec//"[*,"//iy1//":"//iy2//"]",varspec,verb-)
	    sec = "["//l_dq_ext//"]"
	    imcopy(imgin//sec//"[*,"//iy1//":"//iy2//"]",dqspec,verb-)

	    if (l_fl_findneg) {
		printlog("",l_logfile,l_verbose)
		sec = "["//l_sci_ext//"]"
		imdelete("neg_"//imgin,ver-, >& "dev$null")
		imarith(imgin//sec,"*",-1.0,"neg_"//imgin,verbose-)

		# Find negative spectra
		imlog=l_database//"/ap"//"neg_"//imgin
		if (access(imlog)) delete(imlog,ver-, >& "dev$null")
		specred.apfind("neg_"//imgin,1,apertures="",references="",
		    interact=l_fl_inter,find+,recenter-,resize-,edit+,
		    line=l_line,nsum=l_nsum,minsep=5.,maxsep=1000.,
		    order="increasing")
		if (access(imlog)) {
		    match("center",imlog,stop-) | scan(sdum, sdum, ycn)
		    imdelete("neg_"//imgin,verify-, >& "dev$null")
		} else {
		    printlog("ERROR - NSEXTRACT: Cannot find database file "//imlog,
			l_logfile,verbose+)
		    status=1
		    goto clean
		}

		if (l_lower>0) l_lower=-l_lower
		iy1=nint(ycn+l_lower)+1
		iy2=iy1+nint(l_upper-l_lower)-1
		if (iy1 < 0) {
		    printlog("ERROR - NSEXTRACT: image section off image",
			l_logfile,yes)
		    status=1
		    goto clean
		}

		imcopy(imgin//sec//"[*,"//iy1//":"//iy2//"]",specb,verb-)

		y1=ycp-ycn
		y2=mod(y1,int(y1))
		imshift(specb,specb,0.0,y2,inter="linear")
		imarith(scispec,"-",specb,scispec,verb-,noact-)
		imdelete(specb,ver-)

		sec = "["//l_var_ext//"]"
		imcopy(imgin//sec//"[*,"//iy1//":"//iy2//"]",varb,verb-)
		sec = "["//l_dq_ext//"]"
		imcopy(imgin//sec//"[*,"//iy1//":"//iy2//"]",dqb,verb-)

		imarith(varspec,"+",varb,varspec,verb-,noact-)
		imarith(dqspec,"+",dqb,dqspec,verb-,noact-)
		imdelete(varb//","//dqb,ver-)

	    } # end of if(l_fl_findneg)   

	    # If l_fl_project=yes, collapse 2D spectra into 1D
	    if (l_fl_project) {
		imrename(scispec,tsci,verb-)

		imdelete(tdq//".pl",ver-, >& "dev$null")
		imcopy(dqspec,tdq//".pl",verb-)
		imdelete(dqspec,ver-)
		hedit(tsci,"BPM",tdq//".pl",add+,addonly-,del-,veri-,show-,upd+)

		imcombine(tsci,scispec,headers="",bpmasks="",rejmasks="",
		    nrejmasks=imcombtdq,expmasks="",sigmas="",logfile=tmplog,
		    combine="average",reject="none",project=yes,outtype="real",
		    outlimits="",offsets="none",masktype="goodvalue",
		    maskvalue=0,blank=0,scale="none",zero="none",weight="none",
		    nkeep=0)
		imdelete(tsci,ver-)

		type(tmplog, >> l_logfile)
		delete(tmplog,verify-, >& "dev$null") 
		hedit(scispec,"BPM",add-,addonly-,delete+,verify-,show-,update+)

		# multiply by the number of pixels to get the sum
		imgets(scispec,"i_naxis2", >& "dev$null")
		npixr = real(imgets.value)
		imarith(scispec,"*",npixr,scispec,verb-,noact-)

		# variance=N
		tvar=mktemp("tmpvar")
		imrename(varspec,tvar,verb-)
		hedit(tvar,"BPM",tdq//".pl",add+,addonly-,delete-,verify-,show-,
		    update+)

	  	imcombine(tvar,varspec,headers="",bpmasks="",rejmasks="",
		    nrejmasks="",expmasks="",sigmas="",logfile="",
		    combine="average",reject="none",project=yes,outtype="real",
		    outlimits="",offsets="none",masktype="goodvalue",
		    maskvalue=0,blank=0,scale="none",zero="none",weight="none",
		    nkeep=0)
		hedit(varspec,"BPM",add-,addonly-,delete+,verify-,show-,update+)
		delete(tdq//".pl",ver-)

		# multiply by the number of pixels to get the sum
		imarith(varspec,"*",npixr,varspec,verb-,noact-)

		imdelete(tvar,verify-,>& "dev$null")

		# bad pixels have imcombtdq=npixr from imcombine
		expr = "(a=="//npixr//") ? 1 : 0"
		imexpr(expr,dqspec,imcombtdq,dims="auto",intype="auto",
		    outtype="int",refim="auto",bwidth=0,btype="nearest",
		    bpixval=0.,rangech+,verbose+,exprdb="none", >& "dev$null")
		delete(imcombtdq//".pl",ver-)                       
	    } # end of if (l_fl_project)

	} # end of if(l_fl_apall) else

	# Create the output image
	imcopy(scispec,imgout//"["//l_sci_ext//",append]",verbose-)
	imcopy(varspec,imgout//"["//l_var_ext//",append]",verbose-)
	imcopy(dqspec,imgout//"["//l_dq_ext//",append]",verbose-)
	hedit(imgout//"["//l_sci_ext//"]","EXTVER",1,add+,addonly-,delete-,
	    verify-,show-,update+) 
	hedit(imgout//"["//l_var_ext//"]","EXTVER",1,add+,addonly-,delete-,
	    verify-,show-,update+) 
	hedit(imgout//"["//l_dq_ext//"]","EXTVER",1,add+,addonly-,delete-,
	    verify-,show-,update+) 
	imdelete(scispec//","//varspec//","//dqspec,ver-, >& "dev$null")

	# Update header of output image (needs much more!)
	sec = "["//l_sci_ext//"]"
	gemhedit(imgout//"[0]","NSEXTRAC",sdate,"Time stamp for NSEXTRACT")
	hedit(imgout//"[0]","GEM-TLM",sdate,add+,addonly-,delete-,verify-,
	    show-,update+)
    } # end of while(fscan())


    #----------------------------------------------------------------------
clean:
    scanfile=""
    delete(tmpinlist//","//tmpoutlist//","//tmpfile1,ver-, >& "dev$null")
    delete(l_database//"/aplast",ver-, >& "dev$null")
    imdelete(tdq//","//tsci//","//specb,ver-, >& "dev$null")

    specred.logfile=l_logfilesave
    specred.database=l_databasesave
    specred.dispaxis=l_dispaxissave
    specred.verbose=l_verbosesave

    printlog("",l_logfile,l_verbose)
    if (status==0)
	printlog("NSEXTRACT: Exit status good",l_logfile,l_verbose)

    printlog("---------------------------------------------------------------\
	----------------",l_logfile,l_verbose)

end
