# Copyright(c) 2001-2004 Association of Universities for Research in Astronomy, Inc.

procedure nsflat(lampson)

# Derive NIRI bad pixel mask and spectroscopic flat field image
#
# Version  Sept 20, 2002 JJ v1.4 release
#          Aug 19, 2003  KL IRAF 2.12 - new/modified parameters
#                             hedit: addonly
#                             imcombine: headers,bpmasks,expmasks,outlimits
#                                 rejmask->rejmasks, plfile->nrejmasks
#                             imstat: nclip,lsigma,usigma,cache

char   lampson      {"",prompt="Input flat field images"}
char   flatfile     {"",prompt="Output flat field image"}
char   darks        {"",prompt="Input dark images (equal exposure)"}
char   flattitle    {"default",prompt="Title for output flat image"}
char   bpmtitle     {"default",prompt="Title for output bad pixel mask"}
char   bpmfile      {"default",prompt="Name of output bad pixel mask PL file"}
char   statsec      {"default",prompt="Image section to be used for statistics"}
char   fitsec       {"default",prompt="1-d image section for fitting in dispersion direction"}
real   thresh_flo   {0.35,prompt="Lower bad-pixel threshold (fraction of peak) for flats"}
real   thresh_fup   {1.25,prompt="Upper bad-pixel threshold (fraction of peak) for flats"}
real   thresh_dlo   {-20.,prompt="Lower bad-pixel threshold (ADU) for darks"}
real   thresh_dup   {100.,prompt="Upper bad-pixel threshold (ADU) for darks"}
bool   fl_inter     {no,prompt="Set bad pixel cut levels interactively?"}
bool   fl_fixbad    {yes,prompt="Fix bad pixels in the output flat field image?"}
real   fixvalue     {1.0,prompt="Replace bad pixels in output flat with this value"}
char   function     {"spline3",prompt="Fitting function for illumination pattern"}
int    order        {-1,min=-1,prompt="Order of polynomial or number of spline pieces (-1 for default)"}
char   normstat     {"mean",enum="mean|midpt",prompt="Statistic to use to fine-tune normalization."}
char   combtype     {"default",enum="default|average|median",prompt="Type of combine operation"}
char   rejtype      {"avsigclip",enum="none|minmax|ccdclip|crreject|sigclip|avsigclip|pclip",prompt="Type of rejection"}
char   scale        {"none",prompt="Image scaling"}
char   zero         {"none",prompt="Image zeropoint offset"}
char   weight       {"none",prompt="Image weights"}
char   key_exptime  {"EXPTIME",prompt="Exposure time header keyword"}
real   lthreshold   {INDEF,prompt="Lower threshold"}
real   hthreshold   {INDEF,prompt="Upper threshold"}
int    nlow         {1,min=0,prompt="minmax: Number of low pixels to reject"}
int    nhigh        {1,min=0,prompt="minmax: Number of high pixels to reject"}
int    nkeep        {0,min=0,prompt="Minimum to keep or maximum to reject"}  
bool   mclip        {yes,prompt="Use median in sigma clipping algorithms?"}
real   lsigma       {3.,min=0.,prompt="Lower sigma clipping factor"}
real   hsigma       {3.,min=0.,prompt="Upper sigma clipping factor"}
char   snoise       {"0.0",prompt="ccdclip: Sensitivity noise (electrons)"}
real   sigscale     {0.1,min=0.,prompt="Tolerance for sigma clipping scaling correction"}
real   pclip        {-0.5,prompt="pclip: Percentile clipping parameter"}
real   grow         {0.0,min=0.,prompt="Radius (pixels) for neighbor rejection"}
char   key_filter   {"FILTER",prompt="Keyword for filter ID"}
char   key_ron      {"RDNOISE",prompt="Keyword for readout noise in e-"}
char   key_gain     {"GAIN",prompt="Keyword for gain in electrons/ADU"}
char   key_sat      {"SATURATI",prompt="Keyword for saturation in ADU"}
char   key_nonlinear {"NONLINEA",prompt="Header keyword for non-linear regime (ADU)"}
char   sci_ext      {"SCI",prompt="Name or number of science extension"}
char   var_ext      {"VAR",prompt="Name or number of variance extension"}
char   dq_ext       {"DQ",prompt="Name or number of data quality extension"}
bool   fl_vardq     {yes,prompt="Create output variance and data quality frames?"}
char   logfile      {"",prompt="Logfile name"}
bool   verbose      {yes,prompt="Verbose output?"}
int    status       {0,prompt="Exit status (0=good)"}
struct *scanfile    {"",prompt="Internal use only"} 

begin
        
    char   l_lampson, l_darks, l_flatfile, tmpbpmfile, l_bpmfile
    char   l_bpmtitle, l_flattitle, l_key_exptime, l_key_filter, l_filter
    char   l_combtype, l_rejtype, l_scale, l_zero, l_weight, l_normstat
    char   l_key_ron, l_key_gain, l_key_sat
    char   l_snoise, l_logfile, l_title, l_key_nonlinear, l_function
    real   l_ron, l_gain, l_sat, l_lthreshold, l_hthreshold, l_lsigma, l_hsigma
    real   l_grow, l_sigscale, l_pclip, dark1, dark2
    real   l_lower, l_upper, l_sigthresh, l_nonlinear, l_fixvalue
    int    l_nlow, l_nhigh, l_nkeep, l_order, nfiles
    bool   l_fl_dark, l_fl_fixbad, l_fl_vardq
    bool   l_verbose, l_mclip, l_fl_inter, find
    bool   warning, alreadywarned
    char   l_sci_ext, l_var_ext, l_dq_ext
    char   l_sgain, l_sron, l_statsec, l_fitsec
    real   l_thresh_dlo, l_thresh_dup, l_thresh_flo, l_thresh_fup
    char   filelist, scilist, dqlist, infiles[2]
    char   img, dqsum, suf, tmphead
    char   tmpsci, tmpdq, combout[2], combvar[2], combdq[2]
    char   dsub, esub, mask[3], firstimage
    char   project, projfit
    real   norm, stddev, temp
    real   etime[3], rtime
    int    i, j, k, l, nbad, len, dum, xsize, ysize
    int    korder,allotherorder
    char   kstatsec,allotherstatsec
    struct sdate

    #------------------------------------------------------------------------
    # Set default fitting orders and statsecs, by filter:
    korder = 45
    kstatsec="[135:950,10:420]"
    allotherorder = 20
    allotherstatsec="[*,10:420]"

    #------------------------------------------------------------------------
    # Set local parameters
    l_lampson=lampson ; l_darks=darks
    l_flatfile=flatfile ; l_logfile=logfile ; l_key_filter=key_filter
    l_bpmtitle=bpmtitle ; l_flattitle=flattitle ; l_bpmfile=bpmfile
    l_combtype=combtype ; l_rejtype=rejtype ; l_normstat=normstat
    l_scale=scale ; l_zero=zero ; l_weight=weight
    l_key_exptime=key_exptime
    l_lthreshold=lthreshold ; l_hthreshold=hthreshold
    l_nlow=nlow ; l_nhigh=nhigh ; l_nkeep=nkeep
    l_lsigma=lsigma ; l_hsigma=hsigma
    l_verbose=verbose ; l_mclip=mclip
    l_key_gain=key_gain ; l_key_ron=key_ron ; l_key_sat=key_sat
    l_snoise=snoise ; l_sigscale=sigscale ; l_pclip=pclip ; l_grow=grow
    l_sci_ext = sci_ext ; l_dq_ext = dq_ext ; l_var_ext=var_ext
    l_thresh_dlo=thresh_dlo ; l_thresh_dup=thresh_dup
    l_thresh_flo=thresh_flo ; l_thresh_fup=thresh_fup
    l_fl_inter=fl_inter ; l_fl_vardq=fl_vardq 
    l_key_nonlinear=key_nonlinear ; l_fl_fixbad=fl_fixbad 
    l_fixvalue=fixvalue ; l_order=order ; l_function=function
    l_statsec=statsec ; l_fitsec=fitsec

    status=0
    alreadywarned=no
    warning=no
    l_filter="none"

    # Keep parameters from changing by outside world
    cache ("imgets")

    #------------------------------------------------------------------------
    # Define temporary files
    filelist=mktemp("tmpfilelist")
    infiles[1]=mktemp("tmponlist")
    infiles[2]=mktemp("tmpdarklist")
    scilist=mktemp("tmpscilist")
    combout[1]=mktemp("tmpon")
    combout[2]=mktemp("tmpdark")
    combvar[1]=mktemp("tmponvar")
    combvar[2]=mktemp("tmpdrkvar")
    combdq[1]=mktemp("tmpcombdqon")
    combdq[2]=mktemp("tmpcombdqdrk")
    tmpsci=mktemp("tmpsci")
    tmpdq=mktemp("tmpdq")
    tmphead=mktemp("tmphead")
    dsub=mktemp("tmpdsub")
    esub=mktemp("tmpesub")
    mask[1]=mktemp("tmpmask1")
    mask[2]=mktemp("tmpmask2")
    mask[3]=mktemp("tmpmask3")
    tmpbpmfile=mktemp("tmpbpmfile")
    project=mktemp("tmpproject")
    projfit=mktemp("tmpprojfit")
    
    # assign dummy value to tmp file names
    # (they just need to be given a value. real names will be given later.)
    dqsum = "dummy"

    infiles[1]=l_lampson 
    infiles[2]=l_darks

    #------------------------------------------------------------------------
    # Set up log file
    if ((l_logfile=="") || (l_logfile==" ")) {   
	l_logfile=niri.logfile
	if ((l_logfile=="") || (l_logfile==" ")) {  
            l_logfile="niri.log"
            printlog("WARNING - NSFLAT: Both NSFLAT.logfile and niri.logfile \
	        are",l_logfile,verbose+) 
            printlog("                  undefined.  Using niri.log.",
	        l_logfile,verbose+)
	}
    }

    date | scan(sdate)
    printlog("---------------------------------------------------------------\
        -------------",l_logfile,l_verbose)
    printlog("NSFLAT -- "//sdate,l_logfile,l_verbose)
    printlog(" ",l_logfile,l_verbose)

    #------------------------------------------------------------------------
    # Verify that the SCI extension name is not empty, otherwise exit gracefully
    if (l_sci_ext=="" || l_sci_ext==" ") {
        printlog("ERROR - NSFLAT: Extension name sci_ext is missing",
	    l_logfile,verbose+)
        status=1
        goto clean
    } 
    if (l_fl_vardq) {
        if (l_var_ext=="" || l_var_ext==" ") {
	    printlog("ERROR - NSFLAT: Empty variance extension name",
	        l_logfile,verbose+)
	    status=1
	    goto clean
        }
        if (l_dq_ext=="" || l_dq_ext==" ") {
	    printlog("ERROR - NSFLAT: Empty data quality extension name",
	        l_logfile,verbose+)
	    status=1
	    goto clean
        }
    }

    # Set titles
    if (l_flattitle=="default" || l_flattitle=="" || l_flattitle==" ") 
        l_flattitle="FLAT FIELD IMAGE from gemini.niri.nsflat"
    if (l_bpmtitle=="default" || l_bpmtitle=="" || l_bpmtitle==" ") 
        l_bpmtitle="BAD PIXEL MASK from gemini.niri.nsflat"

    # Check if key_exptime is set
    if (l_key_exptime=="" || l_key_exptime==" ")
        printlog("WARNING - NSFLAT: key_exptime not set, cannot check \
	    exposure times.",l_logfile,verbose+)

    # Check threshold values if fl_inter=no
    if (!l_fl_inter) {
        if ((l_thresh_flo > l_thresh_fup) || (l_thresh_dlo > l_thresh_dup)) {
	    printlog("ERROR - NSFLAT: Lower threshold must be lower than the \
	        upper threshold.",l_logfile,verbose+)
            status=1
            goto clean
        }
    }

    # Check if fl_fixbad=yes and fl_vardq=no
    if (l_fl_fixbad && !l_fl_vardq) {
        printlog("WARNING - NSFLAT: Cannot fix bad pixels without bad pixel \
	    mask, but",l_logfile,verbose+)
        printlog("                  fl_vardq=no.  Setting fl_fixbad=no.",
	    l_logfile,verbose+)
        l_fl_fixbad=no
    }  

    # Check for empty statistics section
    if (l_statsec=="" || l_statsec==" ") {
        l_statsec="[*,*]"
        printlog("WARNING - NSFLAT: No statsec defined.  Using "//l_statsec,
	    l_logfile,verbose+)
    }

    # Set fitsec if necessary
    if (l_fitsec=="default") l_fitsec="[*]"
    if (l_fitsec=="" || l_fitsec==" ")
        l_fitsec=substr(l_statsec,1,(stridx(",",l_statsec)-1))//"]"

    #------------------------------------------------------------------------
    # Check output files for empty strings, files that already exist

    if (imaccess(l_flatfile)) {
	printlog("ERROR - NSFLAT: Output file "//l_flatfile//" already exists.",
	    l_logfile,verbose+)
	status=1
	goto clean
    }

    if ((l_lampson=="") || (l_lampson==" ")) {
	printlog("ERROR - NSFLAT: No input lamps-on (or sky) images specified.",
	    l_logfile,verbose+)
	status=1
	goto clean
    }

    if ((l_darks=="") || (l_darks==" "))
	l_fl_dark=no
    else
	l_fl_dark=yes

    # Check if fl_vardq=no and fl_dark=yes
    if (l_fl_dark && !l_fl_vardq) {
	printlog("WARNING - NSFLAT: Darks not needed if no bad pixel mask \
	    is to",l_logfile,verbose+)
	printlog("                  be written (fl_vardq=no).",
	    l_logfile,verbose+)
	l_fl_dark=no
    }

    #--------------------------------------------------------------------------
    # check that files exist and are MEF
    nbad=0
    for (i=1; i<=2; i+=1) {
	if ((i==1) || ((i==2) && (l_fl_dark))) {
	    if (infiles[i] == "" || infiles[i]==" ") {
		printlog("ERROR - NSFLAT: input files not specified",
		    l_logfile,verbose+)
		status=1
		goto clean
	    } 
	    # check existence of list files
	    if (substr(infiles[i],1,1) == "@") {
		len=strlen(infiles[i])
		if (!access(substr(infiles[i],2,len))) {
		    printlog("ERROR - NSFLAT: "//substr(infiles[i],2,len)//\
		        " does not exist.",l_logfile,verbose+)
                    status=1
                    goto clean
	        }
	    }
	    files(infiles[i],sort-, > filelist)

            # Check that all images in the list exist
	    scanfile=filelist
	    nfiles=0
	    while (fscan(scanfile, img) != EOF) {
	        gimverify(img)
	        if (gimverify.status==1) {
                    printlog("ERROR - NSFLAT: File "//img//" not found.",
		        l_logfile,verbose+)
                    nbad+=1
	        } else if (gimverify.status>1) {
                    printlog("ERROR - NSFLAT: File "//img//" not a MEF FITS \
		        image.",l_logfile,verbose+)
                    nbad+=1
	        } else {
                    imgets(img//"[0]","NSCUTSEC", >& "dev$null")
                    if (imgets.value == "0") {
                        printlog("ERROR - NSFLAT: Spectral region of \
			    image "//img//" not extracted with NSCUT.",
			    l_logfile,verbose+)
                        nbad+=1
                    } else {
                        imgets(img//"[0]","NPREPARE", >& "dev$null")
                        if (imgets.value == "0") {
        	            printlog("ERROR - NSFLAT: Image "//img//" not \
			        NPREPAREd.",l_logfile,verbose+)
        	            nbad+=1
                        }
                    }
	        }
	        nfiles+=1

	        # Set default order if set to -1
	        if (i==1 && nfiles==1) {
                    imgets(img//"[0]",l_key_filter, >& "dev$null")
                    if (imgets.value!="0") l_filter=substr(imgets.value,1,1)
                    if (l_order==-1) {
                        if (l_filter=="K") {
        	            l_order=korder
        	            if (l_statsec=="default") l_statsec=kstatsec
                        } else {
        	            l_order=allotherorder
        	            if (l_statsec=="default") l_statsec=allotherstatsec
                        }
                    }
	        }
	    } #end of while loop

	    # check for empty file list
	    if (nfiles==0) {
	        printlog("ERROR - NSFLAT:  No input images meet wildcard \
		    criteria.",l_logfile,verbose+)
	        status=1
	        goto clean
	    }

	    scanfile=""
	    delete(filelist,verify-, >>& "dev$null")
        } #end if(i==1) && etc.
    } 
    #end of for loop

    # Exit if problems found
    if (nbad > 0) {
        printlog("ERROR - NSFLAT: "//nbad//" image(s) either do not exist, \
	    are not MEF files, or",l_logfile,verbose+)
        printlog("                 have not been run through NPREPARE and \
	    NSCUT.",l_logfile,verbose+)
	status=1
	goto clean
    }

    #---------------------------------------------------------------------------
    # Input info to logfile
    printlog("lampson = "//l_lampson,l_logfile,l_verbose)
    if (l_fl_dark) printlog("darks = "//l_darks,l_logfile,l_verbose)
    if (!l_fl_inter) {
        if (l_fl_dark)
	    printlog("thresh_dlo = "//l_thresh_dlo,l_logfile,l_verbose)
        if (l_fl_dark)
	    printlog("thresh_dup = "//l_thresh_dup,l_logfile,l_verbose)
    }
    printlog("fitting function = "//l_function,l_logfile,l_verbose)
    printlog("fit order = "//l_order,l_logfile,l_verbose)
    printlog("1-d fitting section = "//l_fitsec,l_logfile,l_verbose)
    printlog("Normalization statistic = "//l_normstat,l_logfile,l_verbose)
    printlog("Statistics region = "//l_statsec,l_logfile,l_verbose)

    #---------------------------------------------------------------------------
    # Combine, lamps on, then darks

    for (i=1; i<=2; i+=1) {
        # Create tmp FITS file name re-used within this loop
	dqsum=mktemp("tmpdqsum")
	
        if (infiles[i] != "" && infiles[i] != " ") {
	    files(infiles[i],sort-, > filelist)
	    k=0
	    scanfile=filelist
	    # strip out science and dq planes
	    while (fscan(scanfile, img) != EOF) {
	        k=k+1
	        suf = substr(img,strlen(img)-4,strlen(img))
	        # Add the .fits suffix to the input file:
	        if (suf!=".fits" && imaccess(img//".fits"))
                    img = img//".fits"
	        if ((i==1) && (k==1)) {
                    firstimage=img
                    if ((l_flatfile=="") || (l_flatfile==" ")) {
                        suf = substr(img,strlen(img)-4,strlen(img))
                        if (suf==".fits")
        	            l_flatfile=substr(img,1,strlen(img)-5)//"_flat" 
                        else
        	            l_flatfile=img//"_flat"
                        if (l_fl_vardq) {
        	            if (l_bpmfile=="default" || l_bpmfile=="" || l_bpmfile==" ") {
        	                l_bpmfile=substr(l_flatfile,1,strlen(l_flatfile)-4)+"bpm.pl"
                            }
                        }
                    } else {
                        if (l_bpmfile=="default" || l_bpmfile=="" || l_bpmfile==" ") {
        	            l_bpmfile=l_flatfile+"_bpm.pl"
                        }
                    }
                    printlog("flatfile = "//l_flatfile,l_logfile,l_verbose)
                    printlog("bpmfile = "//l_bpmfile,l_logfile,l_verbose)
                    if (imaccess(l_flatfile)) {
                        printlog("ERROR - NSFLAT: Output file "//l_flatfile//\
			    " already exists.",l_logfile,verbose+)
                        status=2
                        goto clean
                    }
                    if (imaccess(l_bpmfile)) {
                        printlog("ERROR - NSFLAT: Output BPM "//l_bpmfile//\
			    " already exists.",l_logfile,verbose+)
                        status=2
                        goto clean
                    }
	            imcopy(img//"[0]",l_flatfile,verbose-)
	        }

                # check gain, readnoise, and saturation values
	        imgets(img//"[0]",l_key_gain, >>& "dev$null")
	        if (imgets.value == "0" || imgets.value=="") {
                    printlog("ERROR - NSFLAT: Could not get gain from header \
		        of image "//img,l_logfile,verbose+)
                    status=1
                    goto clean
	        } else if((k==1) && (i==1)) {
                    l_gain=real(imgets.value)
                    l_sgain=imgets.value
	        } else if(!alreadywarned) {
                    if (abs(real(imgets.value)-l_gain) > 0.5) {
                        printlog("WARNING - NSFLAT: gain values are \
			    different.  Continuing.",l_logfile,verbose+)
                        warning=yes
                    }
	        }

	        imgets(img//"[0]",l_key_ron, >>& "dev$null")
	        if (imgets.value == "0" || imgets.value=="") {
                    printlog("ERROR - NSFLAT: Could not get read noise from \
		        header of image "//img,l_logfile,verbose+)
                    status=1
                    goto clean
	        } else if ((k==1) && (i==1)) {
        	    l_ron=real(imgets.value)
        	    l_sron=imgets.value
	        } else if (!alreadywarned) {
                    if (abs(real(imgets.value)-l_ron) > 0.5) {
                        printlog("WARNING - NSFLAT: read noise values are \
			    different.  Continuing.",l_logfile,verbose+)
                        warning=yes
                    }
	        }

	        imgets(img//"[0]",l_key_sat, >>& "dev$null")
	        if (imgets.value == "0" || imgets.value=="") {
                    printlog("ERROR - NSFLAT: Could not get saturation level \
		        from header of image "//img,l_logfile,verbose+)
                    status=1
                    goto clean
	        } else if ((k==1) && (i==1)) {
                     l_sat=real(imgets.value)
	        } else if (!alreadywarned) {
                    if (abs(real(imgets.value)-l_sat) > 1.) {
                        printlog("WARNING - NSFLAT: saturation values are \
		            different.  Continuing.",l_logfile,verbose+)
                        warning=yes
                    }
	        }

	        imgets(img//"[0]",l_key_nonlinear, >>& "dev$null")
	        if (imgets.value == "0" || imgets.value=="") {
                    printlog("ERROR - NSFLAT: Could not get non-linear level \
		        from header of image "//img,l_logfile,verbose+)
                    status=1
                    goto clean
	        } else if ((k==1) && (i==1)) {
                    l_nonlinear=real(imgets.value)
	        } else if ((i!=3)&&(!alreadywarned)) {
                    if (abs(real(imgets.value)-l_nonlinear) > 1.) {
                        printlog("WARNING - NIFLAT: nonlinear values are \
		            different.",l_logfile,verbose+)
                        warning=yes
                    }      
	        }
	        if (warning) alreadywarned=yes

	        # science extension
	        print(img//"["//l_sci_ext//"]", >> scilist)
	        # DQ extension - set up for imcombine
	        if (imaccess(img//"["//l_dq_ext//"]")) {
                    imcopy(img//"["//l_dq_ext//"]",tmpdq//"_"//k//".pl",
		        verbose-)
                    hedit(img//"["//l_sci_ext//"]","BPM",tmpdq//"_"//k//".pl",
		        add+,addonly-,delete-,verify-,show-,update+)
	        }
	    } # end while(fscan) 
	    scanfile=""

            # combine images----------------------------
	    #special cases for low numbers of images
	    if (k == 1) {
	        printlog("WARNING - NSFLAT: only one input image.",
	            l_logfile,verbose+)
	        imcopy(img//"["//l_sci_ext//"]",combout[i],verbose-)
	        if (imaccess(img//"["//l_dq_ext//"]") && imaccess(img//"["//l_var_ext//"]"))  {
                    imcopy(img//"["//l_dq_ext//"]",combdq[i],verbose-)
                    imcopy(img//"["//l_var_ext//"]",combvar[i],verbose-)
	        } else if (l_fl_vardq) {
        	    l_fl_vardq=no
        	    l_fl_fixbad=no
        	    l_fl_dark=no
        	    printlog("WARNING - NSFLAT: only one input image without \
		        VAR and DQ planes.",l_logfile,verbose+)
        	    printlog("                  Proceeding with fl_vardq=no \
		        and fl_fixbad=no",l_logfile,verbose+)
	        }
	    } else {
	        if (k <= 4) {
                    if (k == 2) {
                        printlog("WARNING - NSFLAT: only two images to \
			    combine, turning off rejection.",l_logfile,verbose+)
                        l_rejtype="none"
                    } else
                        printlog("WARNING - NSFLAT: four or less images to \
		            combine.",l_logfile,verbose+)
	        }
	        if (l_combtype=="default") l_combtype="average"
		imcombine("@"//scilist,combout[i],headers="",bpmasks="",
		    rejmasks="",nrejmasks=combdq[i],expmasks="",
		    sigmas=combvar[i],logfile="STDOUT",combine=l_combtype,
		    reject=l_rejtype,project=no,outtype="real",outlimits="",
		    offsets="none",masktype="goodvalue",maskvalue=0,blank=0,
		    scale=l_scale,zero=l_zero,weight=l_weight,statsec=l_statsec,
		    expname=l_key_exptime,lthreshold=l_lthreshold,
		    hthreshold=l_hthreshold,nlow=l_nlow,nhigh=l_nhigh,
		    nkeep=l_nkeep,mclip=l_mclip,lsigma=l_lsigma,hsigma=l_hsigma,
		    rdnoise=l_sron,gain=l_sgain,snoise=l_snoise,
		    sigscale=l_sigscale,pclip=l_pclip,grow=l_grow,
		    >>& "dev$null$")

	        # make variance image by squaring combvar
	        imarith(combvar[i],"*",combvar[i],combvar[i],pixtype="real",
	            verbose-)
	        # bad pixels have dqsum=n
	        imexpr("(a=="//k//") ? 1 : 0",dqsum,combdq[i],outtype="short",
	            verbose-)
	        imdelete(combdq[i],verify-, >>& "dev$null")
	        imcopy(dqsum,combdq[i],verbose-)
	        imdelete(dqsum,verify-, >>& "dev$null")
	    } # end else 

	    hedit("@"//scilist,"BPM","",add-,addonly-,delete+,verify-,show-,
	        update+)
	    delete(filelist,verify-, >>& "dev$null")
	    delete(scilist,verify-, >>& "dev$null")
	    imdelete(tmpdq//"*.pl",verify-, >>& "dev$null")

	    hedit(combout[i],"BPM",add-,addonly-,delete+,verify-,show-,update+)

        } # end if(infiles[i]!="" etc)

    } # end of for loop

    #--------------------------------------------------------------------------
    # subtract and normalize flats, combine bad pixel masks

    # Create more tmp FITS file names
    dqsum=mktemp("tmpdqsum")

    # darks-------------------------

    if (l_fl_dark) {
        if (l_fl_inter) {
	    imhistogram(combout[2],z1=-100,z2=500,binwidth=INDEF,nbins=512,
	        autoscale+,top_closed-,hist_type="normal",listout-,
		plot_type="line",logy+,device="stdgraph")
	    print("Dark histogram")
	    print("Hit any key to mark lower limit for good pixels")
	    dum=fscan(gcur,l_lower)
	    print("Hit any key to mark upper limit for good pixels")
	    dum=fscan(gcur, l_upper)
	    print(" ")
	    if (l_lower>l_upper) {
	        temp=l_upper
	        l_upper=l_lower
	        l_lower=temp
	    }
        } else {
	    l_lower=l_thresh_dlo
	    l_upper=l_thresh_dup
        }
        imexpr("((a<"//l_lower//") || (a>"//l_upper//") || (b>0)) ? 1 : 0",
	    mask[2],combout[2],combdq[2],outtype="short",verbose-)
        dark1=l_lower
        dark2=l_upper
    } # end if(l_fl_dark)

    # flat--------------------------

    if (l_fl_dark) {
        imstat(combout[1]//l_statsec,fields="midpt,stddev",lower=100.,
	    upper=l_sat,nclip=0,lsigma=INDEF,usigma=INDEF,binwidth=0.1,format-,
	    cache-) | scan(temp,stddev)
        imstat(combout[1]//l_statsec,fields="midpt",lower=(temp-3*stddev),
	    upper=(temp+3*stddev),nclip=0,lsigma=INDEF,usigma=INDEF,
	    binwidth=0.1,format-,cache-) | scan(temp)
        imstat(combout[2]//l_statsec,fields="midpt,stddev",lower=100.,
	    upper=l_sat,nclip=0,lsigma=INDEF,usigma=INDEF,binwidth=0.1,format-,
	    cache-) | scan(norm,stddev)
        imstat(combout[2]//l_statsec,fields="midpt",lower=(norm-3*stddev),
	    upper=(norm+3*stddev),nclip=0,lsigma=INDEF,usigma=INDEF,
	    binwidth=0.1,format-,cache-) | scan(norm)
         if (norm==INDEF || temp==INDEF) {
	    printlog("ERROR - NSFLAT: Statistics failed, possibly due to a \
	        bad statsec.",l_logfile,verbose+)
	    status=1
	    goto clean
         }

         if (temp<norm) {
	     printlog("WARNING - NSFLAT:  Flux level in lampson is lower than \
	         in darks.",l_logfile,verbose+)
	     printlog("                   Proceeding assuming these should be \
	         reversed.",l_logfile,verbose+)
	     imarith(combout[2],"-",combout[1],dsub,pixtype="real",verbose-)
         } else {
	     imarith(combout[1],"-",combout[2],dsub,pixtype="real",verbose-)
         }
         norm=0.

         if (l_fl_vardq) {
	     imexpr("(a>"//l_sat//") ? 4 : ((a>"//l_nonlinear//") ? 2 : 0)",
	         mask[3],combout[1],outtype="short",verbose-) 
	     imarith(combvar[1],"+",combvar[2],esub,pixtype="real",verbose-)
	     addmasks(combdq[1]//","//combdq[2]//","//mask[3],dqsum//".fits",
	         "im1 || im2 || im3",flags=" ")
        }
    } else {
        imcopy(combout[1],dsub,verbose-)
        if (l_fl_vardq) {
	    imexpr("(a>"//l_sat//") ? 4 : ((a>"//l_nonlinear//") ? 2 : 0)",
	        mask[3],combout[1],outtype="short",verbose-) 
	    imcopy(combvar[1],esub,verbose-)
	    imcopy(combdq[1],dqsum,verbose-)
        }
    }

    if (l_fl_vardq)
        nmisc.fixpix(dsub,dqsum,linterp="",cinterp="",ninterp="1",creplace="",
	    >>& "dev$null")

    improject(dsub,project,2,average=yes,highcut=l_sat,lowcut=0.,verbose-)
    fit1d (project//l_fitsec,projfit,"fit",inter=l_fl_inter,sample="*",
        naverage=1,function=l_function, order=l_order,low_reject=3.,
	high_reject=3.,niterate=3,grow=0.)
    imarith(dsub,"/",projfit,dsub,verbose-)

    # divide variance by projfit**2 without an intermediate file
    if (l_fl_vardq) {
        imarith(esub,"/",projfit,esub,verbose-)
        imarith(esub,"/",projfit,esub,verbose-)
    }

    # Get normalization and good pixel limits for flat
    if (l_fl_inter) {
        imhistogram(dsub//l_statsec,z1=INDEF,z2=INDEF,binwidth=INDEF,nbins=512,
	    autoscale+,top_closed-,hist_type="normal",listout-,
	    plot_type="line",logy+,device="stdgraph")
	print("Flat histogram")
	print("Hit any key to mark lower limit for good pixels")   
	dum=fscan(gcur,l_lower)
	print("Hit any key to mark upper limit for good pixels")
	dum=fscan(gcur, l_upper)
	print(" ")
	if (l_lower>l_upper) {
	    temp=l_upper
	    l_upper=l_lower
	    l_lower=temp
	}
	imstat(dsub//l_statsec,fields=l_normstat,lower=l_lower,upper=l_upper,
	    nclip=0,lsigma=INDEF,usigma=INDEF,binwidth=0.1,format-,cache-) | \
	    scan(norm)
    } else {
	imstat(dsub//l_statsec,fields="mode,stddev",lower=100.,upper=l_sat,
	    nclip=0,lsigma=INDEF,usigma=INDEF,binwidth=0.1,format-,cache-) | \
	    scan(norm,stddev)
	imstat(dsub//l_statsec,fields="midpt",lower=(norm-stddev*3),
	    upper=(norm+stddev*3),nclip=0,lsigma=INDEF,usigma=INDEF,
	    binwidth=0.1,format-,cache-) | scan(norm)
	if (norm==INDEF) {
	    printlog("ERROR - NSFLAT: Statistics failed, possibly due to a \
	        bad statsec.",l_logfile,verbose+)
	    status=1
	    goto clean
	}

	l_lower=norm*l_thresh_flo
	l_upper=norm*l_thresh_fup
	if (l_upper > l_sat) {
	    printlog("WARNING - NSFLAT: Upper threshold for good pixels in \
	        the flat field image",l_logfile,verbose+)
	    printlog("                  is larger than the saturation value.  \
	        Resetting upper",l_logfile,verbose+)
	    printlog("                  limit to the saturation value.",
	        l_logfile,verbose+)
	    l_upper=l_sat
	}
	imstat(dsub//l_statsec,fields=l_normstat,lower=l_lower,upper=l_upper,
	    nclip=0,lsigma=INDEF,usigma=INDEF,binwidth=0.1,form-,cache-) | \
	    scan(norm)
	if (norm==INDEF) {
	    printlog("ERROR - NSFLAT: Statistics failed, possibly due to a \
	        bad statsec.",l_logfile,verbose+)
	    status=1
	    goto clean
	}
    }
    imdelete(project//","//projfit,verify-, >>& "dev$null")

    # combine masks------------------
    if (l_fl_vardq) {
	imexpr("((a<"//l_lower//") || (a>"//l_upper//") || (b>0)) ? 1 : 0",
	    mask[1],dsub,dqsum,outtype="short",verbose-)
	if (l_fl_dark)
	    addmasks(mask[1]//","//mask[2],tmpbpmfile//".pl","im1 || im2",
	        flags=" ")
        else
	    imcopy(mask[1],tmpbpmfile//".pl",verbose-)
    }

    # fix bad pixels in flat image
    if (l_fl_fixbad) {
	imcopy(dsub,tmpsci,verbose-)
	imdel(dsub,verify-, >>& "dev$null")
	imexpr("(a==1) ? "//l_fixvalue//" : b",dsub,tmpbpmfile,tmpsci,
	    outtype="real",verbose-)
    }

    # make output MEF file-----------
    hedit(dsub,"i_title",l_flattitle,add+,addonly-,delete-,verify-,
        show-,update+)
    imcopy(dsub,l_flatfile//"["//l_sci_ext//",append]",verbose-)
    hedit(l_flatfile//"["//l_sci_ext//"]","EXTVER",1,add+,addonly-,delete-,
	verify-,show-,update+)

    if (l_fl_vardq) {
	hedit(esub,"i_title",l_flattitle,add+,addonly-,delete-,verify-,
	    show-,update+)
	hedit(tmpbpmfile,"i_title",l_bpmtitle,add+,addonly-,del-,ver-,
	    show-,update+)
	imcopy(esub,l_flatfile//"["//l_var_ext//",append]",verbose-)
	imcopy(tmpbpmfile,l_flatfile//"["//l_dq_ext//",append]",verbose-)
	hedit(l_flatfile//"["//l_var_ext//"]","EXTVER",1,add+,addonly-,delete-,
	    verify-,show-,update+)
	hedit(l_flatfile//"["//l_dq_ext//"]","EXTVER",1,add+,addonly-,delete-,
	    verify-,show-,update+)
    }

    #--------------------------------------------------------------------------
    # update headers in final image PHU
    l_flatfile=l_flatfile//"[0]"
    gemhedit(l_flatfile,"GEM-TLM",sdate,"Last modification with GEMINI")
    printf("%-8s= \'%-18s\' / %-s\n","NSFLAT",sdate,"Time stamp for NSFLAT",
        >> tmphead)
    printf("%-8s= %20.5f / %-s\n","NSFLTNRM",norm,
        "Normalization constant for NSFLAT", >> tmphead)
    printf("%-8s= \'%-18s\' / %-s\n","NSFLTNST",l_normstat,
        "Normalization statistic for NSFLAT", >> tmphead)
    printf("%-8s= \'%-18s\' / %-s\n","NSFLTFUN",l_function,
        "Function fit to illumination by NSFLAT", >> tmphead)
    printf("%-8s= \'%-18s\' / %-s\n","NSFLTORD",l_order,
        "Order of function fit by NSFLAT", >> tmphead)
    printf("%-8s= %20.5f / %-s\n","NSFLTFLO",l_lower,
        "Lower limit for good pixels (fraction)", >> tmphead)
    printf("%-8s= %20.5f / %-s\n","NSFLTFUP",l_upper,
        "Upper limit for good pixels (fraction)", >> tmphead)
    printf("%-8s= \'%-18s\' / %-s\n","NSFLTSTA",l_statsec,
        "Statistics section used by NSFLAT", >> tmphead)
    printf("%-8s= \'%-18s\' / %-s\n","NSFLTFIT",l_fitsec,
        "Limits of flat in dispersion dimension ", >> tmphead)
    if (l_fl_dark) {
        dark1=(int(dark1*100.))/100.
        dark2=(int(dark2*100.))/100.
        printf("%-8s= %20.5f / %-s\n","NSFLTDLO",dark1,
	    "Lower limit for good pixels in dark (ADU)", >> tmphead)
        printf("%-8s= %20.5f / %-s\n","NSFLTDUP",dark2,
	    "Upper limit for good pixels in dark (ADU)", >> tmphead)
    }
    if (l_fl_fixbad)
        printf("%-8s= %20.5f / %-s\n","NSFLTFIX",l_fixvalue,
	    "Bad pixels replaced with this value", >> tmphead)
    printf("%-8s= \'%-18s\' / %-s\n","NSFLTCOM",l_combtype,
        "Combine method used by NSFLAT", >> tmphead)
    if (l_zero!="none")
	printf("%-8s= \'%-18s\' / %-s\n","NSFLTZER",l_zero,
            "Statistic used to compute additive offsets", >> tmphead)
    if (l_scale!="none")
	printf("%-8s= \'%-18s\' / %-s\n","NSFLTSCL",l_scale,
            "Statistic used to compute scale factors", >> tmphead)
    if (l_weight!="none")
	printf("%-8s= \'%-18s\' / %-s\n","NSFLTWEI",l_weight,
            "Statistic used to compute relative weights", >> tmphead)
    if (l_rejtype!="none") {
        printf("%-8s= \'%-18s\' / %-s\n","NSFLTREJ",l_rejtype,
            "Rejection algorithm used by NSFLAT", >> tmphead)
	printf("%-8s= %20.5f / %-s\n","NSFLTLTH",l_lthreshold,
            "Lower threshold before combining", >> tmphead)
	printf("%-8s= %20.5f / %-s\n","NSFLTHTH",l_hthreshold,
            "Upper threshold before combining", >> tmphead)
	printf("%-8s= %20.5f / %-s\n","NSFLTGRW",l_grow,
            "Radius for additional pixel rejection", >> tmphead)
    }
    if (l_rejtype=="minmax") {
	printf("%-8s= %20.0f / %-s\n","NSFLTNLO",l_nlow,
	    "Low pixels rejected (minmax)", >> tmphead)
	printf("%-8s= %20.0f / %-s\n","NSFLTNHI",l_nhigh,
	    "High pixels rejected (minmax)", >> tmphead)
    }
    if (l_rejtype=="sigclip" || l_rejtype=="avsigclip" || l_rejtype=="ccdclip" \
        || l_rejtype=="crreject" || l_rejtype=="pclip") {
	printf("%-8s= %20.5f / %-s\n","NSFLTLSI",l_lsig,
	    "Lower sigma for rejection", >> tmphead)
	printf("%-8s= %20.5f / %-s\n","NSFLTHSI",l_hsig,
	    "Upper sigma for rejection", >> tmphead)
	printf("%-8s= %20.0f / %-s\n","NSFLTNKE",l_nkeep,
	    "Min(max) number of pixels to keep(reject)", >> tmphead)
	printf("%-8s= \'%-18b\' / %-s\n","NSFLTMCL",l_mclip,
	    "Use median in clipping algorithms", >> tmphead)
    } 
    if (l_rejtype=="sigclip" || l_rejtype=="avsigclip" || l_rejtype=="ccdclip" \
	|| l_rejtype=="crreject")
	printf("%-8s= %20.5f / %-s\n","NSFLTSSC",l_sigscale,
	    "Tolerance for sigma clip scaling correction", >> tmphead)
    if (l_rejtype=="pclip")
	printf("%-8s= %20.5f / %-s\n","NSFLTPCL",l_pclip,
	    "Percentile clipping factor used by pclip", >> tmphead)
    if (l_rejtype=="ccdclip")
	printf("%-8s= \'%-18s\' / %-s\n","NSFLTSNO",l_snoise,
	    "Sensitivity noise (e) used by ccdclip", >> tmphead)

    delete(filelist,verify-, >>& "dev$null")
    files(infiles[1],sort-, > filelist)     
    i=1
    scanfile=filelist
    while (fscan(scanfile,img) != EOF) {
        if (substr(img,1,strlen(img)-5) != substr(l_flatfile,1,strlen(l_flatfile)-3)) {
	    printf("%-8s= \'%-18s\' / %-s\n","NSFLON"//str(i),img,
	        "Input flat image", >> tmphead)
	    i+=1
        }
    }
    delete (filelist,verify-, >& "dev$null")
    if (l_fl_dark) {
	files(infiles[2],sort-, > filelist)
	k=1
	scanfile=filelist
	while (fscan(scanfile,img) != EOF) {
	    if (substr(img,1,strlen(img)-5) != substr(l_flatfile,1,strlen(l_flatfile)-3)) {
                printf("%-8s= \'%-18s\' / %-s\n","NSFLOF"//str(k),img,
		    "Input dark", >> tmphead)
                k+=1
	    }
	}
    }

    # put all the new stuff in the header
    mkheader(l_flatfile,tmphead,append+,verbose-)

    # These are not relevant for spectral flats
    hedit(l_flatfile,l_key_ron,add-,addonly-,delete+,verify-,show-,update+)
    hedit(l_flatfile,l_key_sat,add-,addonly-,delete+,verify-,show-,update+)
    hedit(l_flatfile,l_key_nonlinear,add-,addonly-,delete+,verify-,
        show-,update+)
    hedit(l_flatfile,l_key_gain,add-,addonly-,delete+,verify-,show-,update+)

    #--------------------------------------------------------------------------
    # Write the output BPM file, with compete header

    if (l_fl_vardq) {
	imgets(l_flatfile,"ORIGXSIZ", >& "dev$null")
	if (imgets.value!="0") {
	    xsize=int(imgets.value)
	    imgets(l_flatfile,"ORIGYSIZ", >& "dev$null")
	    ysize=int(imgets.value)
	} else {
	    printlog("WARNING - NSFLAT: Can't determine original raw image \
	        size.",l_logfile,verbose+)
	    printlog("                  Using 1024x1024.",l_logfile,verbose+)
	    xsize=1024
	    ysize=1024
        }
        if (substr(l_bpmfile,strlen(l_bpmfile)-2,strlen(l_bpmfile)) != ".pl") 
	    l_bpmfile=l_bpmfile//".pl"
	imgets(l_flatfile,"NSCUTSEC", >& "dev$null")
	mkimage(l_bpmfile,"make",0,ndim=2,dims=xsize//" "//ysize,
	    pixtype="short")
	l_flatfile=substr(l_flatfile,1,strlen(l_flatfile)-3)
	hedit(l_flatfile//"["//l_dq_ext//"]","INHERIT","T",add-,addonly-,
	    delete-,verify-,show-,update+)
	imcopy(l_flatfile//"["//l_dq_ext//"]",l_bpmfile//imgets.value,
	    >& "dev$null")
	hedit(l_flatfile//"["//l_dq_ext//"]","INHERIT","F",add-,addonly-,
	    delete-,verify-,show-,update+)
    }



clean:
    #--------------------------------------------------------------------------
    # clean up

    delete(filelist//","//scilist//","//tmphead,verify-, >>& "dev$null")
    for (i=1; i<=2; i+=1) {
	imdelete(combout[i]//","//combvar[i]//","//combdq[i],verify-,
	    >>& "dev$null")
    }
    imdelete(tmpsci//"*.fits,"//tmpdq//"*.pl,"//dqsum//","//dsub//","//esub,verify-, >>& "dev$null")
    imdelete(tmpbpmfile,ver-, >& "dev$null")
    imdelete(mask[1]//","//mask[2]//","//mask[3],verify-, >>& "dev$null")
    if (status==1) imdelete(l_flatfile,verify-, >>& "dev$null")
    scanfile=""
    # close log file
    if (status==0) {
	printlog(" ",l_logfile,l_verbose)
	printlog("NSFLAT exit status:  good.",l_logfile,l_verbose)
    } else {
	printlog(" ",l_logfile,l_verbose)
	printlog("NSFLAT exit status:  error.",l_logfile,l_verbose)
    }
    printlog("----------------------------------------------------------------\
        ------------",l_logfile,l_verbose)

end
