# Copyright(c) 2000-2004 Association of Universities for Research in Astronomy, Inc.

procedure nssdist(inimages)

# Measure the S-distortion of NIRI spectra
# This task establishes the calibration using stepped images of a star 
# along a slit, or perhaps data from a "pinhole" slit
#
# Version: Sept 20, 2002 JJ v1.4 release
#          Aug 19, 2003  KL IRAF2.12 - new parameter, addonly, in hedit

char  inimages     {prompt="Input NIRI spectra (pinhole or star spectra)"}
char  outsuffix    {"_sdist",prompt="Output spectrum suffix"}
char  key_xoff     {"GXOFF",prompt="Header keyword for instrument x offset (in arcsec)"}
char  key_yoff     {"GYOFF",prompt="Header keyword for instrument y offset (in arcsec)"}
char  key_pixscale {"PIXSCALE",prompt="Image header keyword of pixel scale [arcsec/pix]"}
real  pixscale     {1.,prompt="Pixel scale in arcsec/pixel"}
char  key_dispaxis {"DISPAXIS",prompt="Header keywordd for dispersion axis"}
int   dispaxis     {1,min=1,max=2,prompt="Dispersion axis"}
char  database     {"database",prompt="Directory for files containing feature data"}
real  firstycoord  {0.,min=0.,prompt="y coord of star in first input image"}
char  coordlist    {"",prompt="y-coordinate list for pinhole spectra"}
bool  fl_inter     {no,prompt="Examine identifications interactively"}
bool  fl_dbwrite   {yes,prompt="Write results to database"}
char  section      {"default",prompt="Image section for running identify"}
int   nsum         {30,min=1,prompt="Number of lines or columns to sum"}
char  ftype        {"emission",min="emission|absorption",prompt="Feature type"}
real  fwidth       {10.,min=2,prompt="Feature width in pixels"}
real  cradius      {10.,min=2,prompt="Centering radius in pixels"}
real  threshold    {50.,prompt="Feature threshold for centering"}
real  minsep       {10.,prompt="Minimum pixel separation for features"}
real  match        {-6.,prompt="Coordinate list matching limit, <0 pixels, >0 user"}
char  function     {"chebyshev",min="legendre|chebyshev|spline1|spline3",
                   prompt="Coordinate fitting function"}
int   order        {2,min=1,prompt="Order of coordinate fitting function"}
char  sample       {"",prompt="Coordinate sample regions"}
int   niterate     {3,min=0,prompt="Rejection iterations"}
real  low_reject   {5.,min=0,prompt="Lower rejection sigma"}
real  high_reject  {5.,min=0,prompt="Upper rejection sigma"}
real  grow         {0.,min=0,prompt="Rejection growing radius"}
bool  refit        {yes,prompt="Refit coordinate function when running reidentify"}
int   step         {10,prompt="Steps in lines or columns for reidentification"}
bool  trace        {no,prompt="Use fit from previous step rather than central aperture"}
int   nlost        {10,min=0,prompt="Maximum number of lost features"}
char  aiddebug     {"l",prompt="Debug parameter for aidpars"}
char  sci_ext      {"SCI",prompt="Name or number of science extension"}
char  var_ext      {"VAR",prompt="Name or number of variance extension"}
char  dq_ext       {"DQ",prompt="Name or number of data quality extension"}
char  logfile      {"",prompt="Logfile"}
bool  verbose      {yes,prompt="Verbose"}
int   status       {0,prompt="Exit status (0=Good)"}
struct* scanfile   {"",prompt="Internal use"}


begin

    char   l_inimages, l_outsuffix, l_outspectra, l_prefix
    char   l_section, l_database, l_coordlist, l_units, l_ftype
    char   l_key_pixscale, l_key_xoff, l_key_yoff
    char   l_function, l_sample
    char   l_sci_ext, l_var_ext, l_dq_ext, l_logfile
    bool   l_fl_inter, l_refit, l_trace
    bool   l_fl_trimmed, l_verbose, l_fl_dbwrite
    real   l_crval, l_cdelt, l_crpix, l_fwidth, l_cradius, l_threshold
    real   l_minsep
    real   l_match, l_low_reject, l_high_reject, l_grow, l_firstycoord
    real   l_pixscale
    int    l_dispaxis, l_nsum, l_order, l_niterate, l_step, l_nlost
    char   l_key_dispaxis, l_aiddebug

    char   l_intemp, l_outtemp, tmpinlist
    char   l_outsec, tmptarg, tmpoffset
    char   s_inter, s_dbwrite
    int    ii, nlamps, nout, l_spataxis, n_lines
    bool   l_flsetarray, l_fl_specsec, l_fl_newout
    struct l_struct
    real   yoff
    char   l_temp, img, imgout, sec, unixsec
    char   tmpfile1, tmpcoordlist, tmphead
    int    nfiles, nbad, count, nin, minfound

    #----------------------------------------------------------------------
    tmpinlist = mktemp("tmpinl")
    tmpfile1 = mktemp("tmpfile1")
    tmphead = mktemp("tmphead")
    tmpcoordlist = mktemp("tmpcoordlist")

    # Initialize parameters
    status=0
    l_fl_newout=no

    # cache parameter files
    cache("imgets","niri","keypar")

    # Set the local variables
    l_inimages=inimages ; l_outsuffix=outsuffix
    l_dispaxis=dispaxis ; l_key_dispaxis=key_dispaxis
    l_section=section 
    if ((l_section=="default" || l_section=="" || l_section==" ") && l_dispaxis==1)
        l_section="middle line"
    if ((l_section=="default" || l_section=="" || l_section==" ") && l_dispaxis==2)
        l_section="middle column"

    l_database=database ; l_coordlist=coordlist
    l_fl_inter=fl_inter ; l_nsum=nsum; l_firstycoord=firstycoord
    l_ftype=ftype ; l_fwidth=fwidth ; l_cradius=cradius ; l_threshold=threshold
    l_minsep=minsep ; l_match=match ; l_function=function ; l_order=order
    l_sample=sample ; l_niterate=niterate
    l_low_reject=low_reject ; l_high_reject=high_reject ; l_grow=grow
    l_refit=refit ; l_step=step ; l_trace=trace ; l_nlost=nlost
    l_sci_ext=sci_ext; l_var_ext=var_ext; l_dq_ext=dq_ext
    l_key_pixscale=key_pixscale; l_key_xoff=key_xoff; l_key_yoff=key_yoff
    l_pixscale=pixscale
    l_fl_dbwrite=fl_dbwrite ; l_logfile=logfile ; l_verbose=verbose 

    if (!l_verbose)
        l_aiddebug=""
    else
        l_aiddebug=aiddebug

    #----------------------------------------------------------------------
    if ((l_logfile=="") || (l_logfile==" ")) {
        l_logfile=niri.logfile
	if ((l_logfile=="") || (l_logfile==" ")) {
	    l_logfile="niri.log"
	     printlog("WARNING - NSSDIST:  Both NSSDIST.logfile and \
	         niri.logfile",logfile=l_logfile, verbose+)
	     printlog("                        are empty.  Using niri.log.",
                 logfile=l_logfile, verbose+)
	}
    }

    # Set up the logfile
    printlog("-----------------------------------------------------------\
	-------------------",l_logfile,l_verbose)
    date | scan(l_struct)
    printlog("NSSDIST -- "//l_struct,l_logfile,l_verbose)
    printlog("",l_logfile,l_verbose)

    if (l_fl_inter) s_inter="YES"
    else s_inter="NO"

    if (l_fl_inter && l_fl_dbwrite)
        s_dbwrite="yes"
    else if(!l_fl_inter && l_fl_dbwrite)
        s_dbwrite="YES"
    else
        s_dbwrite="NO"

    #----------------------------------------------------------------------
    # Make lists of input spectra
    #
    if (substr(l_inimages,1,1)=="@") {
        l_temp=substr(l_inimages,2,strlen(l_inimages))
        if (!access(l_temp)) {
	    printlog("ERROR - NSSDIST:  Input list file "//l_temp//" not \
	        found.",l_logfile,verbose+)
	    status=1
	    goto clean
	}
    }

    files(l_inimages,sort-, > tmpfile1)

    # Verify that input images actually exist. 
    nbad=0; nfiles=0
    scanfile = tmpfile1
    while (fscan(scanfile,img) != EOF) {
	gimverify(img); img = gimverify.outname
	if (gimverify.status==1) {
	    printlog("ERROR - NSSDIST: File "//img//" not found.",
	        l_logfile,verbose+)
	    nbad+=1
	} else if (gimverify.status>1) {
	    printlog("ERROR - NSSDIST: File "//img//" not a MEF FITS image.",
	        l_logfile,verbose+)
	    nbad+=1
	} else {
	    imgets(img//"[0]","NSAPPWAV", >& "dev$null")
	    if (imgets.value == "0") {
	   	printlog("ERROR - NSSDIST: spectrum "//img//" not run through \
		    NSAPPWAVE.",l_logfile,verbose+)
		nbad+=1
	    }
	    imgets(img//"[0]","NSSDIST", >& "dev$null")
	    if (imgets.value != "0") {
	        printlog("WARNING - NSSDIST: Image "//img//" has been run",
		    l_logfile,verbose+)
	        printlog("                   through NSSDIST before.",
		    l_logfile,verbose+)
	        l_fl_newout=no
	        if (l_fl_dbwrite) {
                    printlog("WARNING - NSSDIST: Existing database entries \
		        will be overwritten.",l_logfile,verbose+)
	        }
	    } else
	  	l_fl_newout=yes
        }
        nfiles+=1
        print(img, >> tmpinlist)
    } # end while loop

    # check for empty file list
    if (nfiles==0) {
        printlog("ERROR - NSSDIST:  No input images to process.",
	    l_logfile,verbose+)
        status=1
        goto clean
    }

    # Exit if problems found with input files
    if (nbad > 0) {
        printlog("ERROR - NSSDIST: "//nbad//" image(s) either do not exist, \
	    are not MEF files, or",l_logfile, verbose+)
        printlog("                 have not been run through NSAPPWAVE.",
	    l_logfile,verbose+)
	status=1
      	 goto clean
    } # end if (nbad > 0)

    if (nfiles>1) {
	printlog("Combining input files:",l_logfile,l_verbose)
	if (l_verbose) type(tmpinlist)
	type(tmpinlist, >> l_logfile)
	l_fl_newout=yes
    } else
	printlog("Using input file "//l_inimages,l_logfile,l_verbose)

    delete(tmpfile1, ver-)

    #----------------------------------------------------------------------
    # Find offsets for coordlist, combine input using gemcombine

    if (nfiles>1) {
	tmptarg=mktemp("tmptarg")
	tmpoffset=mktemp("tmpoffset")
	scanfile=tmpinlist
	ii=1
	while (ii==1 && fscan(scanfile,l_temp)!=EOF) {
	    if (l_fl_newout) {
	        l_outspectra=l_temp//l_outsuffix
	        printlog("Using output file "//l_outspectra,l_logfile,l_verbose)
	    } else
	 	l_outspectra=l_temp
	    ii+=1
	    if (imaccess(l_outspectra)) {
		printlog("ERROR - NSSDIST: Output file "//l_outspectra//\
		    " already exists.",l_logfile,verbose+
		status=1
		goto clean
	    }
	}
	#get offsets into headers
	gemoffsetlist("@"//tmpinlist,l_temp,99999.,tmptarg,tmpoffset,
	    offsettype="distance",key_xoff=l_key_xoff,key_yoff=l_key_yoff,
	    key_time="",logfile=l_logfile,verbose-, >& "dev$null")
	if (gemoffsetlist.status!=0) {
	    printlog("ERROR - NSSDIST: Problem determining offsets using \
	        GEMOFFSETLIST.",l_logfile,verbose+)
	    status=1
	    goto clean
	}

	#create coordlist
	l_coordlist=tmpcoordlist
	scanfile=tmpinlist
	while (fscan(scanfile,img)!=EOF) {
	    keypar(img//"[0]",l_key_yoff)
	    if (keypar.found) {
		yoff=real(keypar.value)
		keypar(img//"[0]",l_key_pixscale,silent+)
		if (keypar.found) {
		    yoff=l_firstycoord-(yoff/real(keypar.value))
		    print(yoff, >> l_coordlist
		} else {
		    yoff=l_firstcoord-(yoff/l_pixscale)
		    print(yoff, >> l_coordlist)
		}
	    } else {
		printlog("ERROR - NSSDIST: Can't determine the offset \
		    for file "//img,l_logfile,verbose+)
		status=1
		goto clean
	    }
	    # don't delete these or nssdist cannot be rerun, IJ
	    #    hedit(img//"[0]","GXOFF",add-,addonly-,delete+,verify-,show-,update+)
	    #    hedit(img//"[0]","GYOFF",add-,addonly-,delete+,verify-,show-,update+)
	}
        delete(tmptarg//","//tmpoffset,ver-, >& "dev$null")

	# combine input without rejection
	gemcombine("@"//tmpinlist,l_outspectra,combine="average",reject="none",
	    logfile=l_logfile,statsec="[*,*]",scale="none",zero="none",
	    weight="none",sci_ext=l_sci_ext,var_ext=l_var_ext,dq_ext=l_dq_ext,
	    fl_vardq=no,offsets="none",verbose-, >& "dev$null")
	if (gemcombine.status!=0) {
	    printlog("ERROR - NSSDIST: Problem combining input images using \
	        GEMCOMBINE.",l_logfile,verbose+)
	    status=1
	    goto clean
	}

    } else {
	if (l_fl_newout) {
	    l_outspectra=l_inimages//l_outsuffix
	    if (imaccess(l_outspectra) && l_fl_newout) {
	        printlog("ERROR - NSSDIST: Output file "//l_outspectra//\
		    " already exists.",l_logfile,verbose+)
		status=1
		goto clean
	    }
	    printlog("Using output file "//l_outspectra,l_logfile,l_verbose)
	    copy(l_inimages//".fits",l_outspectra//".fits",verbose-)
	} else
	    l_outspectra=l_inimages

	if (!access(l_coordlist) && !l_fl_inter) {
	    l_fl_inter=yes; s_inter="YES"
	    printlog("WARNING - NSSDIST: only 1 input image without \
	        coordlist.  Running",l_logfile,verbose+)
	    printlog("                   interactively to identify the star \
	        or pinhole spectra.",l_logfile,verbose+) 
	} else if (access(l_coordlist) && !l_fl_inter) {
	    printlog("Using coordinate list "//l_coordlist//" to find the \
	        spectra.",l_logfile,verbose+) 
	}
    }

    #----------------------------------------------------------------------
    # All spectra to be processed are now combined in l_outspectra

    imgout=l_outspectra
    sec = "["//l_sci_ext//"]"

    # set dispaxis to the other axis, set section to match that
    l_spataxis=1
    longslit.dispaxis=1
    if (l_dispaxis==1) {
	longslit.dispaxis=2
	l_spataxis=2
    }
    if (section=="default") {
	l_section="middle column"
	if(l_dispaxis==2) l_section="middle line"
    }

    # Get the approximate wavelength info from the header
    imgets(imgout//sec,"CRPIX"//l_spataxis) 
    l_crpix=real(imgets.value)
    imgets(imgout//sec,"CRVAL"//l_spataxis) 
    l_crval=real(imgets.value)
    imgets(imgout//sec,"CD"//l_spataxis//"_"//l_spataxis) 
    l_cdelt=real(imgets.value)

    # check for enough entries in coordlist
    count(l_coordlist) | scan(n_lines)
    if (n_lines<3) {
	printlog("ERROR - NSSDIST: at least 3 positions are needed.",
	    l_logfile,verbose+)
	status=1
	goto clean
    }

    # write offsets to logfile
    printlog(" ",l_logfile,l_verbose)
    printlog("Coordinate list:",l_logfile,l_verbose)
    type(l_coordlist, >> l_logfile)
    if (l_verbose) type(l_coordlist)

    # delete database entry if fl_dbwrite=yes
    if (l_fl_dbwrite)
	delete(l_database//"/id"//imgout,ver-, >& "dev$null")

    if (!l_fl_inter) {
	# autoidentify think that it is a rotated spectrum - but it isn't
	# so need to clean them. Clean the rest as well. 
	hedit (imgout//sec,"LTV1","",add-,addonly-,delete+,verify-,show-,
	    update+)
	hedit (imgout//sec,"LTV2","",add-,addonly-,delete+,verify-,show-,
	    update+)
	hedit (imgout//sec,"LTM1_1","",add-,addonly-,delete+,verify-,show-,
	    update+)
	hedit (imgout//sec,"LTM2_2","",add-,addonly-,delete+,verify-,show-,
	    update+)
	hedit (imgout//sec,"WAT0_001","",add-,addonly-,delete+,verify-,show-,
	    update+)
	hedit (imgout//sec,"WCSDIM","",add-,addonly-,delete+,verify-,show-,
	    update+)

	if (l_verbose)
	    print("Running AUTOIDENTIFY")
	minfound=max(3,nint(0.75*n_lines))
	autoidentify(imgout//sec,l_crval,l_cdelt,coordlist=l_coordlist,units="",
	    interactive=s_inter,section=l_section,nsum=l_nsum,ftype=l_ftype,
	    fwidth=l_fwidth,cradius=l_cradius,threshold=l_threshold,
	    minsep=l_minsep,match=l_match,function=l_function,order=l_order,
	    sample=l_sample,niterate=l_niterate,low_rejec=l_low_reject,
	    high_reject=l_high_reject,grow=l_grow,dbwrite=s_dbwrite,
	    overwrite=l_fl_dbwrite,database=l_database,verbose=l_verbose,
	    logfile=l_logfile,plotfile="",graphic="stdgraph",cursor="",
	    aidpars.reflist=l_coordlist,aidpars.refspec="",
	    aidpars.crpix=l_crpix,aidpars.cddir="increasing",
	    aidpars.crsearc=INDEF,aidpars.cdsearc=INDEF,aidpars.ntarget=n_lines,
	    aidpars.npatter=5,aidpars.nbins=3,aidpars.ndmax=3,aidpars.nneigh=5,
	    aidpars.aidord=4,aidpars.nfound=10,aidpars.sigma=5,
	    aidpars.minrati=0.005,aidpars.rms=0.3,aidpars.fmatch=0.3,
	    aidpars.debug=l_aiddebug)
    } else {
	specred.logfile=l_logfile
	print("")
	print("NSSDIST - Mark features to use")
	print("   Accept coordinate or assign new coordinate to each feature")
	print("   m - mark feature ")
	print("   ? - help, q - exit ")

	identify(imgout//sec,l_crval,l_cdelt,coordlist=l_coordlist,units="",
	    section=l_section,nsum=l_nsum,ftype=l_ftype,fwidth=l_fwidth,
	    cradius=l_cradius,threshold=l_threshold,minsep=l_minsep,
	    match=l_match,function=l_function,order=l_order,sample=l_sample,
	    niterate=l_niterate,low_reject=l_low_reject,
	    high_reject=l_high_reject,grow=l_grow,autowrite=l_fl_dbwrite,
	    database=l_database,graphics="stdgraph",cursor="")
	specred.logfile=""
    }

    if (!access(l_database//"/id"//imgout)) {
	printlog("ERROR - NSSDIST: No s-distortion solution found",
	    l_logfile,verbose+)
	status=1
	goto clean
    }      

    if (l_verbose) print("Running REIDENTIFY")

    reidentify(imgout//sec,imgout//sec,coordlist=l_coordlist,
        interactive=s_inter,section=l_section,newaps=yes,refit=l_refit,
	trace=l_trace,step=l_step,nsum=l_nsum,shift="0.",search=0.,
	nlost=l_nlost,cradius=l_cradius,threshold=l_threshold,addfeatures=no,
	match=6,maxfeatures=250,minsep=l_minsep,override=l_fl_dbwrite,
	database=l_database,verbose=l_verbose,logfile=l_logfile,plotfile="",
	graphics="stdgraph",cursor="")

    #----------------------------------------------------------------------
    # Update header
    date | scan(l_struct)

    if (l_fl_newout) {
	printf("%-8s= \'%-18s\' / %-s\n","NSSDIST",l_struct,
	    "Time stamp for NSSDIST", >> tmphead)
	ii=1
	scanfile=tmpinlist
	while (fscan(scanfile,img) != EOF) {
	    printf("%-8s= \'%-18s\' / %-s\n","NSSDIS00"+ii,img,
	        "NSSDIST input image", >> tmphead)
	    ii+=1
	}
	# put all the new stuff in the header
	mkheader(imgout//"[0]",tmphead,append+,verbose-)
	hedit(imgout//"[0]","GEM-TLM",l_struct,add+,addonly-,delete-,verify-,
	    show-,update+)
    } else {
	gemhedit(imgout//"[0]","NSSDIST",l_struct,"Time stamp for NSSDIST")
	hedit(imgout//"[0]","GEM-TLM",l_struct,add+,addonly-,delete-,verify-,
	    show-,update+)
    }

    # Remove REFSPEC1 since this otherwise refers to pinholes not wavelength
    hedit(imgout//sec,"REFSPEC1",add-,addonly-,delete+,verify-,show-,update+)

    #----------------------------------------------------------------------
    # Rename the idfile so fitcoords can find it
    # sec is the extension name, e.g. [SCI]               
    if (l_fl_dbwrite) {
	# Using UNIX rm and mv commands because cl delete and rename
	# do not allow file names with [] (e.g. a[x,y]):
	delete(".nswave.csh",ver-, >& "dev$null")
	unixsec = "\["//l_sci_ext//"\]"
	print("rm "//l_database//"/id"//imgout//unixsec, > ".nswave.csh")
	print("mv "//l_database//"/id"//imgout//" "//\
	    l_database//"/id"//imgout//unixsec, >> ".nswave.csh")

        !chmod +x .nswave.csh                    
        !.nswave.csh >& .nswave.log
        delete(".nswave.csh,.nswave.log",ver-, >& "dev$null")
    }

clean:
    #----------------------------------------------------------------------
    # Clean up
    scanfile=""
    delete(tmpinlist,ver-, >>& "dev$null")
    delete(tmpfile1,ver-, >>& "dev$null")
    delete(tmphead//","//tmpcoordlist,ver-, >& "dev$null")
    if (status!=0 && l_fl_newout)
        imdelete(l_outspectra//".fits",ver-, >& "dev$null")

    printlog("",l_logfile,l_verbose)
    if (status==0)
	printlog("NSSDIST Exit status good",l_logfile,l_verbose)

    printlog("------------------------------------------------------------\
    	------------------",l_logfile,l_verbose)

end

